function activations = non_maximum_suppression_pyramid(image, activationdir, threshold, xSuppression, ySuppression, verbose)
% Performs non-maximum suppression on an activation map
%
% Returns a 3 rows * num_results cols array
% Row 1: x (column)
% Row 2: y (row)
% Row 3: activation value

addpath('../../lib/util/');

if ( ~exist(image,'file') )
    error('%s not found\n',image);
end
if ( ~exist(activationdir,'file') )
    error('%s not found\n',activationdir);
end
[pathstr, filename] = fileparts(image);
activationfile = [activationdir filename '.activations'];
if ( ~exist(activationfile,'file') )
    error('%s not found\n', activationfile);
end

fid = fopen(activationfile);
line = fgetl(fid);
[windowWidth, windowHeight] = strtok(line);
windowWidth = str2double(windowWidth);
windowHeight = str2double(windowHeight(2:end));

line = fgetl(fid);
scales = parseScales(line);

activations = [];
index = 1;
while ( ~feof(fid) )
   line = fgetl(fid);
   if ( line ~= -1 )
       [x, rem] = strtok(line);
       [y, rem] = strtok(rem(2:end));
       [scale, rem] = strtok(rem(2:end));
       activation = str2double(rem(2:end));
       if (activation >= threshold)
           activations = [activations [index; str2double(x); str2double(y); str2double(scale); activation; true] ]; %#ok<AGROW>
           index = index + 1;
       end
   end
end
fclose(fid);


peaks = [];
if ( length(activations) ~= 0 )
  while ( sum(activations(6,:)) > 0 )
    i = maxActiveLocation(activations);
    if ( maxInNeighborhood(activations, i, xSuppression, ySuppression, scales) )
      left = activations(2,i);
      top = activations(3,i);
      peaks = [peaks [round(left);round(top);round(windowWidth*activations(4,i));round(windowHeight*activations(4,i));activations(5,i)]];
      activations = suppressNeighborhood(activations, i, xSuppression, ySuppression, scales);
    else
      activations(6,i) = false;
    end
  end
end

if ( ~isempty(peaks) )
    activations = filterActivations(peaks);
end

end


function scales = parseScales(line)
rem = line;
scales = [];
while ( ~isempty(rem) )
    [scale rem] = strtok(rem);
    if ( scale ~= -1 )

        scales = [scales str2double(scale)];
        if ( length(rem) > 1 )
            rem = rem(2:end);
        end
    end
end
end

function index = maxActiveLocation(activations)
    activeLocations = activations(:,activations(6,:)==1);
    [maxval,maxloc] = max(activeLocations(5,:));
    index = activeLocations(1,maxloc);
end

%% Neighborhood stuff
function isMax = maxInNeighborhood(activations, i, xSuppression, ySuppression, scales)
    scale = activations(4,i);
    value = activations(5,i);
    scaledXSuppression = scale * xSuppression;
    scaledYSuppression = scale * ySuppression;
    scaleIndex = find(scales == scale);
    scalesOfInterest = scales(scaleIndex);
    if (scaleIndex ~= 1)
        scalesOfInterest = [scalesOfInterest scales(scaleIndex - 1)];
    end
    if (scaleIndex ~= length(scales))
        scalesOfInterest = [scalesOfInterest scales(scaleIndex + 1)];
    end
    
    isMax = true;
    
    for s = 1:length(scalesOfInterest)
        activationsOfInterest = (activations(4,:) == scalesOfInterest(s)) & (activations(5,:) > value) & ...
            (activations(2,:) > activations(2,i) - scaledXSuppression/2) & ...
            (activations(2,:) < activations(2,i) + scaledXSuppression/2) & ...
            (activations(3,:) > activations(3,i) - scaledYSuppression/2) & ...
            (activations(3,:) < activations(3,i) + scaledYSuppression/2);
        if (sum(activationsOfInterest) > 0)
           isMax = false;
        end
    end
    
end

function suppressed = suppressNeighborhood(activations, i, xSuppression, ySuppression, scales)
    scale = activations(4,i);
    scaledXSuppression = scale * xSuppression;
    scaledYSuppression = scale * ySuppression;
    scaleIndex = find(scales == scale);
    scalesOfInterest = scales(scaleIndex);
    if (scaleIndex ~= 1)
        scalesOfInterest = [scalesOfInterest scales(scaleIndex - 1)];
    end
    if (scaleIndex ~= length(scales))
        scalesOfInterest = [scalesOfInterest scales(scaleIndex + 1)];
    end
    
    for s = 1:length(scalesOfInterest)
        activationsOfInterest = (activations(4,:) == scalesOfInterest(s)) & ...
            (activations(2,:) > activations(2,i) - scaledXSuppression/2) & ...
            (activations(2,:) < activations(2,i) + scaledXSuppression/2) & ...
            (activations(3,:) > activations(3,i) - scaledYSuppression/2) & ...
            (activations(3,:) < activations(3,i) + scaledYSuppression/2);
        if (sum(activationsOfInterest) > 0)
           activations(6,activationsOfInterest) = false;
        end
    end
    
    suppressed = activations;
end




%% Filtering
function filteredactivations = filterActivations(activations)

filteredactivations=zeros(5,1);
activationId = 2;

used = zeros(size(activations,2),1);
for i = 1:size(activations,2)
    if ( ~used(i) )
        overlappingActivations = findOverlaps(activations,i);
        if ( overlappingActivations ~= 0 )
            [ value, index ] = max([activations(5, overlappingActivations) activations(5,i)]);

            if ( index <= size(overlappingActivations) )
                filteredactivations(:,activationId) = activations(:, overlappingActivations(index));
                used(overlappingActivations(index)) = 1;
            else
                filteredactivations(:,activationId) = activations(:, i);
                used(i) = 1;
                for b = 1:size(overlappingActivations)
                   used(overlappingActivations(b)) = 1; 
                end
            end

        else
            filteredactivations(:,activationId) = activations(:,i);
            used(i) = 1;
        end
        activationId = activationId + 1;
    end
end

end

function overlaps = findOverlaps(activations,a)

    bottommargin = 0.25;
    topmargin = 0.25;
    sidemargin = 0.25;
    left = activations(1,a);
    top = activations(2,a);
    width = activations(3,a);
    height = activations(4,a);

    numOverlaps = 0;
    for i = 1:size(activations,2)
       if (i ~= a)
           if ( activations(3,i) <= width && ...
                   activations(1,i) < left + width && ...
                   activations(2,i) < top + height && ...
                   activations(1,i) + activations(3,i) > left && ...
                   activations(2,i) + activations(4,i) > top )

               centreX = activations(1,i) + activations(3,i) / 2;
               centreY = activations(2,i) + activations(4,i) / 2;
               if ( centreX > left + sidemargin * width && ...
                       centreX < left + (1-sidemargin) * width && ...
                       centreY > top + topmargin * height && ...
                       centreY < top + (1-bottommargin) * height )
                   numOverlaps = numOverlaps + 1;
                   overlaps(numOverlaps) = i;
               end
           end
       end
    end
    
    if ( ~exist('overlaps', 'var') )
        overlaps = 0;
    end
end