/*! \file BuildModel.cpp
  Implements the construction of a svm model via libsvm.

  This program takes a set of spatial pyramids from positive examplars
  and a set of spatial pyramids from negative examplars, builds a gram
  matrix using the spatial pyramid kernel, and saves the model.

  @author Sancho McCann
*/

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <boost/progress.hpp>
#include <util/util.hpp>
#include <sift/sift.hpp>
#include <spatialpyramid/spatial_pyramid.hpp>
#include <svm/svm.h>
#include <vector>
#include <string>
#include <limits>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
namespace SpatialPyramid = sjm::SpatialPyramid;
using namespace std;

svm_problem* constructSparseProblem( const vector<SpatialPyramid::util::PathAndPyramid> & positivePyramids,
				     const vector<SpatialPyramid::util::PathAndPyramid> & negativePyramids,
				     int levels, const vector<int> & supportVectors );
svm_problem* constructProblem(const vector<SpatialPyramid::util::PathAndPyramid> & positivePyramids,
			      const vector<SpatialPyramid::util::PathAndPyramid> & negativePyramids, int levels);
void loadPyramid( const fs::path & pyramidPath, vector<SpatialPyramid::util::PathAndPyramid> & pyramids );

int main(int argc, char** argv)
{
  set<string> validExtensions;
  vector<string> positivePathNames;
  vector<string> negativePathNames;
  string svmModelFilename;
  bool recursive = false;
  int pyramidLevels;

  validExtensions.insert(".pyramid");
  validExtensions.insert(".cpyramid");

  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help,H", "produce help message")
    ("negative,N", po::value<vector<string> >(), "negative example images/directories")
    ("positive,P", po::value<vector<string> >(), "positive example images/directories")
    ("recursive,R", "flag to indicate if the image search should recurse on sub-directories")
    ("levels,L", po::value<int>(), "levels of the pyramid to use in the kernel")
    ("output,o", po::value<string>(), "svm_model destination file")
    ;

  try {
    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, command_line_options), vm);
    po::notify(vm);
    
    if ( vm.count("help") || !vm.count("positive") || !vm.count("negative") || !vm.count("output") || !vm.count("levels") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    else
      {
	positivePathNames = vm["positive"].as<vector<string> >();
	negativePathNames = vm["negative"].as<vector<string> >();
	pyramidLevels = vm["levels"].as<int>();
	svmModelFilename = vm["output"].as<string>() + ".model";
      }
    recursive = vm.count("recursive");
  } catch ( po::invalid_command_line_syntax ) {
    cout << command_line_options << endl;
    exit(1);
  }

  vector<SpatialPyramid::util::PathAndPyramid> positivePyramids;
  vector<SpatialPyramid::util::PathAndPyramid> negativePyramids;

  vector<string>::const_iterator positiveIt = positivePathNames.begin();
  for ( ; positiveIt != positivePathNames.end(); ++positiveIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(positivePyramids)),
					      *positiveIt,
					      validExtensions,
					      recursive);
    }

  vector<string>::const_iterator negativeIt = negativePathNames.begin();
  for ( ; negativeIt != negativePathNames.end(); ++negativeIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(negativePyramids)),
					      *negativeIt,
					      validExtensions,
					      recursive);
    }

  sort(positivePyramids.begin(), positivePyramids.end());
  sort(negativePyramids.begin(), negativePyramids.end());
  
  svm_parameter svmParameters;
  svmParameters.svm_type = NU_SVC;
  svmParameters.kernel_type = PRECOMPUTED;
  svmParameters.cache_size = 100; /// suggested in svm_train.c
  svmParameters.eps = 0.00001; // suggested in libsvm readbe for NU_SVM. if C_SVM, change to 0.001
  svmParameters.shrinking = 1; // default in svm_train... seems to make no difference
  svmParameters.probability = 0; // we'll use decision values

  // These parameters are defined for completeness of the
  // parameter structure, but are not used when using NU_SVC and
  // PRECOMPUTED kernel
  svmParameters.degree = 3;
  svmParameters.gamma = 0;
  svmParameters.coef0 = 0;
  svmParameters.C = 1;
  svmParameters.nr_weight = 0;
  svmParameters.weight_label = 0;
  svmParameters.weight = 0;
  
  cout << "CONSTRUCTING GRAM MATRIX" << endl;
  cout << "========================" << endl;
  svm_problem* svmProblem = constructProblem(positivePyramids, negativePyramids, pyramidLevels);

  int nr_fold = 10;
  float bestNu = 0.5;
  float bestAccuracy = 0;
  double *target = new double[svmProblem->l]; // macro defined in svm.cpp
  for ( svmParameters.nu = 0.025f; svmParameters.nu <= 0.7f; svmParameters.nu += 0.025f )
    {
      svm_cross_validation(svmProblem,&svmParameters,nr_fold,target);
      std::cout << "svmParameters.nu = " << svmParameters.nu << "\taccuracy = ";
      int totalCorrect = 0;
      int truePositives = 0;
      int falsePositives = 0;
      int groundTruthPositives = 0;
      // compute accuracy for this setting
      for ( int i = 0; i < svmProblem->l; ++i )
	{
	  if (svmProblem->y[i] == 1 )
	    {
	      ++groundTruthPositives;
	    }
	  if (target[i] == svmProblem->y[i])
	    {
	      ++totalCorrect;
	    }
	  if (target[i] == svmProblem->y[i] && target[i] == 1)
	    {
	      ++truePositives;
	    }
	  if (target[i] != svmProblem->y[i] && target[i] == 1)
	    {
	      ++falsePositives;
	    }
	}
      float accuracy = 100.0f * totalCorrect / (float)svmProblem->l;
      float precision = 100.0f * truePositives / (float)(truePositives + falsePositives);
      float recall = 100.0f * truePositives / (float)(groundTruthPositives);
      float fmeasure = 2 * (precision * recall) / (precision + recall);
      std::cout << accuracy << "\tf-measure = " << fmeasure << std::endl;

      if ( accuracy >= bestAccuracy )
	{
	  bestAccuracy = accuracy;
	  bestNu = svmParameters.nu;
	}
    }
  delete[](target);
  
  svmParameters.nu = bestNu;
  std::cout << "Selected setting = " << svmParameters.nu << std::endl;
  
  cout << endl;
  
  cout << "TRAINING SVM MODEL" << endl;
  cout << "==================" << endl;
  svm_model* svmModel = svm_train(svmProblem, &svmParameters);

  svm_save_model(svmModelFilename.c_str(), svmModel);

  vector<int> supportVectors = sjm::util::readSupportVectors(svmModelFilename);

  cout << "CONSTRUCTING SPARSE PROBLEM" << endl;
  cout << "===========================" << endl;
  svm_problem* sparseProblem = constructSparseProblem( positivePyramids, negativePyramids, pyramidLevels, supportVectors );


  svmParameters.nu = 0.001; // don't loose any of the support vectors
  cout << "TRAINING SPARSE MODEL USING " << supportVectors.size() << " SVs" << endl;
  cout << "====================================" << endl;
  svm_model* svmSparseModel = svm_train(sparseProblem, &svmParameters);
  
  fs::path sparsePath = fs::change_extension(svmModelFilename, ".sparse");

  svm_save_model(sparsePath.string().c_str(), svmSparseModel);
  // save support vectors also
  fs::path svPath = fs::change_extension(svmModelFilename, ".sv");
  fs::ofstream supportVectorFile(svPath, ios::binary);
  int numSupportVectors = supportVectors.size();
  supportVectorFile.write((char*)&numSupportVectors, sizeof(numSupportVectors));
  for ( int i = 0; i < numSupportVectors; ++i )
    {
      supportVectorFile.write((char*)&supportVectors[i], sizeof(supportVectors[i]));
    }
  supportVectorFile.close();
  //fs::remove(svmModelFilename); // delete the un-needed non-sparse model

  return 0;
}

svm_problem* constructSparseProblem( const vector<SpatialPyramid::util::PathAndPyramid> & positivePyramids,
				     const vector<SpatialPyramid::util::PathAndPyramid> & negativePyramids,
				     int levels, const vector<int> & supportVectors )
{
  boost::progress_timer t; // times this function
  int problemSize = supportVectors.size();
  boost::progress_display show_progress( problemSize * (problemSize+1) / 2 );
  
  svm_problem* problem = new svm_problem;
  problem->l = problemSize;
  problem->y = new double[problem->l];
  problem->x = new svm_node*[problem->l];

  for ( int i = 0; i < (int)supportVectors.size(); i++ )
    {
      problem->y[i] = (supportVectors[i] < (int)positivePyramids.size());
    }
  
  vector<SpatialPyramid::util::PathAndPyramid> allPyramids = positivePyramids;
  vector<SpatialPyramid::util::PathAndPyramid>::const_iterator negIt = negativePyramids.begin();
  for ( ; negIt != negativePyramids.end(); ++negIt )
    {
      allPyramids.push_back(*negIt);
    }

  for ( int i = 0; i < problemSize; i++ )
    {
      problem->x[i] = new svm_node[problemSize + 2];
    }

  for ( int a = 0; a < (int)supportVectors.size(); ++a )
    {
      problem->x[a][0].index = 0;
      problem->x[a][0].value = a+1;
      for ( int b = a+1; b < (int)supportVectors.size() + 1; ++b ) // take advantage of the symmetry of the gram matrix
	{
	  float match_score = SpatialPyramid::normalizedSpatialPyramidMatchKernel(allPyramids[supportVectors[a]].pyramid, allPyramids[supportVectors[b-1]].pyramid, levels);
	  problem->x[b-1][a+1].index = a+1;
	  problem->x[b-1][a+1].value = match_score;
	  problem->x[a][b].index = b;
	  problem->x[a][b].value = match_score;
	  ++show_progress;
	}
    }
  
  for ( int i = 0; i < (int)supportVectors.size(); ++i )
    {
      problem->x[i][supportVectors.size() + 1].index = -1;
    }
  
  return problem;
}

svm_problem* constructProblem(const vector<SpatialPyramid::util::PathAndPyramid> & positivePyramids,
			      const vector<SpatialPyramid::util::PathAndPyramid> & negativePyramids, int levels)
{
  boost::progress_timer t; // times this function
  int problemSize = positivePyramids.size() + negativePyramids.size();
  boost::progress_display show_progress( ( problemSize * (problemSize+1)) / 2);

  svm_problem* problem = new svm_problem;
  problem->l = (int)(positivePyramids.size() + negativePyramids.size());
  problem->y = new double[problem->l];
  problem->x = new svm_node*[problem->l];
  
  for ( int i = 0; i < (int)positivePyramids.size(); i++ )
    {
      problem->y[i] = 1;
    }
  for ( int i = 0; i < (int)negativePyramids.size(); i++ )
    {
      problem->y[positivePyramids.size() + i] = -1;
    }
  
  vector<SpatialPyramid::util::PathAndPyramid> allPyramids = positivePyramids;
  vector<SpatialPyramid::util::PathAndPyramid>::const_iterator negIt = negativePyramids.begin();
  for ( ; negIt != negativePyramids.end(); ++negIt )
    {
      allPyramids.push_back(*negIt);
    }
  
  for ( int i = 0; i < (int)allPyramids.size(); ++i )
    {
      problem->x[i] = new svm_node[allPyramids.size() + 2];
    }

  for ( int a = 0; a < (int)allPyramids.size(); ++a )
    {
      problem->x[a][0].index = 0;
      problem->x[a][0].value = a+1;
      for ( int b = a+1; b < (int)allPyramids.size() + 1; ++b ) // take advantage of the symmetry of the gram matrix
	{
	  float match_score = SpatialPyramid::normalizedSpatialPyramidMatchKernel(allPyramids[a].pyramid, allPyramids[b-1].pyramid,
										  levels);
	  problem->x[b-1][a+1].index = a+1;
	  problem->x[b-1][a+1].value = match_score;
	  problem->x[a][b].index = b;
	  problem->x[a][b].value = match_score;
	  ++show_progress;
	}
    }
  
  for ( int i = 0; i < (int)allPyramids.size(); ++i )
    {
      problem->x[i][allPyramids.size() + 1].index = -1;
    }
  
  return problem;
}

void loadPyramid( const fs::path & pyramidPath, vector<SpatialPyramid::util::PathAndPyramid> & pyramids )
{
  cout << "Loading " << pyramidPath.string() << endl;

  fs::ifstream pyramidFile(pyramidPath);
  SpatialPyramid::util::PathAndPyramid pathAndPyramid;
  pathAndPyramid.path = pyramidPath;
  SpatialPyramid::load(pathAndPyramid.pyramid, pyramidFile);
  pyramidFile.close();

  pyramids.push_back(pathAndPyramid);
}
