/*! \file BuildPyramids.cpp

  Implements conversion from a sets of codewords to spatial pyramid
  descriptions of the layout of those codewords.

  Allows specification of a fixed aspect ratio within which all
  training images are centered (even if they are smaller).

  @author Sancho McCann
*/
#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <util/util.hpp>
#include <spatialpyramid/spatial_pyramid.hpp>
#include <sift/sift.hpp>
#include <vector>
#include <string>
#include <limits>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
namespace SpatialPyramid = sjm::SpatialPyramid;
using namespace std;

void convertToPyramid( const fs::path & filepath, const set<int> & selectedCodewords, int forcedWidth, int forcedHeight, bool guaranteedSize, bool binaryHistograms, int depth, const string & codewordDirectoryName, const string & outputDirectoryName);

int main(int argc, char** argv)
{
  vector<string> inputFilenames;
  string selectionFilename;
  bool recursive;
  int numFeatures;
  int depth;
  int forcedWidth;
  int forcedHeight;
  bool guaranteedSize = false;
  bool binaryHistograms = false;
  string codewordDirectoryName;
  string outputDirectoryName;

  set<string> validExtensions;
  validExtensions.insert(".bmp");
  validExtensions.insert(".dib");
  validExtensions.insert(".jpeg");
  validExtensions.insert(".jpg");
  validExtensions.insert(".jpe");
  validExtensions.insert(".png");
  validExtensions.insert(".pbm");
  validExtensions.insert(".pgm");
  validExtensions.insert(".ppm");
  validExtensions.insert(".sr");
  validExtensions.insert(".ras");
  validExtensions.insert(".tiff");
  validExtensions.insert(".tif");

  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help", "produce help message")
    ("input,I", po::value<vector<string> >(), "input files/directories")
    ("recursive,R", "flag to indicate if the image search should recurse on sub-directories")
    ("selection,S", po::value<string>(), "feature selection file")
    ("force-width,W", po::value<int>(), "force the aspect ratio of the pyramid")
    ("force-height,H", po::value<int>(), "force the aspect ratio of the pyramid")
    ("codeword-directory,C", po::value<string>(), "codewords stored remotely from images")
    ("num-features,N", po::value<int>(), "use the top N features from the feature selection file")
    ("depth,D", po::value<int>(), "the number of levels of the spatial pyramid")
    ("guaranteed-size,G", "flag to indicate that the forced-width/height is the exact size of the input codeword domain")
    ("output-directory", po::value<string>()->default_value(""), "alternative directory to write pyramids to")
    ("binary-histograms", "flag to indicate histograms are binary rather than accumulative")
    ;
  po::positional_options_description positional;
  positional.add("input", -1);

  try {
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).options(command_line_options).positional(positional).run(),
	      vm);
    po::notify(vm);

    if ( vm.count("help") || !vm.count("input") || !vm.count("selection") || !vm.count("num-features") || !vm.count("depth") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    inputFilenames = vm["input"].as<vector<string> > ();
    selectionFilename = vm["selection"].as<string>();
    numFeatures = vm["num-features"].as<int>();
    depth = vm["depth"].as<int>();
    recursive = vm.count("recursive");
    guaranteedSize = vm.count("guaranteed-size");
    outputDirectoryName = vm["output-directory"].as<string>();
    binaryHistograms = vm.count("binary-histograms");

    if ( (vm.count("force-width") && !vm.count("force-height")) || (!vm.count("force-width") && vm.count("force-height")) )
      {
	cout << "Need to force both width and height to specify an aspect ratio" << endl;
	exit(1);
      }
    else if ( vm.count("force-width") && vm.count("force-height") )
      {
	forcedWidth = vm["force-width"].as<int>();
	forcedHeight = vm["force-height"].as<int>();
      }
    else
      {
	forcedWidth = 0;
	forcedHeight = 0;
      }

    if ( vm.count("codeword-directory") )
      {
	codewordDirectoryName = vm["codeword-directory"].as<string>();
      }

    if ( guaranteedSize )
      {
	validExtensions.clear();
	validExtensions.insert(".codewords");
      }


  } catch ( po::invalid_command_line_syntax ) {
    cout << "Command line syntax error: " << endl;
    cout << command_line_options << endl;
    exit(1);
  }

  if ( !fs::exists(selectionFilename) )
    {
      cout << selectionFilename << " doesn't exist." << endl;
      exit(1);
    }

  int numCodewords;
  vector<float> featureWeights;
  fs::ifstream selectionFile(selectionFilename);
  selectionFile.read((char*)&numCodewords, sizeof(numCodewords));
  for ( int i = 0; i < numCodewords; ++i )
    {
      float weight;
      selectionFile.read((char*)&weight, sizeof(weight));
      featureWeights.push_back(weight);
    }

  set<int> selectedCodewords;
  sjm::util::selection::chooseTopFeatures(featureWeights, selectedCodewords, numFeatures);

  std::cout << inputFilenames.size() << std::endl;

  vector<string>::const_iterator fileIt = inputFilenames.begin();
  for ( ; fileIt != inputFilenames.end(); ++fileIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(convertToPyramid,
							  _1,
							  boost::ref(selectedCodewords),
							  forcedWidth, forcedHeight,
							  guaranteedSize, binaryHistograms,
							  depth,
							  boost::ref(codewordDirectoryName),
							  boost::ref(outputDirectoryName)),
					      *fileIt,
					      validExtensions,
					      recursive);
    }
  
  selectionFile.close();
}

void convertToPyramid( const fs::path & filepath, const set<int> & selectedCodewords, int forcedWidth, int forcedHeight, bool guaranteedSize, bool binaryHistograms, int depth, const string & codewordDirectoryName, const string & outputDirectoryName )
{
  fs::path codewordPath = fs::change_extension(filepath, ".codewords");
  if ( codewordDirectoryName.size() > 0 )
    {
      codewordPath = fs::path(codewordDirectoryName) / codewordPath.leaf();
    }
  fs::path pyramidPath = fs::change_extension(filepath, ".pyramid");
  if ( outputDirectoryName.size() > 0 )
    {
      // Change output path to include user-specified output directory
      pyramidPath = fs::path(outputDirectoryName) / fs::path(pyramidPath.leaf());
    }

  cout << codewordPath.string() << " -> ";
  cout.flush();

  int imageWidth;
  int imageHeight;
  // Get image dimensions
  if ( !guaranteedSize )
    {
      IplImage* cvImage = 0;
      cvImage = cvLoadImage( filepath.string().c_str(), CV_LOAD_IMAGE_GRAYSCALE );
      imageWidth = cvImage->width;
      imageHeight = cvImage->height;
      cvReleaseImage(&cvImage);
    }

  float offsetX = 0;
  float offsetY = 0;
  
  if ( forcedWidth && forcedHeight && !guaranteedSize )
    {
      offsetX = (forcedWidth - imageWidth) / 2.0f;
      offsetY = (forcedHeight - imageHeight) / 2.0f;
    }

  SpatialPyramid::CodewordSet codewordSet;
  int numCodewords;
  fs::ifstream codewordsFile(codewordPath);
  codewordsFile.read((char*)&numCodewords, sizeof(numCodewords));
  for ( int i = 0; i < numCodewords; ++i )
    {
      int codeword;
      float x;
      float y;
      float scale;
      codewordsFile.read((char*)&codeword, sizeof(codeword));
      codewordsFile.read((char*)&x, sizeof(x));
      codewordsFile.read((char*)&y, sizeof(y));
      codewordsFile.read((char*)&scale, sizeof(scale));
      if ( selectedCodewords.find(codeword) != selectedCodewords.end() )
	{
	  int id = std::distance(selectedCodewords.begin(), selectedCodewords.find(codeword));
	  codewordSet.insert((SpatialPyramid::CodewordSet::value_type){id, x+offsetX, y+offsetY, scale});
	}
    }
  codewordsFile.close();
  
  SpatialPyramid::SpatialPyramidType pyramid = 
    SpatialPyramid::buildSpatialPyramid( (forcedWidth ? forcedWidth : imageWidth),
					 (forcedHeight ? forcedHeight : imageHeight),
					 codewordSet, selectedCodewords.size(), depth, binaryHistograms);

  fs::ofstream pyramidFile(pyramidPath, ios::binary);
  SpatialPyramid::save(pyramid, pyramidFile);
  pyramidFile.close();

  cout << pyramidPath.string() << endl;
}
