/*! \file ExtractDescriptors.cpp
  Manages extraction of a dense set of sift descriptors from an image.

  @author Sancho McCann
*/

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <boost/random.hpp>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <vector>
#include <string>
#include <set>
#include <iostream>
#include <boost/timer.hpp>
#include <util/util.hpp>
#include <sift/sift.hpp>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
using namespace std;


void extractDescriptorsFromFile(const fs::path & imagePath, const SIFT::Parameters & parameters, bool force, const string & outputDirectoryName, SIFT::VariateGeneratorType & myRand);
void saveDescriptors(const fs::path & descriptorPath, const SIFT::Parameters & parameters, const SIFT::DescriptorSet & descriptorSet);


int main(int argc, char** argv)
{
  vector<string> inputPathNames;
  string outputDirectoryName;
  bool recursive = false;
  bool force = false;

  SIFT::Parameters siftParameters;
  siftParameters.rotationInvariance = false;
  siftParameters.noNormalize = 0.0f;
  siftParameters.discard = false;
  siftParameters.multiscale = true;
  siftParameters.octaveSteps = 3;
  siftParameters.percentage = 1.0f;
  siftParameters.minRadius = 0;

  // From OpenCv documentation: these are the
  // file types that cvLoadImage can handle
  set<string> validExtensions;
  validExtensions.insert(".bmp");
  validExtensions.insert(".dib");
  validExtensions.insert(".jpeg");
  validExtensions.insert(".jpg");
  validExtensions.insert(".jpe");
  validExtensions.insert(".png");
  validExtensions.insert(".pbm");
  validExtensions.insert(".pgm");
  validExtensions.insert(".ppm");
  validExtensions.insert(".sr");
  validExtensions.insert(".ras");
  validExtensions.insert(".tiff");
  validExtensions.insert(".tif");

  // Parse command line options
  // --------------------------
  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help,H", "produce help message")
    ("input,I", po::value<vector<string> >(), "input images/directory")
    ("recursive,R", "flag to indicate if the image search should recurse on sub-directories")
    ("no-normalize,N", po::value<float>()->default_value(0.0f), "value below which SIFT descriptors are not normalized. low values, like 0.5 up to maybe 2 are appropriate.")
    ("octave-steps,O", po::value<int>()->default_value(3), "number of steps per octave in the Gaussian pyramid")
    ("min-radius,M", po::value<float>()->default_value(0), "minimum size radius to extract descriptors at")
    ("percentage,p", po::value<float>()->default_value(1), "percentage of descriptors to randomly sample (1.0 is fully dense sampling)")
    ("single-scale,S", "forces single scale extraction rather than multiscale extraction")
    ("discard,D", "discard descriptors that are below no-normalize")
    ("output-directory", po::value<string>()->default_value(""), "optionally specify an alternate output directory; the default is to place the .sift files in the same directory as the images")
    ("force,F", "forces overwrite of existing sift file, otherwise, skips processing if the output file already exists")
    ;
  po::positional_options_description positional;
  positional.add("input", -1);

  try {
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).options(command_line_options).positional(positional).run(),
	      vm);
    po::notify(vm);
    
    if ( vm.count("help") || !vm.count("input") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    else
      {
	inputPathNames = vm["input"].as<vector<string> >();
      }
    outputDirectoryName = vm["output-directory"].as<string>();
    siftParameters.noNormalize = vm["no-normalize"].as<float>();    
    recursive = vm.count("recursive");
    siftParameters.multiscale = !(vm.count("single-scale"));
    siftParameters.octaveSteps = vm["octave-steps"].as<int>();
    siftParameters.discard = vm.count("discard");
    siftParameters.percentage = vm["percentage"].as<float>();
    siftParameters.minRadius = vm["min-radius"].as<float>();
    force = vm.count("force");
  } catch ( po::invalid_command_line_syntax ) {
    cout << command_line_options << endl;
    exit(1);
  } catch ( po::unknown_option ) {
    cout << command_line_options << endl;
    exit(1);
  }

  // Random number generator for random sampling of descriptors
  SIFT::RandomGeneratorType random_generator;
  random_generator.seed(time(0));
  SIFT::BernoulliType bernoulli(siftParameters.percentage);
  SIFT::VariateGeneratorType myRand(random_generator, bernoulli);

  // Do the extractions
  // ------------------
  vector<string>::const_iterator it = inputPathNames.begin();
  for ( ; it != inputPathNames.end(); ++it )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(extractDescriptorsFromFile,
							  _1,
							  siftParameters,
							  force,
							  boost::ref(outputDirectoryName),
							  boost::ref(myRand)
							  ),
					      *it, validExtensions, recursive);
    }
  
  return 0;
}

/*!
  Extracts sift descriptors from an input image and writes them to an output file.

  @param imagePath the location of the input image
  @param parameters parameters dictating the options of the sift extraction
  @param force flag to force overwrite of a possibly already existing output file

  @param outputDirectoryName indicates an alternate directory to write
  the sift file to... if this string is 0-length, the output will be
  written in the same directory as the input image

  @param myRand the random variable generator required by the sift
  extraction function for random sampling
*/
void extractDescriptorsFromFile(const fs::path & imagePath, const SIFT::Parameters & parameters, bool force,
				const string & outputDirectoryName, SIFT::VariateGeneratorType & myRand )
{
  IplImage* cvImage = 0;

  cout << imagePath.string() << " -> ";
  cout.flush();

  fs::path descriptorPath = fs::change_extension( imagePath, ".sift" );
  if ( outputDirectoryName.size() > 0 )
    {
      // Change output path to include user-specified output directory
      descriptorPath = fs::path(outputDirectoryName) / fs::path(descriptorPath.leaf());
    }

  // Being careful about overwriting already existing data
  if ( !fs::exists(descriptorPath) || force )
    {
      cvImage = cvLoadImage( imagePath.string().c_str(), CV_LOAD_IMAGE_GRAYSCALE );
      if ( cvImage > 0 )
	{
	  // Passing the cvImage in a way that allows wrapping in a GIL image view
	  SIFT::DescriptorSet descriptorSet = SIFT::ComputeDenseSiftFeaturesGray8((unsigned char*)cvImage->imageData,
										  (ptrdiff_t)cvImage->widthStep,
										  cvImage->width,
										  cvImage->height,
										  0, 0, cvImage->width, cvImage->height,
										  parameters, myRand);
	  cout << descriptorPath.string() << endl;
	  saveDescriptors(descriptorPath, parameters, descriptorSet);
	  cvReleaseImage(&cvImage);
	}
      else
	{
	  cout << "error loading file." << endl;
	}
    }
  else
    {
      cout << descriptorPath.string() << " already exists... skipping" << endl;
    }
}

/*!
  Save the sift descriptor set to file

  @param descriptorPath the output path
  @param parameters the parameters involved with the sift extraction
  @param descriptorSet the set of descriptors to write to file
*/
void saveDescriptors(const fs::path & descriptorPath, const SIFT::Parameters & parameters, const SIFT::DescriptorSet & descriptorSet)
{
  fs::ofstream descriptorFile(descriptorPath, std::ios::binary);

  // Write the parameter settings
  SIFT::writeParameters(parameters, descriptorFile);

  // Write the count and dimensions
  int numDescriptors = descriptorSet.size();
  descriptorFile.write((char*)&numDescriptors, sizeof(int));
  descriptorFile.write((char*)&SIFT::Dimensions, sizeof(int));

  SIFT::DescriptorSet::const_iterator descriptorIterator = descriptorSet.begin();
  for ( ; descriptorIterator != descriptorSet.end(); ++descriptorIterator )
    {
      SIFT::save(*(descriptorIterator->descriptor()), descriptorFile);

      SIFT::DescriptorSet::value_type::IndexType x = descriptorIterator->x();
      SIFT::DescriptorSet::value_type::IndexType y = descriptorIterator->y();
      float scale = descriptorIterator->scale();

      descriptorFile.write((char*)&x, sizeof(SIFT::DescriptorSet::value_type::IndexType));
      descriptorFile.write((char*)&y, sizeof(SIFT::DescriptorSet::value_type::IndexType));
      descriptorFile.write((char*)&scale, sizeof(float));
    }

  descriptorFile.close();
}
