/*! \file FeatureWeighting.cpp
  
  Just counts builds up an aggregate histogram for feature counts
  across all positive images and across all negative images.

  @author Sancho McCann
*/ 

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <util/util.hpp>
#include <sift/sift.hpp>
#include <vector>
#include <string>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
using namespace std;

void accumulateHistograms( const fs::path & codewordPath, vector<vector<int> > & histograms, int numCodewords );

int main(int argc, char** argv)
{
  set<string> validExtensions;
  vector<string> positivePathNames;
  vector<string> negativePathNames;
  string featureSelectionFilename;
  bool recursive = false;
  int numCodewords;

  validExtensions.insert(".codewords");

  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help,H", "produce help message")
    ("negative,N", po::value<vector<string> >(), "negative example images/directories")
    ("positive,P", po::value<vector<string> >(), "positive example images/directories")
    ("recursive,R", "flag to indicate if the image search should recurse on sub-directories")
    ("num", po::value<int>(), "number of codewords")
    ("output,o", po::value<string>(), "feature selection destination file")
    ;

  try {
    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, command_line_options), vm);
    po::notify(vm);
    
    if ( vm.count("help") || !vm.count("positive") || !vm.count("negative") || !vm.count("output") || !vm.count("num") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    else
      {
	positivePathNames = vm["positive"].as<vector<string> >();
	negativePathNames = vm["negative"].as<vector<string> >();
	featureSelectionFilename = vm["output"].as<string>();
	numCodewords = vm["num"].as<int>();
      }
    recursive = vm.count("recursive");
  } catch ( po::invalid_command_line_syntax ) {
    cout << command_line_options << endl;
    exit(1);
  }

  vector<vector<int> > positiveFeatureHistograms;
  vector<vector<int> > negativeFeatureHistograms;

  vector<string>::const_iterator positiveIt = positivePathNames.begin();
  for ( ; positiveIt != positivePathNames.end(); ++positiveIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(accumulateHistograms,
							  _1,
							  boost::ref(positiveFeatureHistograms),
							  numCodewords),
					      *positiveIt,
					      validExtensions,
					      recursive);
    }

  vector<string>::const_iterator negativeIt = negativePathNames.begin();
  for ( ; negativeIt != negativePathNames.end(); ++negativeIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(accumulateHistograms,
							  _1,
							  boost::ref(negativeFeatureHistograms),
							  numCodewords),
					      *negativeIt,
					      validExtensions,
					      recursive);
    }

  fs::ofstream svmTrainingFile(featureSelectionFilename, ios::binary);

  int numData = negativeFeatureHistograms.size() + positiveFeatureHistograms.size();
  int dimensions = numCodewords;
  svmTrainingFile.write((char*)&numData, sizeof(numData));
  svmTrainingFile.write((char*)&dimensions, sizeof(dimensions));
  for ( int i = 0; i < (int)positiveFeatureHistograms.size(); i++ )
    {
      for ( int j = 0; j < numCodewords; j++ )
	{
	  int element = positiveFeatureHistograms[i][j];
	  svmTrainingFile.write((char*)&element, sizeof(element));
	}
    }
  for ( int i = 0; i < (int)negativeFeatureHistograms.size(); i++ )
    {
      for ( int j = 0; j < numCodewords; j++ )
	{
	  int element = negativeFeatureHistograms[i][j];
	  svmTrainingFile.write((char*)&element, sizeof(element));
	}
    }
  int positive = 1;
  int negative = 2;
  for ( int i = 0; i < (int)positiveFeatureHistograms.size(); i++ )
    {
      svmTrainingFile.write((char*)&positive, sizeof(int));
    }
  for ( int i = 0; i < (int)negativeFeatureHistograms.size(); i++ )
    {
      svmTrainingFile.write((char*)&negative, sizeof(int));
    }

  svmTrainingFile.close();

  return 0;
}

void accumulateHistograms( const fs::path & codewordPath, vector<vector<int> > & histograms, int numCodewords )
{
  int numWords;

  fs::ifstream codewordFile(codewordPath, ios::binary);
  codewordFile.read((char*)&numWords, sizeof(int));

  vector<int> histogram(numCodewords, 0);
  for ( int i = 0; i < numWords; i++ )
    {
      SIFT::DescriptorSet::value_type::IndexType dummy;
      float floatDummy;
      int word;
      codewordFile.read((char*)&word, sizeof(int));
      
      codewordFile.read((char*)&dummy, sizeof(dummy)); // x
      codewordFile.read((char*)&dummy, sizeof(dummy)); // y
      codewordFile.read((char*)&floatDummy, sizeof(floatDummy)); // scale

      histogram[word]++;
    }
  histograms.push_back(histogram);

  codewordFile.close();
}
