/*! \File LocateObjects.cpp

  Uses the learned svm model and set of examplars to produce
  activations in test images.

  @author Sancho McCann
*/

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <vector>
#include <string>
#include <set>
#include <iostream>
#include <limits>
#include <cmath>
#include <functional>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <util/util.hpp>
#include <sift/sift.hpp>
#include <histograms/integral_histogram.hpp>
#include <spatialpyramid/spatial_pyramid.hpp>
#include <boost/bind.hpp>
#include <svm/svm.h>
#include <boost/timer.hpp>
#include <boost/progress.hpp>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
namespace SpatialPyramid = sjm::SpatialPyramid;
using namespace std;

struct SvmModels
{
  svm_model* svmModel;
  svm_model* colourSvmModel;
};

struct LocationOfInterest
{
  int tlx;
  int tly;
  float scale;
  int width;
  int height;
  float value;
  
  bool operator < (const LocationOfInterest & l) const
  {
    return scale < l.scale;
  }
};

struct SiftAndColourCollection
{
  vector<SpatialPyramid::util::PathAndPyramid> siftPyramids;
  vector<SpatialPyramid::util::PathAndPyramid> allColourPyramids;
  vector<SpatialPyramid::util::PathAndPyramid> positiveColourPyramids;

};

struct PathNames
{
  string codewords;
  string activations;
};

struct WindowSize
{
  int width;
  int height;
};

struct Resolution
{
  int x;
  int y;
};

struct ExtraOptions
{
  int pyramidLevels;
  bool binaryHistograms;
  int numFeatures;
  float colourWeight;
};

void createActivationPyramid( const fs::path & imagePath,
			      const PathNames & pathNames,
			      const SiftAndColourCollection & pyramids,
			      const set<int> & selectedCodewords,
			      SvmModels models,
			      const vector<int> & supportVectors,
			      ExtraOptions options,
			      Resolution resolution,
			      WindowSize windowSize);
void makeTestNode(const SpatialPyramid::SpatialPyramidType & pyramid, const vector<int> & supportVectors,
		  const vector<SpatialPyramid::util::PathAndPyramid> & trainingPyramids, svm_node* testNode, int pyramidLevels);
void loadPyramid( const fs::path & pyramidPath, vector<SpatialPyramid::util::PathAndPyramid> & pyramids );

bool multiScaleTrain;

int main(int argc, char** argv)
{
  vector<string> positivePathNames;
  vector<string> negativePathNames;
  vector<string> positiveColourPaths;
  vector<string> negativeColourPaths;
  vector<string> imagePathNames;
  string svmModelBasename;
  string colourSvmModelBasename;
  string featureWeightingFilename;
  ExtraOptions extraOptions;
  extraOptions.binaryHistograms = false;
  bool recursive = false;
  Resolution res;
  WindowSize windowSize;
  multiScaleTrain = false;
  PathNames pathNames;

  // From OpenCv documentation: these are the
  // file types that cvLoadImage can handle
  set<string> validExtensions;
  validExtensions.insert(".bmp");
  validExtensions.insert(".dib");
  validExtensions.insert(".jpeg");
  validExtensions.insert(".jpg");
  validExtensions.insert(".jpe");
  validExtensions.insert(".png");
  validExtensions.insert(".pbm");
  validExtensions.insert(".pgm");
  validExtensions.insert(".ppm");
  validExtensions.insert(".sr");
  validExtensions.insert(".ras");
  validExtensions.insert(".tiff");
  validExtensions.insert(".tif");

  set<string> pyramidExtension;
  pyramidExtension.insert(".pyramid");

  set<string> colourPyramidExtension;
  colourPyramidExtension.insert(".cpyramid");

  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help", "produce help message")
    ("images,I", po::value<vector<string> >(), "input images/directory")
    ("codeword-directory,D", po::value<string>(), "testing codeword files maintained separately from images")
    ("activation-directory,A", po::value<string>()->default_value(""), "activations stored separately")
    ("negative,N", po::value<vector<string> >(), "negative example images/directories")
    ("positive,P", po::value<vector<string> >(), "positive example images/directories")
    ("positive-colour", po::value<vector<string> >(), "negative colour pyramids")
    ("negative-colour", po::value<vector<string> >(), "positive colour pyramids")
    ("recursive,R", "flag to indicate if the image search should recurse on sub-directories")
    ("levels,L", po::value<int>(), "levels of the pyramid to use in the kernel")
    ("svm-model,M", po::value<string>(), "svm_model destination file")
    ("colour-svm-model", po::value<string>(), "colour svm model")
    ("selection-file,S", po::value<string>(), "feature weighting file")
    ("x-spacing,X", po::value<int>()->default_value(5), "resolution of sweeping window")
    ("y-spacing,Y", po::value<int>()->default_value(5), "resolution of sweeping window")
    ("window-width,W", po::value<int>(), "width of sliding window")
    ("window-height,H", po::value<int>(), "height of sliding window")
    ("multi-scale", "flag to perform multi-scale sliding window and return an activation pyramid instead of map")
    ("multi-scale-train", "do sweeping window assuming multi-scale training data")
    ("binary-histograms", "flag to indicate histograms are binary rather than accumulative")
    ("num-features,n", po::value<int>(), "how many top features are being used?")
    ("colour-weight,", po::value<float>()->default_value(0.5), "weighting given to colour pyramids")
    ;
  po::positional_options_description positional;
  positional.add("images", -1);

  try {
    po::variables_map vm;
    po::store(po::command_line_parser(argc, argv).options(command_line_options).positional(positional).run(),
	      vm);
    po::notify(vm);
    
    if ( vm.count("help") || !vm.count("images") || !vm.count("negative") ||
	 !vm.count("positive") || !vm.count("levels") ||
	 !vm.count("svm-model") || !vm.count("window-width") || !vm.count("window-height") ||
	 !vm.count("codeword-directory") ||
	 !vm.count("num-features") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    else
      {
	imagePathNames = vm["images"].as<vector<string> >();
	negativePathNames = vm["negative"].as<vector<string> >();
	positivePathNames = vm["positive"].as<vector<string> >();
	negativeColourPaths = vm["negative-colour"].as<vector<string> >();
	positiveColourPaths = vm["positive-colour"].as<vector<string> >();
	recursive = vm.count("recursive");
	extraOptions.pyramidLevels = vm["levels"].as<int>();
	svmModelBasename = vm["svm-model"].as<string>();
	colourSvmModelBasename = vm["colour-svm-model"].as<string>();
	featureWeightingFilename = vm["selection-file"].as<string>();
	res.x = vm["x-spacing"].as<int>();
	res.y = vm["y-spacing"].as<int>();
	windowSize.width = vm["window-width"].as<int>();
	windowSize.height= vm["window-height"].as<int>();
	multiScaleTrain = vm.count("multi-scale-train");
	pathNames.codewords = vm["codeword-directory"].as<string>();
	pathNames.activations = vm["activation-directory"].as<string>();
	extraOptions.binaryHistograms = vm.count("binary-histograms");
	extraOptions.numFeatures = vm["num-features"].as<int>();
	extraOptions.colourWeight = vm["colour-weight"].as<float>();
      }
  } catch ( po::invalid_command_line_syntax ) {
    cout << "Invalid command line syntax" << endl;
    cout << command_line_options << endl;
    exit(1);
  }

  //cout << svmModelBasename << endl;

  //fs::path svmSparseModelPath = fs::change_extension(svmModelBasename, ".sparse");
  fs::path svmModelPath = fs::change_extension(svmModelBasename, ".model");
  fs::path supportVectorsPath = fs::change_extension(svmModelBasename, ".sv");

  fs::path colourSvmModelPath = fs::change_extension(colourSvmModelBasename, ".model");

  //cout << svmSparseModelPath << endl;
  //cout << supportVectorsPath << endl;

  if ( !fs::exists(svmModelPath) )
    {
      cout << svmModelPath.string() << " does not exist." << endl;
      exit(1);
    }
  if ( !fs::exists(colourSvmModelPath) )
    {
      cout << colourSvmModelPath.string() << " does not exist." <<  endl;
      exit(1);
    }
  if ( !fs::exists(supportVectorsPath) )
    {
      cout << supportVectorsPath.string() << " does not exist." << endl;
    }

  vector<SpatialPyramid::util::PathAndPyramid> positivePyramids;
  vector<SpatialPyramid::util::PathAndPyramid> negativePyramids;
  vector<SpatialPyramid::util::PathAndPyramid> positiveColourPyramids;
  vector<SpatialPyramid::util::PathAndPyramid> negativeColourPyramids;

  for ( vector<string>::const_iterator positiveIt = positivePathNames.begin(); positiveIt != positivePathNames.end(); ++positiveIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(positivePyramids)),
					      *positiveIt,
					      pyramidExtension,
					      recursive);
    }

  for ( vector<string>::const_iterator negativeIt = negativePathNames.begin(); negativeIt != negativePathNames.end(); ++negativeIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(negativePyramids)),
					      *negativeIt,
					      pyramidExtension,
					      recursive);
    }
  
  for ( vector<string>::const_iterator positiveIt = positiveColourPaths.begin(); positiveIt != positiveColourPaths.end(); ++positiveIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(positiveColourPyramids)),
					      *positiveIt,
					      colourPyramidExtension,
					      recursive);
    }

  for ( vector<string>::const_iterator negativeIt = negativeColourPaths.begin(); negativeIt != negativeColourPaths.end(); ++negativeIt )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(loadPyramid,
							  _1,
							  boost::ref(negativeColourPyramids)),
					      *negativeIt,
					      colourPyramidExtension,
					      recursive);
    }

  sort(positivePyramids.begin(), positivePyramids.end());
  sort(negativePyramids.begin(), negativePyramids.end());
  sort(positiveColourPyramids.begin(), positiveColourPyramids.end());
  sort(negativeColourPyramids.begin(), negativeColourPyramids.end());


  SvmModels svmModels;
  svmModels.colourSvmModel = 0;
  svmModels.svmModel = 0;
  
  svmModels.svmModel = svm_load_model(svmModelPath.string().c_str());
  svmModels.colourSvmModel = svm_load_model(colourSvmModelPath.string().c_str());

  if ( !fs::exists(featureWeightingFilename) )
    {
      cout << featureWeightingFilename << " doesn't exist." << endl;
      exit(1);
    }

  int numCodewords;
  vector<float> featureWeights;
  fs::ifstream selectionFile(featureWeightingFilename);
  selectionFile.read((char*)&numCodewords, sizeof(numCodewords));
  for ( int i = 0; i < numCodewords; ++i )
    {
      float weight;
      selectionFile.read((char*)&weight, sizeof(weight));
      featureWeights.push_back(weight);
    }

  set<int> selectedCodewords;
  sjm::util::selection::chooseTopFeatures(featureWeights, selectedCodewords, extraOptions.numFeatures);

  vector<SpatialPyramid::util::PathAndPyramid> allPyramids;
  vector<SpatialPyramid::util::PathAndPyramid> allColourPyramids;

  allPyramids = positivePyramids;
  for ( unsigned int i = 0; i < negativePyramids.size(); ++i )
    {
      allPyramids.push_back(negativePyramids[i]);
    }
  positivePyramids.clear();
  negativePyramids.clear();
  
  allColourPyramids = positiveColourPyramids;
  for ( unsigned int i = 0; i < negativeColourPyramids.size(); ++i )
    {
      allColourPyramids.push_back(negativeColourPyramids[i]);
    }

  //For using sparse model
  /*
  vector<int> supportVectors;
  fs::ifstream supportVectorFile(supportVectorsPath, ios::binary);
  int numSupportVectors;
  supportVectorFile.read((char*)&numSupportVectors, sizeof(numSupportVectors));
  for ( int i = 0; i < numSupportVectors; ++i )
    {
      int element;
      supportVectorFile.read((char*)&element, sizeof(element));
      supportVectors.push_back(element);
    }
  */
    
  // For using dense model, taking no advantage of its support vectors
  vector<int> supportVectors;
  for ( int i = 0; i < (int)allPyramids.size(); ++i )
    {
      supportVectors.push_back(i);
    }

  SiftAndColourCollection collection;
  collection.siftPyramids = allPyramids;
  collection.allColourPyramids = allColourPyramids;
  collection.positiveColourPyramids = positiveColourPyramids;
  for ( vector<string>::const_iterator inputIterator = imagePathNames.begin(); inputIterator != imagePathNames.end(); ++inputIterator )
    {
      sjm::util::recursiveFunctionApplication(boost::bind(createActivationPyramid, _1,
							  boost::ref(pathNames),
							  boost::ref(collection),
							  boost::ref(selectedCodewords),
							  svmModels, boost::ref(supportVectors),
							  extraOptions,
							  res,
							  windowSize),
					      *inputIterator,
					      validExtensions,
					      recursive);
    }
  
  return 0;
}

void createActivationPyramid( const fs::path & imagePath,
			      const PathNames & pathNames,
			      const SiftAndColourCollection & pyramids,
			      const set<int> & selectedCodewords,
			      SvmModels models,
			      const vector<int> & supportVectors,
			      ExtraOptions options,
			      Resolution res,
			      WindowSize windowSize)
{
  // Set up paths
  fs::path codewordPath = fs::change_extension(imagePath, ".codewords");
  fs::path activationsPath = fs::change_extension(imagePath, ".activations");

  codewordPath = fs::path(pathNames.codewords) / codewordPath.leaf();
  activationsPath = fs::change_extension(codewordPath, ".activations");
  
  if (pathNames.activations.size() > 0)
    {
      activationsPath = fs::path(pathNames.activations) / activationsPath.leaf();
    }
  
  // Get image size
  IplImage* cvImage = 0;
  cvImage = cvLoadImage( imagePath.string().c_str(), CV_LOAD_IMAGE_GRAYSCALE );
  int imageWidth = cvImage->width;
  int imageHeight = cvImage->height;
  cvReleaseImage(&cvImage);

  // Load the codewords
  SpatialPyramid::CodewordSet codewordSet;
  int numCodewords;
  fs::ifstream codewordsFile(codewordPath);
  codewordsFile.read((char*)&numCodewords, sizeof(numCodewords));
  for ( int i = 0; i < numCodewords; ++i )
    {
      int codeword;
      float x;
      float y;
      float scale;
      codewordsFile.read((char*)&codeword, sizeof(codeword));
      codewordsFile.read((char*)&x, sizeof(x));
      codewordsFile.read((char*)&y, sizeof(y));
      codewordsFile.read((char*)&scale, sizeof(scale));
      if ( selectedCodewords.find(codeword) != selectedCodewords.end() )
	{
	  int id = std::distance(selectedCodewords.begin(), selectedCodewords.find(codeword));
	  codewordSet.insert((SpatialPyramid::CodewordSet::value_type){id, x, y, scale});
	}
    }
  codewordsFile.close();

  // Get scales: we'll only scan the image at as many scales as codewords were extracted.
  set<float> scales;
  SpatialPyramid::CodewordSet::const_iterator codewordIterator = codewordSet.begin();
  for ( ; codewordIterator != codewordSet.end(); ++codewordIterator )
    {
      scales.insert(codewordIterator->scale);
    }
  
  svm_node* testNode = new svm_node[supportVectors.size()+1]; // For svm-ing

  // Write the start of the activation file
  fs::ofstream activationFile(activationsPath);
  activationFile << windowSize.width << " " << windowSize.height << endl;
  float baseScale = *(scales.begin());
  for ( set<float>::const_iterator scaleIterator = scales.begin(); scaleIterator != scales.end(); ++scaleIterator )
    activationFile << (*scaleIterator)/baseScale << " ";
  activationFile << endl;
  
  // Locations stored by the less discriminative level of the cascade
  // search for search by the next discriminative level
  vector<LocationOfInterest> interestLocations;
  sjm::histograms::IntegralHistogram colourHistogram;

  // Initial colour sweep
  { 
    // Get colour information
    boost::gil::rgb16_image_t image;
    boost::gil::jpeg_read_and_convert_image(imagePath.string(), image);
    boost::gil::rgb16_image_t downsampledImage( (image.width() + 1) / 2, (image.height() + 1) / 2);
    boost::gil::copy_pixels(boost::gil::subsampled_view(boost::gil::view(image), 2, 2), boost::gil::view(downsampledImage));

    boost::gil::hsv32f_image_t hsvImage( boost::gil::view( downsampledImage ).dimensions() );
    boost::gil::copy_pixels( boost::gil::color_converted_view<boost::gil::hsv32f_pixel_t>( boost::gil::view( downsampledImage ))
			     , boost::gil::view( hsvImage ));
    boost::gil::hsv32f_view_t hsvView = boost::gil::view(hsvImage);
    
    std::cout << "Building integral histogram..."; std::cout.flush();
    boost::timer histogramTimer;
    colourHistogram = sjm::histograms::integralHistogramFromView( hsvView );
    std::cout << "completed in " << histogramTimer.elapsed() << "s." << std::endl;

    boost::timer timer;
    set<float>::const_iterator scaleIterator = scales.begin();
    for ( ; scaleIterator != scales.end(); ++scaleIterator )
      {
	float scaleRatio = *scaleIterator / baseScale;
	WindowSize scaledWindow = { int(windowSize.width * scaleRatio), int(windowSize.height * scaleRatio) };
 	int scaledYSpacing = res.y;
 	int scaledXSpacing = res.x;
	int xSteps = int(floor((imageWidth - 1 - scaledWindow.width) / (float)scaledXSpacing) + 1);
	int ySteps = int(floor((imageHeight - 1 - scaledWindow.height) / (float)scaledYSpacing) + 1);
	
	if ( xSteps > 0 && ySteps > 0 )
	  {
	    double decisionValue;
	    for ( int topLeftY = 0; topLeftY + scaledWindow.height < imageHeight; topLeftY += scaledYSpacing )
	      {
		for ( int topLeftX = 0; topLeftX + scaledWindow.width < imageWidth; topLeftX += scaledXSpacing )
		  {
		    SpatialPyramid::SpatialPyramidType colourPyramid = 
		      SpatialPyramid::pyramidFromIntegralHistogram(colourHistogram,
								   topLeftX/2, topLeftY/2, scaledWindow.width/2, scaledWindow.height/2, 1);

		    makeTestNode(colourPyramid, supportVectors, pyramids.allColourPyramids, testNode, 1);
		    
		    svm_predict_values(models.colourSvmModel, testNode, &decisionValue);

		    if ( decisionValue > 0 )
		      {
			interestLocations.push_back( (LocationOfInterest){topLeftX, topLeftY, *scaleIterator, scaledWindow.width, scaledWindow.height, decisionValue} );
			
		      }
		    
		  } // end x sweep
		//cout << endl;
	      } // end y sweep
	  } // end if sweep is possible      
      } // end iteration over scales
    std::cout << "Colour sweep in " << timer.elapsed() << " s" << std::endl;
  } // end colour sweep

  for ( int cascadeLevel = 1; cascadeLevel <= options.pyramidLevels; ++cascadeLevel )
    {
      std::cout << interestLocations.size() << " locations of interest remaining" << std::endl;
  
//       if ( cascadeLevel == 1 )
// 	{
// 	  boost::timer timer;
// 	  set<float>::const_iterator scaleIterator = scales.begin();
// 	  for ( ; scaleIterator != scales.end(); ++scaleIterator )
// 	    {
// 	      SpatialPyramid::CodewordSet inScaleCodewords;
// 	      SpatialPyramid::CodewordSet::const_iterator codewordIterator = codewordSet.begin();
// 	      for ( ; codewordIterator != codewordSet.end(); ++codewordIterator )
// 		{
// 		  if ( !multiScaleTrain )
// 		    {
// 		      if ( abs(codewordIterator->scale - *scaleIterator) < numeric_limits<float>::epsilon() )
// 			{
// 			  inScaleCodewords.insert(*codewordIterator);
// 			}
// 		    }
// 		  else
// 		    {
// 		      if ( codewordIterator->scale > *scaleIterator - numeric_limits<float>::epsilon() )
// 			{
// 			  inScaleCodewords.insert(*codewordIterator);
// 			}
// 		    }
		  
// 		}
	      
// 	      sjm::histograms::IntegralHistogram codewordHistogram =
// 		sjm::histograms::integralHistogramFromCodewordSet( inScaleCodewords, imageWidth, imageHeight, options.numFeatures );
	      
// 	      float scaleRatio = *scaleIterator / baseScale;
// 	      WindowSize scaledWindow = { int(windowSize.width * scaleRatio), int(windowSize.height * scaleRatio) };
// 	      int scaledYSpacing = res.y;
// 	      int scaledXSpacing = res.x;
	      
// 	      int xSteps = int(floor((imageWidth - 1 - scaledWindow.width) / (float)scaledXSpacing) + 1);
// 	      int ySteps = int(floor((imageHeight - 1 - scaledWindow.height) / (float)scaledYSpacing) + 1);
	      
// 	      if ( xSteps > 0 && ySteps > 0 )
// 		{
// 		  for ( int topLeftY = 0; topLeftY + scaledWindow.height < imageHeight; topLeftY += scaledYSpacing )
// 		    {
// 		      for ( int topLeftX = 0; topLeftX + scaledWindow.width < imageWidth; topLeftX += scaledXSpacing )
// 			{
// 			  SpatialPyramid::SpatialPyramidType pyramid = 
// 			    SpatialPyramid::pyramidFromIntegralHistogram(codewordHistogram,
// 									 topLeftX, topLeftY, scaledWindow.width, scaledWindow.height, cascadeLevel);
// 			  SpatialPyramid::SpatialPyramidType colourPyramid = 
// 			    SpatialPyramid::pyramidFromIntegralHistogram(colourHistogram,
// 									 topLeftX, topLeftY, scaledWindow.width, scaledWindow.height, cascadeLevel);
			  

// 			  makeTestNode(pyramid, supportVectors, pyramids.siftPyramids, testNode, cascadeLevel);
// 			  double siftDecisionValue = 0;
// 			  svm_predict_values(models.svmModel, testNode, &siftDecisionValue);
			  
// 			  makeTestNode(colourPyramid, supportVectors, pyramids.allColourPyramids, testNode, cascadeLevel);
// 			  double colourDecisionValue = 0;
// 			  svm_predict_values(models.colourSvmModel, testNode, &colourDecisionValue);
			  
// 			  float decisionValue = ((1 - options.colourWeight) * siftDecisionValue + options.colourWeight * colourDecisionValue);
// 			  if ( decisionValue > 0 )
// 			    {
// 			      interestLocations.push_back( (LocationOfInterest){topLeftX, topLeftY, *scaleIterator, scaledWindow.width, scaledWindow.height, decisionValue} );
// 			      if ( options.pyramidLevels == 1 ) // if we're only using one level, write the results now
// 				{
// 				  activationFile << topLeftX << " " << topLeftY << " " << scaleRatio << " " << decisionValue << endl;
// 				}
// 			    }
			  
// 			} // end x sweep
// 		      //cout << endl;
// 		    } // end y sweep
// 		}      
// 	    }
// 	  std::cout << timer.elapsed() << " s for bag-of-features level" << std::endl;
	    
// 	} // end if ( cascadeLevel == 1 )
//       else
	{
	  boost::timer timer;
	  std::sort(interestLocations.begin(), interestLocations.end());
	  vector<LocationOfInterest> newInterestLocations;
	  set<float> interestScales;
	  for ( vector<LocationOfInterest>::const_iterator it = interestLocations.begin();
		it != interestLocations.end(); ++it )
	    {
	      interestScales.insert(it->scale);
	    }

	  set<float>::const_iterator scaleIterator = interestScales.begin();
	  int scaleIndex = -1;
	  for ( ; scaleIterator != interestScales.end(); ++scaleIterator )
	    {
	      scaleIndex++;
	      SpatialPyramid::CodewordSet inScaleCodewords;
	      SpatialPyramid::CodewordSet::const_iterator codewordIterator = codewordSet.begin();
	      for ( ; codewordIterator != codewordSet.end(); ++codewordIterator )
		{
		  if ( !multiScaleTrain )
		    {
		      if ( abs(codewordIterator->scale - *scaleIterator) < numeric_limits<float>::epsilon() )
			{
			  inScaleCodewords.insert(*codewordIterator);
			}
		    }
		  else
		    {
		      if ( codewordIterator->scale > *scaleIterator - numeric_limits<float>::epsilon() )
			{
			  inScaleCodewords.insert(*codewordIterator);
			}
		    }
		  
		}
	      
// 	      sjm::histograms::IntegralHistogram codewordHistogram =
// 		sjm::histograms::integralHistogramFromCodewordSet( inScaleCodewords, imageWidth, imageHeight, options.numFeatures );
	      
	      float scaleRatio = *scaleIterator / baseScale;
	      
	      vector<LocationOfInterest>::const_iterator locationIt = interestLocations.begin();
	      for ( ; locationIt != interestLocations.end(); ++locationIt )
		{
		  if ( locationIt->scale == *scaleIterator )
		    {
		      int width = locationIt->width;
		      int height = locationIt->height;
		      int topLeftX = locationIt->tlx;
		      int topLeftY = locationIt->tly;

		      SpatialPyramid::SpatialPyramidType pyramid =
			SpatialPyramid::buildSpatialPyramid(width, height, inScaleCodewords, options.numFeatures,
							    cascadeLevel, options.binaryHistograms, topLeftX, topLeftY);
							    

// 		      SpatialPyramid::SpatialPyramidType pyramid = 
// 			SpatialPyramid::pyramidFromIntegralHistogram(codewordHistogram,
// 									   topLeftX, topLeftY, width, height, cascadeLevel);
		      SpatialPyramid::SpatialPyramidType colourPyramid = 
			SpatialPyramid::pyramidFromIntegralHistogram(colourHistogram,
									   topLeftX/2, topLeftY/2, width/2, height/2, cascadeLevel);
		      

		      makeTestNode(pyramid, supportVectors, pyramids.siftPyramids, testNode, cascadeLevel);
		      double siftDecisionValue = 0;
		      svm_predict_values(models.svmModel, testNode, &siftDecisionValue);
		      
		      makeTestNode(colourPyramid, supportVectors, pyramids.allColourPyramids, testNode, cascadeLevel);
		      double colourDecisionValue = 0;
		      svm_predict_values(models.colourSvmModel, testNode, &colourDecisionValue);
		      
		      double decisionValue = ((1 - options.colourWeight) * siftDecisionValue + options.colourWeight * colourDecisionValue);

		      if (decisionValue > 0)
			{
			  if ( cascadeLevel == options.pyramidLevels )
			    {
			      activationFile << topLeftX << " " << topLeftY << " " << scaleRatio << " " << decisionValue << endl;
			    }
			  else
			    {
			      newInterestLocations.push_back( (LocationOfInterest){topLeftX, topLeftY, *scaleIterator, width, height, decisionValue} );
			    }
			}
		    }
		}
	    } // end scale traversal
	  
	  interestLocations = newInterestLocations;
	  std::cout << timer.elapsed() << " s for " << cascadeLevel << " level" << std::endl;
	} // end cascadeLevel != 1 case
    } // end cascadeLevel loop
  
  
  activationFile.close();
  delete[] testNode;
}

void makeTestNode(const SpatialPyramid::SpatialPyramidType & pyramid, const vector<int> & supportVectors,
		  const vector<SpatialPyramid::util::PathAndPyramid> & trainingPyramids, svm_node* testNode, int pyramidLevels)
{
  for ( int i = 0; i < (int)supportVectors.size(); ++i )
    {
      int t = supportVectors[i];
      float matchScore = SpatialPyramid::normalizedSpatialPyramidMatchKernel(pyramid, trainingPyramids[t].pyramid, pyramidLevels);
      testNode[i].index = i+1;
      testNode[i].value = matchScore;
    } // end make sparse test node
  
  
  testNode[supportVectors.size()].index = -1;
  testNode[supportVectors.size()].value = 0;
}


void loadPyramid( const fs::path & pyramidPath, vector<SpatialPyramid::util::PathAndPyramid> & pyramids )
{
  //cout << "Loading " << pyramidPath.string() << endl;

  fs::ifstream pyramidFile(pyramidPath);
  SpatialPyramid::util::PathAndPyramid pathAndPyramid;
  pathAndPyramid.path = pyramidPath;
  SpatialPyramid::load(pathAndPyramid.pyramid, pyramidFile);
  pyramidFile.close();

  pyramids.push_back(pathAndPyramid);
}

