/*! SubsetCodewords.cpp

  Produces a codeword file by taking an initial codeword file and
  extracting a subset of codewords defined by a provided
  sub-window. The subset is described by coordinates relative to the
  given sub-window's corner.

  Useful when you're extracting many negative windows from one large
  test image. For example, I run a test of an initial model over a
  bunch of known negative images, and then take the locations that are
  reported as false positives as negative examples in the next round.

  @author Sancho McCann
*/

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <util/util.hpp>
#include <sift/sift.hpp>
#include <spatialpyramid/spatial_pyramid.hpp>
#include <vector>
#include <string>

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace SIFT = sjm::SIFT;
namespace SpatialPyramid = sjm::SpatialPyramid;
using namespace std;

int main(int argc, char** argv)
{
  int tlx, tly, brx, bry;
  fs::path inputPath, outputPath;
  bool singleScaleOnly;

  po::options_description command_line_options("Command line options");
  command_line_options.add_options()
    ("help,H", "produce help message")
    ("tlx", po::value<int>(), "the top left x coordinate of the sub-window")
    ("tly", po::value<int>(), "the top left y coordinate of the sub-window")
    ("brx", po::value<int>(), "the bottom right x coordinate of the sub-window")
    ("bry", po::value<int>(), "the bottom right y coordinate of the sub-window")
    ("input,i", po::value<string>(), "the input set of codewords")
    ("output,o", po::value<string>(), "a subset of codewords")
    ("single-scale,s", "flag to extract only the finest scale")
    ;

  try {
    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, command_line_options), vm);
    po::notify(vm);
    
    if ( vm.count("help") || !vm.count("tlx") || !vm.count("tly") || !vm.count("brx") || !vm.count("bry") ||
	 !vm.count("input") || !vm.count("output") )
      {
	cout << command_line_options << endl;
	exit(1);
      }
    else
      {
	tlx = vm["tlx"].as<int>();
	tly = vm["tly"].as<int>();
	brx = vm["brx"].as<int>();
	bry = vm["bry"].as<int>();
	inputPath = fs::path(vm["input"].as<string>());
	outputPath = fs::path(vm["output"].as<string>());
	singleScaleOnly = vm.count("single-scale");
      }
  } catch ( po::invalid_command_line_syntax ) {
    cout << command_line_options << endl;
    exit(1);
  }
  
  int numWords;
  fs::ifstream codewordFile(inputPath, ios::binary);
  codewordFile.read((char*)&numWords, sizeof(int));

  set<float> scaleSet;
  SpatialPyramid::CodewordSet subset;
  for ( int i = 0; i < numWords; i++ )
    {
      typedef SIFT::DescriptorSet::value_type::IndexType IndexType;
      int word;
      IndexType x;
      IndexType y;
      float scale;
      
      codewordFile.read((char*)&word, sizeof(word));
      codewordFile.read((char*)&x, sizeof(x));
      codewordFile.read((char*)&y, sizeof(y));
      codewordFile.read((char*)&scale, sizeof(scale));
      scaleSet.insert(scale);

      if ( x >= tlx-1 && x < brx && y >= tly-1 && y < bry )
	{
	  x = x - (tlx-1);
	  y = y - (tly-1);
	  subset.insert((SpatialPyramid::LocatedCodeword){word, x, y, scale});
	}
    }
  codewordFile.close();

  if ( singleScaleOnly )
    {
      float desiredScale = *(scaleSet.begin());
      SpatialPyramid::CodewordSet singleScaleSubset;
      for ( SpatialPyramid::CodewordSet::iterator i = subset.begin(); i != subset.end(); ++i )
	{
	  if ( abs(i->scale - desiredScale) < numeric_limits<float>::epsilon() )
	    {
	      singleScaleSubset.insert(*i);
	    }
	}
      subset.clear();
      subset = singleScaleSubset;
    }

  int numSubset = subset.size();
  fs::ofstream subsetFile(outputPath, ios::binary);  
  subsetFile.write((char*)&numSubset, sizeof(int));
  for ( SpatialPyramid::CodewordSet::iterator i = subset.begin(); i != subset.end(); ++i )
    {
      typedef SIFT::DescriptorSet::value_type::IndexType IndexType;
      int word;
      IndexType x;
      IndexType y;
      float scale;
      word = i->id; x = i->x; y = i->y; scale = i->scale;
      
      subsetFile.write((char*)&word, sizeof(word));
      subsetFile.write((char*)&x, sizeof(x));
      subsetFile.write((char*)&y, sizeof(y));
      subsetFile.write((char*)&scale, sizeof(scale));
    }
  subsetFile.close();

  return 0;
}

void accumulateHistograms( const fs::path & codewordPath, vector<vector<int> > & histograms, int numCodewords )
{
  int numWords;

  fs::ifstream codewordFile(codewordPath, ios::binary);
  codewordFile.read((char*)&numWords, sizeof(int));
  
  vector<int> histogram(numCodewords, 0);
  for ( int i = 0; i < numWords; i++ )
    {
      SIFT::DescriptorSet::value_type::IndexType dummy;
      float floatDummy;
      int word;
      codewordFile.read((char*)&word, sizeof(int));
      codewordFile.read((char*)&dummy, sizeof(dummy)); // x
      codewordFile.read((char*)&dummy, sizeof(dummy)); // y
      codewordFile.read((char*)&floatDummy, sizeof(floatDummy)); // scale

      histogram[word]++;
    }
  histograms.push_back(histogram);

  codewordFile.close();
}
