#pragma once

#include <map>
#include <vector>
#include <utility>
#include <boost/gil/gil_all.hpp>
#include <boost/gil/extension/io/jpeg_io.hpp>
#include <boost/gil/extension/io/jpeg_dynamic_io.hpp>
#include <boost/gil/extension/toolbox/hsv.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include "sparse_histogram.hpp"
#include "histogram_defs.hpp"
#include "../spatialpyramid/codeword_set.hpp"

namespace sjm { namespace histograms {
    
  boost::numeric::ublas::vector<float> regionHistogram( IntegralHistogram & i, int l, int t, int r, int b )
  {
    return i[std::make_pair(r,b)] - i[std::make_pair(r,t)] - i[std::make_pair(l,b)] + i[std::make_pair(l,t)];
  }

  IntegralHistogram integralHistogramFromCodewordSet( SpatialPyramid::CodewordSet & codewordSet, int width, int height, int dimensions )
  {
    IntegralHistogram integralHistogram;

    typedef sjm::SpatialPyramid::CodewordSet::nth_index<1>::type codeword_set_by_y;
    codeword_set_by_y& y_index = codewordSet.get<1>();

    for (int y = 0; y < height; ++y )
      {
	codeword_set_by_y::iterator y_lower = y_index.lower_bound(y);
	codeword_set_by_y::iterator y_upper = y_index.upper_bound(y+1);
	for (int x = 0; x < width; ++x )
	  {
	    boost::numeric::ublas::vector<float> nextHistogram;
	    if ( y == 0 and x == 0 )
	      {
		nextHistogram = boost::numeric::ublas::zero_vector<float>(dimensions);
	      }
	    else if ( y == 0 and x != 0 )
	      {
		nextHistogram = integralHistogram[std::make_pair(x-1,y)];
	      }
	    else if ( y != 0 and x == 0 )
	      {
		nextHistogram = integralHistogram[std::make_pair(x, y-1)];
	      }
	    else
	      {
		nextHistogram = integralHistogram[std::make_pair(x-1,y)] + integralHistogram[std::make_pair(x,y-1)] - integralHistogram[std::make_pair(x-1,y-1)];
	      }

	    for ( codeword_set_by_y::iterator it = y_lower; it != y_upper; ++it )
	      {
		if ( it->x >= x && it->x < x+1 )
		  {
		    nextHistogram[it->id] += 1;
		  }
	      }
	    
	    integralHistogram[std::make_pair(x,y)] = nextHistogram;
	  }
      }
    
    return integralHistogram;
  }

  IntegralHistogram integralHistogramFromView( boost::gil::hsv32f_view_t & hsvView )
  {
    IntegralHistogram integralHistogram;
    
    for ( int y = 0; y < hsvView.height(); ++y )
      {
	boost::gil::hsv32f_view_t::x_iterator hsv_x_it = hsvView.row_begin( y );
	for ( int x = 0; x < hsvView.width(); ++x )
	  {
	    float h = boost::gil::get_color(hsv_x_it[x], boost::gil::hsv_color_space::hue_t());
	    float s = boost::gil::get_color(hsv_x_it[x], boost::gil::hsv_color_space::saturation_t());
	    float v = boost::gil::get_color(hsv_x_it[x], boost::gil::hsv_color_space::value_t());

	    boost::numeric::ublas::vector<float> nextHistogram;
	    if ( y == 0 and x == 0 )
	      {
		nextHistogram = boost::numeric::ublas::zero_vector<float>(100);
	      }
	    else if ( y == 0 and x != 0 )
	      {
		nextHistogram = integralHistogram[std::make_pair(x-1,y)];
	      }
	    else if ( y != 0 and x == 0 )
	      {
		nextHistogram = integralHistogram[std::make_pair(x, y-1)];
	      }
	    else
	      {
		nextHistogram = integralHistogram[std::make_pair(x-1,y)] + integralHistogram[std::make_pair(x,y-1)] - integralHistogram[std::make_pair(x-1,y-1)];
	      }
	    
	    if ( s > 0.1 and v > 0.2 )
	      {
		int sBin = (int)(s * 10);
		sBin = sBin % 10;
		int hBin = (int)(h * 10);
		hBin = hBin % 10;
		
		for ( int soff = -1; soff <= 1; ++soff )
		  {
		    int s = sBin + soff;
		    if ( s >= 0 && s < 10 ) // if the adjacent bin exists
		      {
			for ( int hoff = -1; hoff <= 1; ++hoff )
			  {
			    int h = (hBin + hoff) % 10;
			    if ( h < 0 )
			      h += 10;
			    int histogramBin = s * 10 + h;
			    nextHistogram[histogramBin] += 1;
			  }
		      }
		  }
	      }

	    integralHistogram[std::make_pair(x,y)] = nextHistogram;
	  }
      }
    
    return integralHistogram;
  }

  IntegralHistogram integralHistogramFromImage( std::string imagePath )
  {
    boost::gil::rgb16_image_t image;
    boost::gil::jpeg_read_and_convert_image(imagePath, image);
    boost::gil::hsv32f_image_t hsvImage( boost::gil::view( image ).dimensions() );
    boost::gil::copy_pixels( boost::gil::color_converted_view<boost::gil::hsv32f_pixel_t>( boost::gil::view( image ))
		      , boost::gil::view( hsvImage ));
    
    boost::gil::hsv32f_view_t hsvView = boost::gil::view(hsvImage);

    return integralHistogramFromView(hsvView);
  }

}}
