#include <boost/python.hpp>
#include <boost/python/list.hpp>
#include <boost/python/iterator.hpp>
#include <boost/python/suite/indexing/vector_indexing_suite.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/gil/extension/io/png_io.hpp>
#include <boost/foreach.hpp>
#include <stdexcept>
#include <string>
#include <iterator>
#include "sift.hpp"

using namespace boost::python;
using namespace sjm::SIFT;
namespace fs = boost::filesystem;

SiftExtractor::SiftExtractor(Parameters params)
{
  this->params = params;
  cvImage = 0;
}

SiftExtractor::~SiftExtractor()
{
  if (cvImage)
    {
      cvReleaseImage(&cvImage);
    }
}

void
SiftExtractor::loadImage(std::string imagePath)
{
  if ( !fs::exists(imagePath) )
    {
      throw std::runtime_error("File not found: " + imagePath);
    }
  
  cvImage = cvLoadImage( imagePath.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
  if ( cvImage == 0 )
    {
      throw std::runtime_error("Error loading image: " + imagePath);
    }
}

void
SiftExtractor::buildScalePyramid()
{
  if ( !cvImage )
    {
      throw std::runtime_error("Attempt to process before loading image.");
    }

  float x_difference_operator[] = {1, 0, -1};
  float y_difference_operator[] = {-1, 0, 1};
  gil::kernel_1d_fixed<float, 3> x_kernel(x_difference_operator, 1);
  gil::kernel_1d_fixed<float, 3> y_kernel(y_difference_operator, 1);

  gil::gray8c_view_t original = gil::interleaved_view(cvImage->width,
						      cvImage->height,
						      (const gil::gray8_pixel_t*)cvImage->imageData,
						      (ptrdiff_t)cvImage->widthStep);

  float sigmaRatio = std::pow(2.0f, 1.0f / params.octaveSteps);
  int minSize = 2 * BorderDist + 2;
  int octave = 0;
  int step = 0;

  float octSize = 1.0f;
  float sigma = 0.5;

  imagePyramid.push_back(gil::gray8_image_t(original.width() * 2, original.height() * 2));
  gil::resize_view(original, gil::view(imagePyramid[0]), gil::bilinear_sampler());
  xgradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
  ygradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
  gil::convolve_rows_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), x_kernel, gil::view(xgradPyramid.back()), gil::convolve_option_extend_constant);
  gil::convolve_cols_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), y_kernel, gil::view(ygradPyramid.back()), gil::convolve_option_extend_constant);
  sigma *= 2.0;
  octSize *= 0.5;
  
  float additionalSigma = std::sqrt(InitSigma * InitSigma - sigma*sigma);
  gaussianBlur(gil::view(imagePyramid[0]), additionalSigma);
  scales.push_back(InitSigma);
  octSizes.push_back(octSize);
    
  while ( imagePyramid.back().width() > minSize and
	  imagePyramid.back().height() > minSize )
    {
      sigma = InitSigma;
      for ( step = 0; step < params.octaveSteps; ++step )
	{
	  // Increase scale
	  float increaseSigma = sigma * std::sqrt(sigmaRatio * sigmaRatio - 1.0f);
	  if ( step < params.octaveSteps - 1 )
	    {
	      imagePyramid.push_back(gil::gray8_image_t( imagePyramid.back().width(), imagePyramid.back().height() ) );
	      gil::copy_pixels(gil::view(imagePyramid[imagePyramid.size()-2]), gil::view(imagePyramid.back()));
	      gaussianBlur(gil::view(imagePyramid.back()), increaseSigma);

	      xgradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
	      ygradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
	      gil::convolve_rows_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), x_kernel, gil::view(xgradPyramid.back()), gil::convolve_option_extend_constant);
	      gil::convolve_cols_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), y_kernel, gil::view(ygradPyramid.back()), gil::convolve_option_extend_constant);

	      sigma *= sigmaRatio;
	      scales.push_back(sigma);
	      octSizes.push_back(octSize);
	    }
	  else
	    {
	      // Increase scale into a temp image...
	      gil::gray8_image_t tempImage(imagePyramid.back().width(), imagePyramid.back().height());
	      gil::copy_pixels(gil::view(imagePyramid.back()), gil::view(tempImage));
	      gaussianBlur(gil::view(tempImage), increaseSigma);

	      xgradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
	      ygradPyramid.push_back(gil::gray16s_image_t(imagePyramid.back().dimensions()));
	      gil::convolve_rows_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), x_kernel, gil::view(xgradPyramid.back()), gil::convolve_option_extend_constant);
	      gil::convolve_cols_fixed<gil::gray16s_pixel_t>(gil::view(imagePyramid.back()), y_kernel, gil::view(ygradPyramid.back()), gil::convolve_option_extend_constant);

	      sigma *= sigmaRatio;
	      
	      // and downsample onto the back of the pyramid
	      imagePyramid.push_back(gil::gray8_image_t( (tempImage.width() + 1) / 2, (tempImage.height() + 1) / 2));
	      gil::copy_pixels(gil::subsampled_view(gil::view(tempImage), 2, 2), gil::view(imagePyramid.back()));
	      scales.push_back(InitSigma);
	      octSize *= 2;
	      octSizes.push_back(octSize);
	      ++octave;
	    }
	}
    }
}

void
SiftExtractor::outputPyramid()
{
  std::vector<gil::gray8_image_t>::const_iterator image = imagePyramid.begin();
  std::vector<gil::gray16s_image_t>::const_iterator ximage = xgradPyramid.begin();
  std::vector<gil::gray16s_image_t>::const_iterator yimage = ygradPyramid.begin();
  
  int i = 0;
  for ( ; image != imagePyramid.end(); ++image, ++ximage, ++yimage )
    {
      gil::png_write_view( "image-" + boost::lexical_cast<std::string>(i) + ".png", gil::const_view(*image) );
      ++i;
    }
}

boost::python::list
SiftExtractor::scaleRange(float minRadius, float maxRadius)
{
  boost::python::list range;

  float minScale = scaleFromRadius(minRadius);
  float maxScale = scaleFromRadius(maxRadius);

  float sigmaRatio = std::pow(2.0f, 1.0f / params.octaveSteps);
  for ( unsigned int scaleIndex = 0; scaleIndex < scales.size(); scaleIndex++ )
    {
      float sigmaRelativeToOriginal = InitSigma * std::pow(sigmaRatio, (float)scaleIndex);
      if ( sigmaRelativeToOriginal >= minScale and sigmaRelativeToOriginal <= maxScale )
	{
	  range.append(scaleIndex);
	}
    }

  return range;
}

std::vector<boost::python::list>
SiftExtractor::getDescriptorSet(std::string imagePath)
{
  std::vector<boost::python::list> pythonResultSet;
  IplImage* cvImage = 0;

  RandomGeneratorType random_generator;
  random_generator.seed(time(0));
  BernoulliType bernoulli(params.percentage);
  VariateGeneratorType myRand(random_generator, bernoulli);

  cvImage = cvLoadImage( imagePath.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
  if ( cvImage > 0 )
    {
      // Passing the cvImage in a way that allows wrapping in a GIL image view
      DescriptorSet descriptorSet = ComputeDenseSiftFeaturesGray8((unsigned char*)cvImage->imageData,
								  (ptrdiff_t)cvImage->widthStep,
								  cvImage->width,
								  cvImage->height,
								  0, 0, cvImage->width, cvImage->height,
								  params, myRand);

      for ( DescriptorSet::const_iterator d = descriptorSet.begin(); d != descriptorSet.end(); ++d )
	{
	  boost::python::list descriptor;
	  BOOST_FOREACH(SiftDescriptorType::value_type bin, *(d->descriptor()))
	    {
	      descriptor.append(bin);
	    }
	  descriptor.append(d->x());
	  descriptor.append(d->y());
	  descriptor.append(d->scale());
	  pythonResultSet.push_back(descriptor);
	}
      
      cvReleaseImage(&cvImage);
    }
  else
    {
      throw std::runtime_error("Error loading image: " + imagePath);
    }

  return pythonResultSet;
}

boost::python::list
SiftExtractor::getDescriptor(float x, float y, unsigned int scaleIndex)
{
  if ( scaleIndex >= imagePyramid.size() )
    {
      throw std::out_of_range( "scaleIndex out of imagePyramid range" );
    }

  float sigma = scales[scaleIndex];
  float octSize = octSizes[scaleIndex];
  
  boost::shared_ptr<SiftDescriptorType> potentialDescriptor;
  extractDescriptor(gil::view(xgradPyramid[scaleIndex]),
		    gil::view(ygradPyramid[scaleIndex]),
		    y / octSize, // row
		    x / octSize, // column
		    sigma,
		    params,
		    potentialDescriptor);

  boost::python::list descriptor;
  if ( isValidDescriptor(potentialDescriptor) )
    {
      BOOST_FOREACH(SiftDescriptorType::value_type bin, *potentialDescriptor)
	{
	  descriptor.append(bin);
	}
    }

  return descriptor;
}

BOOST_PYTHON_MODULE(sift_ext)
{  
  class_<std::vector<boost::python::list> >("DescriptorSet")
    .def("__iter__", boost::python::iterator<std::vector<boost::python::list> >())
    //    .def(boost::python::vector_indexing_suite<std::vector<boost::python::list> >())
    ;

  class_<sjm::SIFT::Parameters>("SiftParameters")
    .def_readwrite("rotationInvariance", &sjm::SIFT::Parameters::rotationInvariance)
    .def_readwrite("noNormalize", &sjm::SIFT::Parameters::noNormalize)
    .def_readwrite("discard", &sjm::SIFT::Parameters::discard)
    .def_readwrite("octaveSteps", &sjm::SIFT::Parameters::octaveSteps)
    .def_readwrite("percentage", &sjm::SIFT::Parameters::percentage)
    .def_readwrite("minRadius", &sjm::SIFT::Parameters::minRadius)
    .def_readwrite("multiscale", &sjm::SIFT::Parameters::multiscale)
    ;

  def("scaleFromRadius", sjm::SIFT::scaleFromRadius);

  class_<sjm::SIFT::SiftExtractor>("SiftExtractor", init<sjm::SIFT::Parameters>())
    .def("loadImage", &sjm::SIFT::SiftExtractor::loadImage)
    .def("buildScalePyramid", &sjm::SIFT::SiftExtractor::buildScalePyramid)
    .def("outputPyramid", &sjm::SIFT::SiftExtractor::outputPyramid)
    .def("scaleRange", &sjm::SIFT::SiftExtractor::scaleRange)
    .def("getDescriptor", &sjm::SIFT::SiftExtractor::getDescriptor)
    .def("getDescriptorSet", &sjm::SIFT::SiftExtractor::getDescriptorSet)
    ;
}


