#pragma once

#include <boost/gil/gil_all.hpp>
#include <boost/gil/gil_concept.hpp>
#include <boost/gil/extension/io/png_io.hpp>
#include <boost/gil/extension/io/jpeg_io.hpp>
#include <boost/gil/extension/numeric/convolve.hpp>
#include <boost/gil/extension/numeric/kernel.hpp>
#include <boost/gil/extension/numeric/resample.hpp>
#include <boost/gil/extension/numeric/sampler.hpp>
#include <boost/numeric/conversion/bounds.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/algorithm.hpp>
#include <boost/lambda/if.hpp>
#include <boost/lambda/loops.hpp>
#include <boost/lambda/switch.hpp>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <boost/multi_array.hpp>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/mem_fun.hpp>
#include <boost/thread.hpp>
#include <boost/random.hpp>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <stdexcept>
#include <cassert>
#include <string>

#if defined(SJM_PYSIFT)
#include <boost/python.hpp>
#include <boost/python/list.hpp>
#endif

namespace sjm {
  /*!  
    Functions and structures implementing dense SIFT feature
    extraction

    This release of this code has been stripped of the source from
    David Lowe's demo program for the extraction of invariant
    keypoints available at http://www.cs.ubc.ca/~lowe/keypoints

    It will not producing meaningful descriptors in its current state

    @author Sancho McCann
  */
  namespace SIFT
  {
    namespace fs = boost::filesystem;
    namespace ublas = boost::numeric::ublas;
    namespace bl = boost::lambda;
    namespace gil = boost::gil;
  
    typedef ublas::vector<float> SiftDescriptorType; ///< The type that the sift descriptors will be stored in
    typedef ublas::zero_vector<int> SiftDescriptorZeroType; ///< The type that produces a zero sift descriptor
  
    typedef boost::multi_array<float, 3> IndexArrayType; ///< Used locally to build the SIFT descriptor
    typedef boost::multi_array<float, 3> BinCacheType; ///< Used locally to cache bins for cheap dense sift extraction

    typedef boost::mt19937 RandomGeneratorType; ///< Type exposed to allow user-code to construct the VariateGeneratorType
    typedef boost::bernoulli_distribution<float> BernoulliType; ///< Type exposed to allow user-code to construct the VariateGeneratorType
    typedef boost::variate_generator<RandomGeneratorType &, BernoulliType> VariateGeneratorType; ///< The VariateGenerator used in random sampling

    /*!
      Parameters that effect the behaviour of the SIFT extraction
    */
    struct Parameters
    {
      Parameters()
      {
	rotationInvariance = false;
	noNormalize = 0;
	discard = false;
	octaveSteps = 3;
	multiscale = true;
	percentage = 1;
	minRadius = 0;
      }

      bool rotationInvariance; ///< Describe orientations relative to the dominant orientation of the patch
      float noNormalize; ///< Skip normalization for descriptors with initial magnitude less than this value
      bool discard; ///< Discard descriptors that are not normalized
      int octaveSteps; ///< Number of steps per octave (doubling) of the scale-space parameter
      bool multiscale; ///< Perform multi-scale extraction (setting this to false will cause extraction of only the finest scale
      float percentage; ///< Percentage of descriptors to extract (randomly sampled)
      float minRadius; ///< Descriptors will not be extracted at radius smaller than this
    };

    ///@todo all of these should be parameters
    const int ValidDescriptorLength = 128; ///< All true descriptors will have this length
    const int Dimensions = ValidDescriptorLength; ///< Alias
    const int InvalidDescriptorLength = 1; ///< Invalid descriptors will be assigned this length
  
    const int IndexSize = 4; ///< Number of bins along one side of the sift descriptor window
    const int OriSize = 8; ///< Number of orientation bins in the sift descriptor
    const int MagFactor = 3; ///< Determines the size of each bin
    const float OriSigma = 1.5; ///< Multiplier for scale to weight contributions to the orientation histogram
    const int OriBins = 36; ///< Number of histogram bins when finding peaks in orientation
    const int BorderDist = 5; ///< The minimum distance from the border that a feature will be extracted at
    const float InitSigma = 1.6; ///< The amount of initial blur of the doubled and additionally blurred image

    /*!
      Utility function to give the radius of a descriptor in pixels
      of the original image given the scale of the descriptor.

      @param the scale, relative to the ideal original image (stored with the descriptor)
      @return float the radius
    */
    float radiusFromScale(float scale)
    {
      float spacing = scale * MagFactor;
      float radius = (1.414f * spacing * (IndexSize+1)) / 4.0;
      return radius;
    }

    /*!
      Utility function to give the scale required to extract a
      descriptor of the desired radius.

      @param float the radius
      @return float the required scale
    */
    float scaleFromRadius(float radius)
    {
      float sigma = (4.0 * radius) / (1.414f * MagFactor * (IndexSize+1));
      return sigma;
    }
    
    /*!
      Utility function for writing SIFT parameters to file.
      
      @param parameters the parameters struct
      @param outputStream an open binary output stream
    */
    inline void writeParameters(const Parameters & parameters, fs::ofstream & outputStream)
    {
      outputStream.write((char*)&parameters.rotationInvariance, sizeof(parameters.rotationInvariance));
      outputStream.write((char*)&parameters.noNormalize, sizeof(parameters.noNormalize));
      outputStream.write((char*)&parameters.discard, sizeof(parameters.discard));
      outputStream.write((char*)&parameters.multiscale, sizeof(parameters.multiscale));
      outputStream.write((char*)&parameters.octaveSteps, sizeof(parameters.octaveSteps));
      outputStream.write((char*)&parameters.percentage, sizeof(parameters.percentage));
      outputStream.write((char*)&parameters.minRadius, sizeof(parameters.minRadius));
    }

    /*!
      Utility function for reading SIFT parameters to file.
      
      @param parameters the parameters struct
      @param inputStream an open binary input stream
    */
    inline void readParameters(Parameters & parameters, fs::ifstream & inputStream)
    {
      inputStream.read((char*)&parameters.rotationInvariance, sizeof(parameters.rotationInvariance));
      inputStream.read((char*)&parameters.noNormalize, sizeof(parameters.noNormalize));
      inputStream.read((char*)&parameters.discard, sizeof(parameters.discard));
      inputStream.read((char*)&parameters.multiscale, sizeof(parameters.multiscale));
      inputStream.read((char*)&parameters.octaveSteps, sizeof(parameters.octaveSteps));
      inputStream.read((char*)&parameters.percentage, sizeof(parameters.percentage));
      inputStream.read((char*)&parameters.minRadius, sizeof(parameters.minRadius));
    }
    
    /*!
      Utility function for saving SIFT descriptors to file.
   
      @param descriptor a SIFT descriptor
      @param outputStream an open file output stream
    */
    template <typename X>
    inline void save(const ublas::vector<X> & descriptor, fs::ofstream & outputStream)
    {
      typename ublas::vector<X>::const_iterator it = descriptor.begin();
      for ( ; it != descriptor.end(); ++it )
	{
	  X element = *it;
	  outputStream.write((char*)&element, sizeof(X));
	}
    }
  
    /*!
      Utility function for loading SIFT descriptors from file.
   
      @param descriptor the destination SIFT descriptor into which the next descriptor from file is read into
      @param inputStream an open file input stream
    */
    template <typename X>
    inline void load(ublas::vector<X> & descriptor, fs::ifstream & inputStream)
    {
      X element = 0;
      typename ublas::vector<X>::iterator it = descriptor.begin();
      for ( ; it != descriptor.end(); ++it )
	{
	  inputStream.read((char*)&element, sizeof(X));
	  *it = element;
	}
    }
  
    /*!
      A SIFT descriptor combined with its location and scale at which it was found in the image.
    */
    template <typename T, typename FeatureDescriptorType>
    class Keypoint
    {
    public:
      typedef T IndexType; ///< Exposes the type that the location information is stored in (int, float, etc.)
    
      /*!
	Constructor
     
	@param x the x pixel at which this descriptor was extracted from
	@param y the y pixel at which this descriptor was extracted from
	@param scale the scale at which this descriptor was extracted
	@param descriptor the SIFT descriptor
      */
      explicit Keypoint(T x, T y, float scale, boost::shared_ptr<FeatureDescriptorType> descriptor) : m_x(x), m_y(y), m_scale(scale), m_descriptor(descriptor) {};
     /*!
	Destructor... no special behaviour
      */
      ~Keypoint() {};
    
      /*!
	Get x
	@return the x coordinate at which this descriptor was extracted from
      */
      T x() const { return m_x; }
    
      /*!
	Get y
	@return the y coordinate at which this descriptor was extracted from
      */
      T y() const { return m_y; }
    
      /*!
	Get (x,y)
	@return a pair representing the x-y coordinates at which this descriptor was extracted from
      */
      std::pair<T, T> location() const { return std::make_pair(m_x, m_y); }
    
      /*!
	Gets the scale of the descriptor
	@return the scale
      */
      float scale() const { return m_scale; }
    
      /*!
	Get descriptor
	@return a const pointer to the stored descriptor
      */
      const boost::shared_ptr<FeatureDescriptorType> descriptor() const { return m_descriptor; }
    
    private:
      T m_x; ///< The x coordinate
      T m_y; ///< The y coordinate
      float m_scale; ///< The scale at which this descriptor was extracted at
      boost::shared_ptr<FeatureDescriptorType> m_descriptor; ///< The descriptor
    };
  
    typedef Keypoint<float, SiftDescriptorType> DescriptorType; ///< The type that the located descriptors will be exposed to user code as
  
    /*!
      Used only as a helper type for easy extraction of the scale index in the descriptor set.
    */
    struct scale{};
  
    /*!
      The type that a set of descriptors will be stored in. This is a boost::multi_index set, with ordered indices into both the x and y
      components of a descriptor's location. This allows for relatively fast 2d range finding --- not optimal, but faster than an maintaining only
      a single index as a standard stl set or multiset would allow.
    */
    typedef boost::multi_index_container<
      DescriptorType,
      boost::multi_index::indexed_by<
      // sort set by x and y locations for quick location retrieval
      boost::multi_index::ordered_non_unique<boost::multi_index::const_mem_fun<DescriptorType, DescriptorType::IndexType, &DescriptorType::x> >,
      boost::multi_index::ordered_non_unique<boost::multi_index::const_mem_fun<DescriptorType, DescriptorType::IndexType, &DescriptorType::y> >,
      boost::multi_index::ordered_non_unique<boost::multi_index::tag<scale>, boost::multi_index::const_mem_fun<DescriptorType, float, &DescriptorType::scale> >
    >
    > DescriptorSet;
  
    /*!
      Descriptor distance functor object.
   
      Useful to downstream users if needing a distance function between feature descriptors (clustering).
      If using a different structure for the SiftFeature, define another distance functor for users of this library.
    */
    struct SiftFeatureDistanceFunctor
    {
      /*!
	The L2 norm.
	@param a a descriptor
	@param b another descriptor
     
	@return L2Norm between two descriptors in feature space
      */
      float operator()(const boost::shared_ptr<SiftDescriptorType> a, const boost::shared_ptr<SiftDescriptorType> b)
      {
	return ublas::norm_2(*a - *b);
      }
    };
    
    /*!
      Gives an invalid sift descriptor.
   
      @return a pointer to an SIFT::SiftDescriptorType object with length InvalidDescriptorLength
    */
    inline boost::shared_ptr<SiftDescriptorType> createInvalidDescriptor()
    {
      return boost::shared_ptr<SiftDescriptorType>(new SiftDescriptorType(InvalidDescriptorLength));
    }
  
    /*!
      Tests descriptor validity.
   
      @param descriptor the validity of this descriptor will be checked
      @return true if the descriptor is of valid length, false otherwise
    */
    inline bool isValidDescriptor(const boost::shared_ptr<SiftDescriptorType> descriptor)
    {
      return (int)descriptor->size() == ValidDescriptorLength;
    }
  
    /*!
      Returns the dominant orientation within the support of a requested SIFT descriptor.
   
      x_grad and y_grad must implement the gil::ImageViewConcept
   
      Implementation taken from Lowe's research code
   
      @param x_grad a GIL view into the horizontal gradient image
      @param y_grad a GIL view into the vertical gradient image
      @param frow the centre of the descriptor in the image
      @param fcol the centre of the descriptor in the image
      @param scale the scale of this descriptor (relative to the base scale of the current octave) - determines the support of the descriptor
      @return the dominant orientation in [-pi,pi]
    */
    template <typename SrcView>
    inline float getDominantOrientation(const SrcView& x_grad, const SrcView& y_grad, float frow, float fcol, float scale)
    {
      // Source removed

      return 0;
    }
  
    /*!
      The final placement of a gradient's contribution into the descriptor.
   
      @param indexArray the bins representing the descriptor - this is eventually what becomes the SIFT descriptor
      @param magnitude the magnitude of the gradient being added
      @param orientation the orientation of the gradient being added
      @param rx the rotated relative row position of the gradient being added to the descriptor
      @param cx the rotated relative column position of the gradient being added to the descriptor
    */
    template <typename ArrayOrView>
    inline void placeInIndex(ArrayOrView & indexArray, float magnitude, float orientation, float rx, float cx)
    {
      // Source removed
    }
  
    /*!
      Places a gradient from a given integer location in the original image into the appropriate (x,y,r)
      SIFT descriptor bin relative to the centre of this descriptor.
   
      Ported directly from Lowe's available research code.
   
      @param index the array of bins that samples are being accumulated into - this becomes the descriptor eventually
      @param x_grad the horizontal gradient values
      @param y_grad the vertical gradient values
      @param r the integer row (y) in the original image of the sample being considered
      @param c the integer column (x) in the original image of the sample being considered
      @param rpos the relative row offset from the centre of the descriptor
      @param cpos the relative column offset from the centre of the descriptor
      @param rx the rotated relative row offset from the centre of the descriptor
      @param cx the rotated relative column offset from the centre of the descriptor
      @param angle the dominant orientation of this descriptor - only non-zero when rotation invariance is turned on
    */
    template <typename SrcView, typename ArrayOrView>
    inline void addSample(ArrayOrView & index, const SrcView& x_grad, const SrcView& y_grad,
			  int r, int c, float rpos, float cpos, float rx, float cx, float angle)
    {
      // Source removed
    }
  
    /*!
      Gets a SIFT descriptor from gradient images at a requested location.
   
      Implementation from Lowe's research code.
   
      @param x_grad a GIL view into the image containing the horizontal gradient values
      @param y_grad a GIL view into the image containing the vertical gradient values
      @param fractional_row the centre of the descriptor in the image
      @param fractional_column the centre of the descriptor in the image
      @param scale the scale relative to the scale of the base of the octave at which this descriptor will be extracted
      @param parameters parameters controlling the sift descriptor (rotation variance, etc.)
      @param siftDescriptor passed by reference to return the result of this function
    */
    template <typename SrcView>
    inline void extractDescriptor(const SrcView& x_grad, const SrcView& y_grad,
				  float fractional_row, float fractional_column,
				  float scale, const Parameters & parameters,
				  boost::shared_ptr<SiftDescriptorType> & siftDescriptor)
    {
      siftDescriptor = boost::shared_ptr<SiftDescriptorType>(new SiftDescriptorType(ValidDescriptorLength));

      // source removed
    }

    /*!
      Populates a subview of the cached bins for fast sift extraction
   
      Implementation from Lowe's research code.
   
      @param x_grad a GIL view into the image containing the horizontal gradient values
      @param y_grad a GIL view into the image containing the vertical gradient values
      @param fractional_row the centre of the descriptor in the image
      @param fractional_column the centre of the descriptor in the image
      @param scale the scale relative to the scale of the base of the octave at which this descriptor will be extracted
      @param parameters parameters controlling the sift descriptor (rotation variance, etc.)
      @param siftDescriptor passed by reference to return the result of this function
    */
    template <typename SrcView, typename ArrayView>
    inline void populateBins(const SrcView& x_grad, const SrcView& y_grad, ArrayView & indexArray,
			     float fractional_row, float fractional_column,
			     float scale, const Parameters & parameters)
    {
      // Source removed
    }

    template <typename SrcView>
    inline BinCacheType getCachedBins(const SrcView& x_grad, const SrcView& y_grad,
				      float scale, const Parameters & parameters)
    {
      float spacing = scale * MagFactor;
      
      int xBins = int(x_grad.width() / spacing);
      int yBins = int(y_grad.height() / spacing);

      BinCacheType bins(boost::extents[xBins][yBins][OriSize]);

      for ( int y = 2; y < yBins - 1; y += IndexSize )
	{
	  for ( int x = 2; x < xBins - 1; x += IndexSize )
	    {
	      typedef BinCacheType::index_range range;
	      BinCacheType::array_view<3>::type subview =
		bins[boost::indices[range(x - IndexSize/2, x + IndexSize/2)][range(y - IndexSize/2, y + IndexSize/2)][range(0, OriSize)]];

	      float fx = x * spacing;
	      float fy = y * spacing;
	      
	      populateBins(x_grad, y_grad, subview, fy, fx, scale, parameters); 
	    }
	}
      
      return bins;
    }

    inline void getDescriptorFromBins(const BinCacheType & bins, int x, int y, float scale, const Parameters & parameters,
				      boost::shared_ptr<SiftDescriptorType> & siftDescriptor)
    {
      typedef BinCacheType::index_range range;
      BinCacheType::const_array_view<3>::type indexArray =
	bins[boost::indices[range(x - IndexSize/2, x + IndexSize/2)][range(y - IndexSize/2, y + IndexSize/2)][range(0, OriSize)]];
      
      typedef BinCacheType::const_array_view<3>::type::index IndexType;
      siftDescriptor = boost::shared_ptr<SiftDescriptorType>(new SiftDescriptorType(ValidDescriptorLength));

      // Unroll index into the sift descriptor vector
      int v = 0;
      for ( IndexType i = 0; i < IndexSize; ++i )
	{
	  for ( IndexType j = 0; j < IndexSize; ++j )
	    {
	      for ( IndexType k = 0; k < OriSize; ++k )
		{
		  (*siftDescriptor)[v++] = indexArray[i][j][k];
		}
	    }
	}
      
    }

    /*!
      Performs a Gaussian blur on an image view.
   
      Ported from Lowe's research code.
   
      @param src a view into a mutable image - must satisfy the concept of gil::MutableImageViewConcept
      @param sigma the standard deviation of the Gaussian kernel in pixel coordinates
    */
    template <typename SrcView>
    inline void gaussianBlur(SrcView& src, float sigma)
    {
      gil::gil_function_requires<gil::ImageViewConcept<SrcView> >();
      gil::gil_function_requires<gil::MutableImageViewConcept<SrcView> >();
      int width = (int)(sigma * 3.0 * 2 + 1);
      width = std::max(3, width);
      
      if ( width % 2 == 0 )
	{
	  ++width; // Make odd
	}
      
      float* gaussian = new float[width];
      float normalization = 0;
      for ( int i = 0; i < width; ++i )
	{
	  float x = i - width / 2;
	  gaussian[i] = std::exp(- x * x / (2.0 * sigma * sigma));
	  normalization += gaussian[i];
	}
      
      for ( int i = 0; i < width; ++i )
	{
	  gaussian[i] /= normalization;
	}
      
      gil::kernel_1d<float> kernel(gaussian,width,width/2);
      
      gil::convolve_rows<gil::gray32f_pixel_t>(src,kernel,src,gil::convolve_option_extend_constant);
      gil::convolve_cols<gil::gray32f_pixel_t>(src,kernel,src,gil::convolve_option_extend_constant);
      
      delete [] gaussian;
    }

    /*!
      Internal namespace to hide unnecessary details from user-code
    */
    namespace detail {
      /*!
	Parameters prescribing how the thread in the threaded
	extraction should behave. These are collected into a single
	struct because boost::bind has an upper limit on the number of
	parameters that may be used.
      */
      struct ThreadParams
      {
	float minRow; ///< The minimum row that is part of this thread's extraction grid
	float maxRow; ///< One past the maximum row of this thread's extraction grid
	float rowIncrement;
	float minCol;
	float maxCol;
	float colIncrement;
	float sigma; ///< At what scale is this extraction happening relative to the current octave (dictates the radius of the descriptor)
	float sigmaRelativeToIdealOriginal; ///< Like sigma above, but relative to the original, unscaled image
	float octSize; ///< Allows relating rows and columns in this possible scaled down view to rows and columns in the original
      };
      
      struct ThreadParamsCache
      {
	int minRow;
	int maxRow;
	float sigma;
	float sigmaRelativeToIdealOriginal;
	float octSize;
      };
      
      template<typename ViewType>
      struct GradViews
      {
	const ViewType& x_grad;
	const ViewType& y_grad;
      };
      
    }

    
    
    /*!
      The thread that will extract descriptors from a subregion of an image at a certain scale

      @param x_grad the view into the x-gradient data of the image; x_grad must satisfy the gil::ImageViewConcept
      @param y_grad the view into the y-gradient data of the image; y_grad must satisfy the gil::ImageViewConcept
      @param geometryParams parameters dictating over what region this extraction will occur and the scale of the views that are being passed in
      @param parameters parameters dictating options of the sift extraction such as rotation invariance, ignore thresholds, etc.
      @param descriptors the DescriptorSet into which extracted descriptors will be written
      @param descriptorAccessMutex manages access to the DescriptorSet descriptors by possibly multiple threads
      @param myRand the random variable generator for random sampling of descriptors
    */
    template<typename SrcView>
    inline void extractDescriptorsFromRegion(const detail::GradViews<SrcView>& views,
					     const detail::ThreadParams & geometryParams, const Parameters & parameters, DescriptorSet & descriptors,
					     boost::mutex & descriptorAccessMutex, VariateGeneratorType & myRand)
    {
      for ( float row = geometryParams.minRow; row < geometryParams.maxRow; row+=geometryParams.rowIncrement )
	{
	  for ( float col = geometryParams.minCol; col < geometryParams.maxCol; col+=geometryParams.colIncrement )
	    {
	      // A bernoulli random variable
	      if ( myRand() )
		{
		  boost::shared_ptr<SiftDescriptorType> potentialDescriptor;
		  extractDescriptor(views.x_grad, views.y_grad, row, col, geometryParams.sigma, parameters, potentialDescriptor);
		  
		  if (isValidDescriptor(potentialDescriptor))
		    {
		      DescriptorType d(col*geometryParams.octSize,row*geometryParams.octSize,geometryParams.sigmaRelativeToIdealOriginal,potentialDescriptor);
		      
		      // Wait for, then lock this mutex for the rest
		      // of the "if" scope. This mutex is shared only
		      // with other threads running this function.
		      boost::mutex::scoped_lock L(descriptorAccessMutex);
		      descriptors.insert( d );
		  
		    }
		}
	    }
	}
    }
    
    inline void extractDescriptorsFromRegion_Cached(const BinCacheType & bins,
						    const detail::ThreadParamsCache & geometryParams, const Parameters & parameters, DescriptorSet & descriptors, boost::mutex & descriptorAccessMutex, VariateGeneratorType & myRand)
    {
      float spacing = geometryParams.sigma * MagFactor;
      
      for ( int y = std::max(2, geometryParams.minRow); y <= std::min(geometryParams.maxRow, int(bins.shape()[1])-2); ++y )
	{
	  for ( int x = 2; x < int(bins.shape()[0]) - 1; ++x )
	    {
	      float col = x * spacing;
	      float row = y * spacing;
	      
	      if ( myRand() )
		{
		  boost::shared_ptr<SiftDescriptorType> potentialDescriptor;
		  getDescriptorFromBins(bins, x, y, geometryParams.sigma, parameters, potentialDescriptor);
		  
		  if (isValidDescriptor(potentialDescriptor))
		    {
		      DescriptorType d(col*geometryParams.octSize,row*geometryParams.octSize,geometryParams.sigmaRelativeToIdealOriginal,potentialDescriptor);		      
		      // Wait for, then lock this mutex for the rest
		      // of the "if" scope. This mutex is shared only
		      // with other threads running this function.
		      boost::mutex::scoped_lock L(descriptorAccessMutex);
		      descriptors.insert( d );
		    }
		}
	    }
	}
    }
				      
    /*! A generic algorithm to compute a set of dense SIFT descriptors
      over a GIL image view. This function computes SIFT descriptors
      for each square in a dense grid covering the image. First, the x
      and y gradients are computed for the entire image. Second, the
      SIFT descriptors are calculated for small subimage views
      completely covering the image in a dense grid.
    
      @param src an image view satisfying GIL's ImageViewConcept
      @param parameters a struct of parameters prescribing how sift descriptors will be extracted
      @param myRand a bernoulli random variable generator for random sampling of descriptors
      @return a boost::multi_index DescriptorSet
    */
    template <typename SrcView>
    inline DescriptorSet denseSiftFeatures(const SrcView& src, const Parameters & parameters, VariateGeneratorType & myRand)
    {
      gil::gil_function_requires<gil::ImageViewConcept<SrcView> >();
      
      gil::gray8_image_t my_gray_image(src.dimensions());
      gil::copy_and_convert_pixels(src, gil::view(my_gray_image));
      gil::gray8_view_t my_gray_view = gil::view(my_gray_image);
      
      DescriptorSet descriptors; // the Located SIFT descriptors
      
      float minScale = scaleFromRadius(parameters.minRadius);

      float octSize = 1.0f;
      float sigma = 0.5f; // before doubling
      
      gil::gray8_image_t currImage(my_gray_image.width() * 2, my_gray_image.height() * 2);
      gil::resize_view(gil::view(my_gray_image), gil::view(currImage), gil::bilinear_sampler());
      sigma *= 2.0f;
      octSize *= 0.5;
      float additionalSigma = std::sqrt(InitSigma * InitSigma - sigma*sigma);
      
      gaussianBlur(gil::view(currImage), additionalSigma);
      
      float sigmaRatio = std::pow(2.0f, 1.0f / parameters.octaveSteps);
      int minSize = 2 * BorderDist + 2;
      int octave = 0;
      bool completedOneScale = false;
      while ( currImage.width() > minSize and currImage.height() > minSize and (parameters.multiscale or (not completedOneScale) ) )
	{
	  sigma = InitSigma;
	  for ( int i = 0; i < parameters.octaveSteps and (parameters.multiscale or (not completedOneScale)); ++i )
	    {
	      float sigmaRelativeToIdealOriginal = InitSigma * std::pow(2.0f, octave) * std::pow(sigmaRatio, i);
	      if ( sigmaRelativeToIdealOriginal >= minScale )
		{
		  // GET DESCRIPTORS FROM THIS SCALE IMAGE
		  // =====================================
		  // Allocate space for the gradient images
		  gil::gray16s_image_t x_grad_image(currImage.dimensions());
		  gil::gray16s_image_t y_grad_image(currImage.dimensions());
		  
		  // Store the x and y gradients of the image
		  float x_difference_operator[] = {1, 0, -1};
		  float y_difference_operator[] = {-1, 0, 1};
		  ///@todo These calls are not optimized for gradient calculation... should write my own.
		  gil::kernel_1d_fixed<float, 3> x_kernel(x_difference_operator, 1);
		  gil::kernel_1d_fixed<float, 3> y_kernel(y_difference_operator, 1);
		  gil::convolve_rows_fixed<gil::gray16s_pixel_t>(gil::view(currImage), x_kernel, view(x_grad_image), gil::convolve_option_extend_constant);
		  gil::convolve_cols_fixed<gil::gray16s_pixel_t>(gil::view(currImage), y_kernel, view(y_grad_image), gil::convolve_option_extend_constant);
		  
		  // Get views into these images
		  gil::gray16s_view_t x_grad = gil::view(x_grad_image);
		  gil::gray16s_view_t y_grad = gil::view(y_grad_image);

		  BinCacheType bins = getCachedBins(x_grad, y_grad, sigma, parameters);

#if defined(MULTI_THREAD)
		  boost::mutex descriptorAccess;

		  int middleRow = bins.shape()[1] / 2;
		  int firstRowSecondHalf = middleRow + 1;
		  
		  boost::function0<void> firstHalf = boost::bind(extractDescriptorsFromRegion_Cached, boost::cref(bins),
								 boost::cref((detail::ThreadParamsCache){2, middleRow,
								       sigma, sigmaRelativeToIdealOriginal, octSize}),
								 boost::cref(parameters), boost::ref(descriptors), boost::ref(descriptorAccess), boost::ref(myRand));
		  boost::function0<void> secondHalf = boost::bind(extractDescriptorsFromRegion_Cached, boost::cref(bins),
								  boost::cref((detail::ThreadParamsCache){firstRowSecondHalf, bins.shape()[1] - 2,
												     sigma, sigmaRelativeToIdealOriginal, octSize}),
								  boost::cref(parameters), boost::ref(descriptors), boost::ref(descriptorAccess), boost::ref(myRand));
		  
		  boost::thread t1(firstHalf);
		  boost::thread t2(secondHalf);
		  
		  t1.join();
		  t2.join();

#else
		  float spacing = sigma * MagFactor;

		  for ( unsigned int y = 2; y < bins.shape()[1]-1; ++y )
		    {
		      for ( unsigned int x = 2; x < bins.shape()[0]-1; ++x )
			{
			  float col = x * spacing;
			  float row = y * spacing;
			  if ( myRand() )
			    {
			      boost::shared_ptr<SiftDescriptorType> potentialDescriptor;
			      getDescriptorFromBins(bins, x, y, sigma, parameters, potentialDescriptor);
			      
			      if (isValidDescriptor(potentialDescriptor))
				{
				  DescriptorType d(col*octSize,row*octSize,sigmaRelativeToIdealOriginal,potentialDescriptor);
				  descriptors.insert( d );
				}
			    }
			}
		    }
#endif
 		  completedOneScale = true;
		}
	      
	      // Blur to next scale in the octave
	      float increaseSigma = sigma * std::sqrt(sigmaRatio * sigmaRatio - 1.0f);
	      gaussianBlur(gil::view(currImage), increaseSigma);
	      sigma *= sigmaRatio;
	    }
        
	  // Downsample
	  gil::gray8_image_t tempImage((currImage.width() + 1) / 2, (currImage.height() + 1) / 2);
	  gil::copy_pixels(gil::subsampled_view(gil::view(currImage), 2, 2), gil::view(tempImage));
	  currImage = tempImage;
	  octSize *= 2;
	  ++octave;
	}      
      return descriptors;
    }
  
    /*! Wrapper function for an image of unsigned 8 bit pixels. This
      function wraps the given raw data into an appropriate GIL view
      and passes the view to the generic denseSiftFeatures function.
    
      @see denseSiftFeatures()
      @param src_pixels a pointer to the start of the raw data
      @param src_row_bytes the number of bytes in each row of the image not necessarily equal to w due to padding (sometimes called the widthstep)
      @param w the width in pixels of the image represented by the src_pixels
      @param h the height in pixels of the image represented by the src_pixels
      @param subregionTopLeftX the top left of the subregion of the full image to extract descriptors from - set to 0 to use entire image
      @param subregionTopLeftY the top left of the subregion of the full image to extract descriptors from - set to 0 to use entire image
      @param subregionWidth the width of the region of interest - set to imageWidth to use entire image
      @param subregionHeight the height of the region of interest - set to imageHeight to use entire image
      @param parameters these dictate the behaviour of the sift extraction
      @param myRand a bernoulli random variable generator for random sampling of sift descriptors
      @return a boost::multi_index DescriptorSet
    */
    inline DescriptorSet ComputeDenseSiftFeaturesGray8( const unsigned char* src_pixels, ptrdiff_t src_row_bytes, int w, int h, int subregionTopLeftX, int subregionTopLeftY, int subregionWidth, int subregionHeight, const Parameters & parameters, VariateGeneratorType & myRand )
    {
      gil::gray8c_view_t src = gil::interleaved_view(w, h, (const gil::gray8_pixel_t*)src_pixels, src_row_bytes);
      gil::gray8c_view_t subimageView = gil::subimage_view(src, subregionTopLeftX, subregionTopLeftY, subregionWidth, subregionHeight);
    
      return denseSiftFeatures(subimageView, parameters, myRand );
    }

#if defined(SJM_PYSIFT)
    class SiftExtractor
    {
    public:
      SiftExtractor(Parameters params);
      ~SiftExtractor();

      void loadImage(std::string imagePath);
      void buildScalePyramid();
      void outputPyramid();
      std::vector<boost::python::list> getDescriptorSet(std::string imagePath);
      boost::python::list getDescriptor(float x, float y, unsigned int scaleIndex);
      boost::python::list scaleRange(float minRadius, float maxRadius);
      
    private:
      Parameters params;
      IplImage* cvImage;
      std::vector<float> scales;
      std::vector<float> octSizes;
      std::vector<gil::gray8_image_t> imagePyramid;
      std::vector<gil::gray16s_image_t> xgradPyramid;
      std::vector<gil::gray16s_image_t> ygradPyramid;
    };
#endif
  } // end SIFT namespace
} // end sjm namespace
