#pragma once

#include <iostream>
#include <string>
#include <cstring>
#include <vector>
#include <algorithm>
#include <utility>
#include <boost/lexical_cast.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_sparse.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/algorithm.hpp>
#include <boost/lambda/if.hpp>
#include <boost/lambda/loops.hpp>
#include <boost/lambda/switch.hpp>
#include <boost/gil/gil_all.hpp>
#include <boost/gil/extension/io/jpeg_io.hpp>
#include <boost/gil/extension/io/jpeg_dynamic_io.hpp>
#include <boost/gil/extension/toolbox/hsv.hpp>
#include "../sift/sift.hpp"
#include "codeword_set.hpp"
#include "../histograms/sparse_histogram.hpp"
#include "../histograms/integral_histogram.hpp"


namespace sjm
{
  /*! Functions and structures implementing Spatial Pyramid Match kernel (Lazebnik et al)
    @author Sancho McCann
  */
  namespace SpatialPyramid
  {
    namespace fs = boost::filesystem;
    namespace ublas = boost::numeric::ublas;
    namespace bl = boost::lambda;
    namespace gil = boost::gil;
    namespace hsv = boost::gil::hsv_color_space;

    typedef sjm::histograms::SparseHistogram<float> HistogramType; ///< The type that histograms will be returned in
    typedef sjm::histograms::SparseHistogram<float> HistogramZeroType; ///< The type that produces a zero histogram
    
    /*!
      This type will hold the spatial pyramid.
      It is a vector of vectors of histograms.
      Each vector represents one level of the spatial pyramid, holding 1, 2, 4, etc HistogramType histograms
      The order of spatial bins in each histogram is in reading order (left-right, top-bottom)
    */
    typedef std::vector<std::vector<HistogramType> > SpatialPyramidBins;
    
    /*!
      A spatial pyramid description of a located codeword set
    */
    struct SpatialPyramidType
    {
      int totalFeatures; ///< Cache of the total number of codewords associated with this pyramid description
      SpatialPyramidBins bins; ///< The bins of the pyramid
    };

    /*!
      Utility function for saving spatial pyramids to file.
     
      @param pyramid a spatial pyramid structure
      @param outputStream an open file output stream
    */
    void save(const SpatialPyramidType & pyramid, fs::ofstream & outputStream)
    {
      int depth = (int)pyramid.bins.size();
      int numChannels = pyramid.bins[0][0].size();
      int totalFeatures = pyramid.totalFeatures;

      outputStream.write((char*)&depth, sizeof(depth));
      outputStream.write((char*)&numChannels, sizeof(numChannels));
      outputStream.write((char*)&totalFeatures, sizeof(totalFeatures));
      for (int level = 0; level < depth; ++level)
	{
	  for (int bin = 0; bin < (int)pyramid.bins[level].size(); ++bin)
	    {
	      int size = pyramid.bins[level][bin].size();
	      outputStream.write((char*)&size, sizeof(size));
	      for (HistogramType::const_iterator it = pyramid.bins[level][bin].begin();
		   it != pyramid.bins[level][bin].end();
		   ++it)
		{
		  int index = it->first;
		  float value = it->second;
		  outputStream.write((char*)&index, sizeof(index));
		  outputStream.write((char*)&value, sizeof(value));
		}
	    }
	  
	}
    }

    /*!
      Utility function for saving spatial pyramids to file.
     
      @param pyramid a spatial pyramid structure
      @param outputPath the path to write to
    */
    void save_(const SpatialPyramidType & pyramid, const std::string & outputPath )
    {
      fs::ofstream pyramidFile(outputPath, std::ios::binary);
      save(pyramid, pyramidFile);
      pyramidFile.close();
    }
    
    /*!
      Utility function for loading spatial pyramids from file.
     
      @param pyramid the destination pyramid structure into which the next pyramid from the inputStream will be read
      @param inputStream an open file input stream
    */
    void load(SpatialPyramidType & pyramid, fs::ifstream & inputStream)
    {
      int depth;
      int numChannels;
      int totalFeatures;
      inputStream.read((char*)&depth, sizeof(depth));
      inputStream.read((char*)&numChannels, sizeof(numChannels));
      inputStream.read((char*)&totalFeatures, sizeof(totalFeatures));

      pyramid.totalFeatures = totalFeatures;
      pyramid.bins.clear();
      for (int level = 0; level < depth; ++level)
	{
	  pyramid.bins.push_back( std::vector<HistogramType>( (1 << level) * (1 << level), HistogramZeroType() ) );
	}
      for (int level = 0; level < depth; ++level)
	{
	  for (int bin = 0; bin < (int)pyramid.bins[level].size(); ++bin)
	    {
	      int size = 0;
	      inputStream.read((char*)&size, sizeof(size));
	      for (int i = 0; i < size; ++i)
		{
		  int index;
		  float value;
		  inputStream.read((char*)&index, sizeof(index));
		  inputStream.read((char*)&value, sizeof(value));
		  
		  pyramid.bins[level][bin][index] = value;
		}
	    }
	}
    }
    
    SpatialPyramidType pyramidFromIntegralHistogram( sjm::histograms::IntegralHistogram & integralHistogram,
						     int tlx, int tly, int width, int height, int pyramid_levels )
    {
      SpatialPyramidType pyramid;
      int totalFeatures = 0;

      for ( int level = 0; level < pyramid_levels; ++level )
	{
	  pyramid.bins.push_back( std::vector<HistogramType>( (1 << level) * (1 << level), HistogramZeroType() ) );
	}
      
      for ( int level = pyramid_levels - 1; level >= 0; level-- )
	{
	  int numDivisions = ( 1 << level );
	  float binWidth = (float)width/numDivisions;
	  float binHeight = (float)height/numDivisions;
	  
	  for ( int x_bin = 0; x_bin < numDivisions; ++x_bin )
	    {
	      int left = tlx + int(x_bin * binWidth + 0.5);
	      int right = tlx + int((x_bin+1) * binWidth + 0.5);
	      for ( int y_bin = 0; y_bin < numDivisions; ++y_bin )
		{
		  int top = tly + int(y_bin * binHeight + 0.5);
		  int bottom = tly + int((y_bin+1) * binHeight + 0.5);
		  
		  boost::numeric::ublas::vector<float> region = sjm::histograms::regionHistogram(integralHistogram, left, top, right, bottom );
		  pyramid.bins[level][y_bin * numDivisions + x_bin] = region;
		  if ( level == 0 )
		    {
		      totalFeatures = int(boost::numeric::ublas::sum(region) + 0.5);
		    }
		}
	    }
	}

      pyramid.totalFeatures = totalFeatures;
      return pyramid;
    }

    SpatialPyramidType colourPyramidFromView( const gil::hsv32f_view_t & hsvView, int tlx, int tly, int width, int height, int pyramid_levels )
    {
      SpatialPyramidType pyramid;
      int totalFeatures = 0;

      for ( int level = 0; level < pyramid_levels; ++ level)
	{
	  pyramid.bins.push_back( std::vector<HistogramType>( (1 << level) * (1 << level), HistogramZeroType() ) );
	}
      
      for ( int level = pyramid_levels - 1; level >= 0; level-- )
	{
	  int numDivisions = ( 1 << level );
	  float binWidth = (float)width / numDivisions;
	  float binHeight = (float)height / numDivisions;
	  
	  for ( int y = std::max(0, tly); y < std::min(tly+height,hsvView.height()); ++y )
	    {
	      gil::hsv32f_view_t::x_iterator hsv_x_it = hsvView.row_begin( y );
	      for ( int x = std::max(0, tlx); x < std::min(tlx+width,hsvView.width()); ++x )
		{
		  int x_bin = (int)( ((float)x-(float)tlx)/(float)binWidth );
		  int y_bin = (int)( ((float)y-(float)tly)/(float)binHeight );
		  
		  float h = gil::get_color(hsv_x_it[x], hsv::hue_t());
		  float s = gil::get_color(hsv_x_it[x], hsv::saturation_t());
		  float v = gil::get_color(hsv_x_it[x], hsv::value_t());
		  
		  if ( s > 0.1 and v > 0.2 )
		    {
		      int sBin = (int)(s * 10);
		      sBin = sBin % 10;
		      int hBin = (int)(h * 10);
		      hBin = hBin % 10;
		      
		      for ( int soff = -1; soff <= 1; ++soff )
			{
			  int s = sBin + soff;
			  if ( s >= 0 && s < 10 ) // if the adjacent bin exists
			    {
			      for ( int hoff = -1; hoff <= 1; ++hoff )
				{
				  int h = (hBin + hoff) % 10;
				  if ( h < 0 )
				    h += 10;
				  int histogramBin = s * 10 + h;
				  pyramid.bins[level][y_bin * numDivisions + x_bin][histogramBin] += 1; // Accumulate histogram
				  if ( level == 0 )
				    ++totalFeatures;
				}
			    }
			}
		    }
		}
	    }
	} // end level iteration
      
      pyramid.totalFeatures = totalFeatures;
      return pyramid;
    }

    /*! Produces a spatial pyramid of colour histograms
     */
    SpatialPyramidType colourPyramid( const std::string & imagePath, int tlx, int tly, int width, int height, int pyramid_levels )
    {
      gil::rgb16_image_t image;
      gil::jpeg_read_and_convert_image(imagePath, image);
      gil::hsv32f_image_t hsvImage( gil::view( image ).dimensions() );
      gil::copy_pixels( gil::color_converted_view<gil::hsv32f_pixel_t>( gil::view( image ))
			, gil::view( hsvImage ));

      gil::hsv32f_view_t hsvView = gil::view(hsvImage);

      SpatialPyramidType pyramid = colourPyramidFromView(hsvView, tlx, tly, width, height, pyramid_levels);
      return pyramid;
    }
    
    /*! Produces a spatial pyramid description of the layout of codeword features as described in Lazebnik, Schmid, and Ponce
      @param topLeftX the top left of the subwindow of the codewprd set to take to make a pyramid with
      @param topLeftY the top left of the subwindow of the codeword set to take to make a pyramid with
      @param width the width of the image associated with this set of codewords - needed for forming the spatial pyramid
      @param height the height of the image associated with this set of codewords - needed for forming the spatial pyramid
      @param pyramid_levels the number of pyramid levels to construct, going from coarse to fine
      @param binary flag to cause building of binary indicator vectors rather than accumulative histograms
      @param histogramSize number of bins in the histogram
      @param codewordSet a collection of located codewords
      @return a SpatialPyramidType
    */
    SpatialPyramidType buildSpatialPyramid(int width, int height, const CodewordSet & codewordSet,
					   int histogramSize, int pyramid_levels, bool binary, int topLeftX = 0, int topLeftY = 0)
    {
      SpatialPyramidType spatialPyramidResult;

      int totalFeatures = 0;
      // The map will be built from the finest requested level down to
      // the coarsest level, so that the histograms from the finer
      // histograms can just be aggreated to make coarser level
      // histograms
      ///@todo: do what the above comment actually says to do, right now, this is pretty inefficient
      for (int level = 0; level < pyramid_levels; ++level)
	{
	  spatialPyramidResult.bins.push_back( std::vector<HistogramType>( (1 << level) * (1 << level), HistogramZeroType() ) );
	}
    
      for (int level = pyramid_levels - 1; level >= 0; level--)
	{
	  int numDivisions = (1 << level);
	  float binWidth = (float)width / numDivisions;
	  float binHeight = (float)height / numDivisions;

	  CodewordSet::const_iterator codewordIterator = codewordSet.begin();
	  for ( ; codewordIterator != codewordSet.end(); ++codewordIterator )
	    {
	      if ( codewordIterator->x >= topLeftX && codewordIterator->x < topLeftX + width &&
		   codewordIterator->y >= topLeftY && codewordIterator->y < topLeftY + height )
		{
		  assert(codewordIterator->id >= 0 && codewordIterator->id < histogramSize);
		  
		  int x_bin = (int) ( (codewordIterator->x - topLeftX) / binWidth);
		  int y_bin = (int) ( (codewordIterator->y - topLeftY) / binHeight);
		  
		  if ( !binary )
		    {
		      spatialPyramidResult.bins[level][y_bin * numDivisions + x_bin][codewordIterator->id] += 1; // Accumulative histogram
		      if ( level == 0 )
			++totalFeatures;
		    }
		  else
		    {
		      // Only count once per codeword in this bin
		      if ( !spatialPyramidResult.bins[level][y_bin * numDivisions + x_bin][codewordIterator->id] )
			{
			  if ( level == 0 )
			    ++totalFeatures;
			  spatialPyramidResult.bins[level][y_bin * numDivisions + x_bin][codewordIterator->id] = 1; // Binary histogram
			}
		    }
		}
	    } // end for all codewords
	} // end for each level
      
      spatialPyramidResult.totalFeatures = totalFeatures;
      return spatialPyramidResult;
    }
    
    /*!
      Print the spatial pyramid to std::cout.
      A utility function for debugging.
     
      @param pyramid the spatial pyramid structure
      @param numLevels the number of levels being used by the spatial pyramid
      @param channel which channel to print out
    */
    void printSpatialPyramid(SpatialPyramidType & pyramid, int numLevels, unsigned int channel)
    {
      for (int level = 0; level < numLevels; ++level)
	{
	  std::cout << "LEVEL " << level << std::endl;
	  for (int ybin = 0; ybin < (1 << level); ++ybin)
	    {
	      for (int xbin = 0; xbin < (1 << level); ++xbin)
		{
		  if ( pyramid.bins[level][ybin * (1<<level) + xbin].find(channel) != pyramid.bins[level][ybin * (1<<level) + xbin].end() )
		    {
		      std::cout << pyramid.bins[level][ybin * (1<<level) + xbin][channel] << "  ";
		    }
		  else
		    {
		      std::cout << 0 << " ";;
		    }
		}
	      std::cout << std::endl;
	    }
	}
    }

    /*!
      The pyramid match kernel.
      Takes advantage of the sparsity of the histograms.

      @param a one spatial pyramid
      @param b another spatial pyramid
      @param pyramid_levels the number of levels in the pyramid to use in the match kernel

      @return float the unnormalized match score
    */
    float fastMatchKernel(const SpatialPyramidType & a, const SpatialPyramidType & b, int pyramid_levels)
    {
      float accumulator = 0;

      float first_level = 0;
      for (int bin = 0; bin < (int)a.bins[0].size(); ++ bin)
	{
	  if ( a.bins[0][bin].size() < b.bins[0][bin].size() )
	    {
	      HistogramType::const_iterator a_it = a.bins[0][bin].begin();
	      for ( ; a_it != a.bins[0][bin].end(); ++a_it )
		{
		  if ( b.bins[0][bin].find(a_it->first) != b.bins[0][bin].end() )
		    {
		      first_level += std::min(a_it->second, b.bins[0][bin].find(a_it->first)->second);
		    }
		}
	    }
	  else
	    {
	      HistogramType::const_iterator b_it = b.bins[0][bin].begin();
	      for ( ; b_it != b.bins[0][bin].end(); ++b_it )
		{
		  if ( a.bins[0][bin].find(b_it->first) != a.bins[0][bin].end() )
		    {
		      first_level += std::min(a.bins[0][bin].find(b_it->first)->second, b_it->second);
		    }
		}
	    }
	}

      accumulator = (1.0f/(1 << (pyramid_levels-1))) * first_level;
      
      for (int level = 1; level < pyramid_levels; ++level)
	{
	  float level_accumulator = 0;
	  for (int bin = 0; bin < (int)a.bins[level].size(); ++bin)
	    {
	      if ( a.bins[level][bin].size() < b.bins[level][bin].size() )
		{
		  HistogramType::const_iterator a_it = a.bins[level][bin].begin();
		  for ( ; a_it != a.bins[level][bin].end(); ++a_it )
		    {
		      if ( b.bins[level][bin].find(a_it->first) != b.bins[level][bin].end() )
			{
			  level_accumulator += std::min(a_it->second, b.bins[level][bin].find(a_it->first)->second);
			}
		    }
		}
	      else
		{
		  HistogramType::const_iterator b_it = b.bins[level][bin].begin();
		  for ( ; b_it != b.bins[level][bin].end(); ++b_it )
		    {
		      if ( a.bins[level][bin].find(b_it->first) != a.bins[level][bin].end() )
			{
			  level_accumulator += std::min(a.bins[level][bin].find(b_it->first)->second, b_it->second);
			}
		    }
		}

	    }
	  accumulator += (1.0f/(1 << (pyramid_levels-1-level+1))) * level_accumulator;
	}

      return accumulator;

    }
    
    /*!
      Computes the normalized spatial pyramid match score between two
      spatial pyramids. Normalization is necessary because large images
      will match better than small images (there are just more features
      in each bin in large images). Normalization is done by dividing
      the score by the square route of the sum of each spatial pyramids
      self-match score. This properly bounds the output of this match
      kernel to [0,1].
     
      @param a A spatial pyramid
      @param b A spatial pyramid
      @param pyramid_levels The number of levels of the pyramid to use

      @return the normalized match score
    */
    float normalizedSpatialPyramidMatchKernel(const SpatialPyramidType & a, const SpatialPyramidType & b, int pyramid_levels)
    {
      float crossSimilarity = fastMatchKernel(a, b, pyramid_levels);

      if ( a.totalFeatures == 0 || b.totalFeatures == 0 )
      {
      	  return 0;
      	}
      else
      {
      	  return crossSimilarity / sqrt((float)(a.totalFeatures) * (float)(b.totalFeatures));
      }
    }
    
    
    namespace util
    {
      /*!  The file path and loaded pyramid representation from that
	path.

	Useful for managing the construction of the svm models where
	it is important to have a canonical ordering of the exemplars.
      */
      struct PathAndPyramid
      {
	fs::path path; ///< The file path of the pyramid
	SpatialPyramid::SpatialPyramidType pyramid; ///< The loaded spatial pyramid
	
	/*!
	  Comparison operator. Simply uses the lexicographical ordering of the path.
	  
	*/
	bool operator < (const PathAndPyramid & other) const
	{
	  return path.string() < other.path.string();
	}
      };
      
    } 
    
  }
}
