#pragma once
#include <string>
#include <vector>
#include <set>
#include <iostream>
#include <boost/algorithm/string.hpp>
#include <boost/function.hpp>
#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

namespace sjm {
  /*!
    Just some utility functions I use a lot in my code
    
    @author Sancho McCann
  */
  namespace util {
    /*!
      Forces a slash at the end of a path. Useful for concatenating strings that are supposed to represent paths.
      
      @param directoryName a string to enforce the slash on
    */
    inline void enforceDirectorySlash(std::string & directoryName)
    {
      if ( directoryName.substr(directoryName.size() - 1) != "/" )
	{
	  directoryName = directoryName + "/";
	}
    }
    
    /*!
      Checks whether or not a given extension is in a list of extensions
      
      @param extension the exension query
      @param extensionList the list of valid extensions
      @return bool returns true if the given extension is valid
    */
    inline bool extensionCheck(const std::string & extension, const std::set<std::string> & extensionList)
    {
      return extensionList.find(boost::algorithm::to_lower_copy(extension)) != extensionList.end();
    }

    /*!
      Applies a function that operates on a file to the given file.

      @param func the void function object that takes a file path as a parameter
      @param filePath a file path
    */
    inline void applyFunctionToFile(boost::function1<void, const boost::filesystem::path&> func, const boost::filesystem::path & filePath)
    {
      func(filePath);
    }

    /*!  Takes a function that operates on a file and applies it to
      any files in a given path hierarchy with valid extensions.

      @param func the void function object that takes a file path as a parameter
      @param filePath a file path
      @param validExtensions the set of extensions that func can operate on
      @param recursive flag to cause recursive call of this function with any directories found in filePath
    */
    inline void recursiveFunctionApplication(boost::function1<void, const boost::filesystem::path&> func,
					     const boost::filesystem::path & filePath,
					     const std::set<std::string> & validExtensions,
					     bool recursive)
    {
      if ( boost::filesystem::exists(filePath) )
	{
	  if ( boost::filesystem::is_directory(filePath) )
	    {
	      if ( recursive )
		{
		  boost::filesystem::directory_iterator dit(filePath);
		  boost::filesystem::directory_iterator end;

		  for ( ; dit != end; ++dit )
		    {
		      recursiveFunctionApplication(func, *dit, validExtensions, recursive);
		    }
		}
	    }
	  else if ( extensionCheck(boost::filesystem::extension(filePath), validExtensions) )
	    {
	      applyFunctionToFile(func, filePath);
	    } // end if-else directory/file check
	} // end if file exists
    } // end recursive function application


    /*!  Reads the indices of support vectors found in the model file
      produced by libsvm.

      @param svmModelFilename libsvm's model file output
      @return std::vector<int> the collection of support vector indices
    */
    std::vector<int> readSupportVectors( std::string svmModelFilename )
    {
      std::vector<int> supportVectors;
      boost::filesystem::ifstream svmModelFile(svmModelFilename);
      char line[512];
      svmModelFile.getline(line, 512); // svm_type
      svmModelFile.getline(line, 512); // kernel_type
      svmModelFile.getline(line, 512); // nr_class
      svmModelFile.getline(line, 512); // total_sv
      std::string totalSV(line);
      std::vector<std::string> tokens;
      boost::algorithm::split(tokens, totalSV, boost::algorithm::is_any_of(" "));
      int numSupportVectors = boost::lexical_cast<int>(tokens[1]);
      svmModelFile.getline(line, 512); // rho
      svmModelFile.getline(line, 512); // label
      svmModelFile.getline(line, 512); // nr_sv
      svmModelFile.getline(line, 512); // SV
      for ( int i = 0; i < numSupportVectors; ++i )
	{
	  svmModelFile.getline(line, 512);
	  std::string supportVector(line);
	  boost::algorithm::trim(supportVector);
	  boost::algorithm::split(tokens, supportVector, boost::algorithm::is_any_of(":"));
	  int supportVectorId = boost::lexical_cast<int>(tokens[1]) - 1; // stored in file as 1-based... changing to zero-based
	  supportVectors.push_back(supportVectorId);
	}
      svmModelFile.close();
        
      return supportVectors;
    }
    
    namespace selection {
      /*!
	Creates a list of indices ordered by their feature weighting.

	@param featureWeights a vector of weights for the features
	@param selectedCodewords the destination of the selection results
	@param numToSelect how many features to select
      */
      void chooseTopFeatures( const std::vector<float> & featureWeights, std::set<int> & selectedCodewords, int numToSelect )
      {
	std::vector<bool> choosen(featureWeights.size(), false); // keeps track of features already chosen
	for ( int selected = 0; selected < numToSelect; selected++ )
	  {
	    int bestFeature = -1;
	    float bestWeight = 0;
	    for ( int i = 0; i < (int) featureWeights.size(); i++ )
	      {
		if ( featureWeights[i] > bestWeight && !choosen[i] )
		  {
		    bestWeight = featureWeights[i];
		    bestFeature = i;
		  }
	      }
	    assert(bestFeature != -1);
	    choosen[bestFeature] = true;
	    selectedCodewords.insert(bestFeature);
	  }
  
      }
    } // end namespace selection
    
  } // end namespace util
} // end namespace sjm
