%%% All robot initialization that must be performed after reading
%%% configuration files is done from this script.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% save old images and create a new images directory
%! util/rmri.bash


% Connect to robot and get PTU parameters
if configPars.usingRobot
    powerbot('connect',configPars.robotName,7001);
end;

if configPars.havePTU
  ptu_lim = powerbot('getlim');
  ptu_res = powerbot('getres');
  ptu_lim(1:2) = [-90 90];  % Constrain range of ptu_lim to -90..90
  cameraPars.ptu_lim = ptu_lim;
  state = powerbot('getpos');
  pan = state(1);
  tilt = state(2);
  fprintf('[InitializeRobot.m]: PTU Setup: pan=%d tilt=%d\n',pan,tilt);
else
    %configPars.exploreOnly = true;
end;

if configPars.usingLaptop && ~configPars.exploreOnly
  
  if cameraPars.bbfl
    camera('connect',1);
    camera('whitebalance',51,30);
    camera('setup_capture',1);
    camera('framerate',15);
  end;
  
  if cameraPars.canonfl
      cameraCapture('start');
      cameraCapture('setsize',cameraPars.res_name{cameraPars.canon_res});
  end;
  
end;

if configPars.usingRobot
    % Get initial odometry
    odom = powerbot('getodom');
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Perform SLAM initialization
fprintf('[InitializeRobot.m]: Initializing SLAM mapping and object list...\n');

if ~configPars.usingRobot
    odom = [ 0 0 0];
end;

performSlam('start',odom);
oGrid = performSlam('get_map');
mappingPars.logFile = fopen('map.log', 'wt');
mappingPars.curObjId = 1;
mappingPars.mapId = 1;
mappingPars.objs = [];
mappingPars.coverageMap = zeros(size(oGrid),'uint8');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Perform planner initialization
fprintf('[InitializeRobot.m]: Initializing planner...\n');
planningPars.explored = zeros(size(oGrid)); % Binary map of explored frontier points
planningPars.OBSTACLE_THRESHOLD = 0.6;
planningPars.YAW_THRESH = 10 * pi / 180;
planningPars.MAX_COST = 16384;
planningPars.exploreDone = false;
planningPars.minViewDist = 14;
planningPars.maxViewDist = 35;

% This parameter determines the most seconds that the robot will spend trying to
% achieve a goal before it aborts the goal and replans.
planningPars.maxGoalTime = 99999;%20;

% These sequence numbers will index every worldState.
planningPars.sequenceNum = 0;

% When moving along a path, how many points should we 'read ahead' to check
% where we are along the path?
planningPars.lookAhead = 12;

% When sampling over potential goals, how large should our batch size be?
planningPars.batchSize = 20;

% When sampling over potential goals, how many epochs should we run before we
% decide to quit, at most? (Convergence may allow us to sample less than this)
planningPars.maxEpochs = 100;

% How much more important is exploring a frontier square as opposed to gaining a
% full coverage shot? Ie: how many time units would I have to spend on a full
% coverage shot before I preferred spending one time unit exploring the frontier?
planningPars.frontierImportance = 8;

% How much more important is taking a photograph of an object as opposed to
% gaining a full coverage shot? Ie: how many time units would I have to spend on
% coverage before I preferred spending one time unit taking a photograph?
planningPars.photographImportance = 3;

% How much less valuable is a coverage shot of the same location? Ie: for a
% given pixel/location that has been covered once, by what factor should its
% contribution to scoring decrease? ie: x.^(-numShots)
planningPars.coverageScoreDegradation = 1.15;

% How much time, in seconds, to spend on goal planning before returning control
% to the calling function. See sampleFromGoalSpace.m.
planningPars.sampleTimeLimit = 2.0;

% Dilate/Erode the oGrid by how much in order to compute the obstacle_map?
% This parameter should tailor robot size.
planningPars.dilationRadius = 6;

% How many grid cells should we insert between waypoints?
planningPars.samplingDiscretization = 3;

% Should we perform an initial 'swivel' turn before proceeding to a goal?
planningPars.performInitialTurns = false; 

% How far ahead does the robot prefer to set its intermediate waypoints? 
% This distance is in metres. Low values will cause the stop-start behaviour
% where matlab finds out "late" that we arrived at a waypoint. Higher values
% will cause the robot to rely on vfh more often.   
planningPars.favoredWaypointDistance = 1.5 

if configPars.experiment > 2
  
  % Load training data
  fprintf('[InitializeRobot.m]: Loading webtrain.mat...\n');
  load('/home/lciuser/turntable_data_selected/webtrain.mat');
  planningPars.trainList = fileList;
  
  % Define BoF training images to be the first 4 available training images
  for k=1:size(fileList,2),
    planningPars.trainList(k).bagImages=1:size(fileList(k).list,2);
  end
  
  % List of [done err objId topScore objId2 2ndScore] for each class of object we're looking for
  planningPars.lookBackList = zeros(length(fileList),8);
  planningPars.lookBackList(:,3:8) = -1;
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialize runtime variables 
goal = [];
configPars.numZoomedPics = 0;
configPars.planPicNum = 0;
configPars.nextDisplayTime = toc;
configPars.benchmarkTime = 0;
if configPars.timeLimit > 0
    planningPars.timeLimit = toc + configPars.timeLimit;
end;

planningPars.hasGoal = false;
planningPars.numGoals = 0;
planningPars.numSuccesses = 0;

if configPars.savePlanMovie
    configPars.planMovie = avifile('data/planMovie.avi');
    configPars.planMovie.FPS = 2;
end;

% Do some initial slam from where we start!
fprintf('[InitializeRobot.m]: Performing initial SLAM scans.\n');
doSlam(5,true);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Program constants
global STATE_NEW;
global STATE_INITIAL_TURN_START;
global STATE_INITIAL_TURN;
global STATE_TRAVERSAL;
global STATE_FINAL_TURN_START;
global STATE_FINAL_TURN;
global STATE_PHOTOGRAPH;
global STATE_FINISHED;
global STATE_ABORTED;

STATE_NEW = 1;
STATE_INITIAL_TURN = 2;
STATE_TRAVERSAL = 3;
STATE_FINAL_TURN = 4;
STATE_PHOTOGRAPH = 5;
STATE_FINISHED = 6;
STATE_ABORTED = 7;
STATE_INITIAL_TURN_START = 8;
STATE_FINAL_TURN_START = 9;

planningPars.BEHAVIOUR_ONE = 1;
planningPars.BEHAVIOUR_TWO = 2;
planningPars.BEHAVIOUR_THREE = 3;

planningPars.BEHAVIOUR_TIME_ONE = configPars.timeLimit;% 12*60;
planningPars.BEHAVIOUR_TIME_TWO = 30*60;
planningPars.BEHAVIOUR_TIME_THREE = configPars.timeLimit;

planningPars.behaviourDesc{1} = 'BEHAVIOUR_ONE';
planningPars.behaviourDesc{2} = 'BEHAVIOUR_TWO';
planningPars.behaviourDesc{3} = 'BEHAVIOUR_THREE';

planningPars.currentBehaviour = planningPars.BEHAVIOUR_ONE;

planningPars.hasGoal = false;
planningPars.numGoals = 0;
planningPars.numSuccesses = 0;
%planningPars.visitedTables = [];
planningPars.visitedTableRegions = [];
%planningPars.lastTable = [];
planningPars.lastTableRegion = [];
planningPars.TABLE_DISTANCE = 5; %How far away you need to look from a previously visited point on a table

if configPars.savePlanMovie
    configPars.planMovie = avifile('data/planMovie.avi');
    configPars.planMovie.FPS = 2;
end;
