% RunRobot.m
%
% This script is intended to provide a high-level driver for planner
% experiments on Curious George. It can also be modified to provide the
% competition runtime behaviour for the robot.
% 
% The motivation for this code is to provide a testing and operations
% framework that is more flexible than the SRVC_Phase2 script.
% 

% Puts all directories into command path.
LoadPath;

% Disconnect connections, clear the command window.
clc;
clear variables;
fprintf('\n\nCurious George 2008 starting up...\n\n');

% Start the timer.
tic;
global configPars;

% Configure the robot by evaluating configuration files into the current
% workspace. This includes:
% Reading the config parameters. (configPars)
% Reading the camera parameters. (cameraPars)
% Reading the gaze parameters. (gazePars)
ConfigureRobot;

% Reboot all possible servers
killall;

% Perform other initialization:
% Connect to powerbot. (powerbot('connect') etc.)
% Initialize SLAM mapping. (mappingPars)
% Initialize planning structures. (planningPars)
InitializeRobot;

% Do some random walking to bootstrap our SLAM module.
% The wheels have to be moving during this process, so here goes!
for i=1:configPars.numWalks
    [mappingPars, configPars ] = doRandomWalk2(true,mappingPars,planningPars,cameraPars,configPars,gazePars,1);
end;
%
% Main behaviour loop:
% 
% while termination criteria are not met, repeat the following:
%
%   - get odometry, speed, map, and updated objects.
%       - compose these variables into one object, called our 'worldState'.
%
%   - if there is no plan in effect, ask the planner to generate one.
%
%   - call executeStep on the current plan, passing in the worldState.
%
%   - if this function does not take long, we can spend some time on extra
%   tasks like planning ahead.
%
termination = 0;
%planningPars.currentBehaviour = 3;
while ~termination
    
    fprintf(1,'Current behaviour is %s.\n', planningPars.behaviourDesc{planningPars.currentBehaviour} );
    
    if planningPars.currentBehaviour == 1 && toc > planningPars.BEHAVIOUR_TIME_ONE
        
        planningPars.currentBehaviour = 2;
      
    end;
   
    if planningPars.currentBehaviour == 2 && toc > planningPars.BEHAVIOUR_TIME_TWO
        
        planningPars.currentBehaviour = 3;
        
    end;
    
    % Bump the sequence number
    planningPars.sequenceNum = planningPars.sequenceNum + 1;
    
    if configPars.benchmarkTime == 0
        configPars.benchmarkTime = toc;
    end;
    
    if configPars.PRINTLEVEL >= 4 && mod(planningPars.sequenceNum,10000)==0
        timenow = toc;
        timediff = timenow - configPars.benchmarkTime;
        if timediff > 0
            framerate = 10000 / timediff;
            fprintf('[RunRobot.m]: FPS = %.2f\n', framerate);
        end;
        configPars.benchmarkTime = timenow;
    end;
    
    % Do we have a current goal? 
    if (isempty(goal)) 
        
        worldState = generateWorldState ( cameraPars, gazePars, mappingPars, configPars, planningPars );
        
%         if ~configPars.exploreOnly
%             [configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars);
%         end;
%   

        if planningPars.currentBehaviour == planningPars.BEHAVIOUR_ONE
            
            [ goal planningPars ] = generateGoalTable( worldState, cameraPars, configPars, mappingPars, planningPars );
            
            if goal.type == 'x'
                [mappingPars, configPars ] = doRandomWalkBB(toc+60,mappingPars,planningPars,cameraPars,configPars,gazePars,0);
            end;
                        
        elseif planningPars.currentBehaviour == planningPars.BEHAVIOUR_TWO
            
            [mappingPars, configPars ] = doRandomWalkBB(planningPars.BEHAVIOUR_TIME_TWO,mappingPars,planningPars,cameraPars,configPars,gazePars,1);
            
            
        elseif planningPars.currentBehaviour == planningPars.BEHAVIOUR_TWO || planningPars.currentBehaviour == planningPars.BEHAVIOUR_THREE
            
            % Nothing here for now
            [mappingPars, configPars ] = doRandomWalkBB(planningPars.BEHAVIOUR_TIME_THREE,mappingPars,planningPars,cameraPars,configPars,gazePars,1);
            
        end;
        
        % If there is no such goal, we have to generate one from our current state.
        %[ goal planningPars ] = generateGoal ( worldState, cameraPars, configPars, mappingPars, planningPars );
        if ( isempty(goal) )
            fprintf('[RunRobot.m]: No goal available.\n');
            break;
        elseif goal.type == 'x'
            if configPars.PRINTLEVEL >= 2
                fprintf('[RunRobot.m]: goal.type == ''x''. Walking randomly.\n');
            end;
		fprintf('[RunRobot.m] Executing the random walk.\n');
            [mappingPars, configPars ] = doRandomWalk2(true,mappingPars,planningPars,cameraPars,configPars,gazePars,0);
		fprintf('[RunRobot.m] Executing exploreLocation.\n');
            [configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars);
		fprintf('[RunRobot.m] Done exploreLocation.\n');
            
            goal.state = STATE_FINISHED ;
        else
            planningPars.hasGoal = true;
        end;
    else
        UpdateWorldState;
    end;
    
    % Render visual output to window and disk.
    DisplayFrame;
    
    % Once a long-term goal has been established, this function will execute on that plan.
    if goal.type ~= 'x'
        [ worldState goal planningPars ] = executeFrame ( worldState, goal, configPars, planningPars, mappingPars, gazePars, cameraPars );
    end;
    
    % If the goal is finished, say so here.
    if goal.state == STATE_FINISHED
        if configPars.PRINTLEVEL >= 2
            fprintf('[RunRobot.m]: Goal successful.\n');
        end;
        planningPars.numSuccesses = planningPars.numSuccesses + 1;
        planningPars.hasGoal = false;
        goal = [];
    elseif goal.state == STATE_ABORTED
        if configPars.PRINTLEVEL >= 2
            fprintf('[RunRobot.m]: Goal aborted.\n');
        end;
        planningPars.hasGoal = false;
        goal = [];
        
%         if rand > 0.5
%             [configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars);
%         end;
    end;
    
    % Finally, check if our termination criteria are met.
    termination = checkTerminationCriteria ( planningPars, configPars );
    
end;

% Perform some tasks before ending the script.
FinalizeRobot;

