function [ worldState goal planningPars ] = executeFrame ( worldState, goal, configPars, planningPars, mappingPars, gazePars, cameraPars )
% [ worldState goal planningPars ] = executeFrame ( worldState, goal )
%   
%   Executes one frame of activity. A frame consists of:
%   
% worldState := A set of collected data about the world and the robot that is
%   collected/read from sensors every frame. This structure may also contain
%   information persisted from frame to frame as an internal representation of
%   the robot's world.
%   
% goal := A set of information determining the robot's frame-by-frame medium-term
%   plans. Specifically, contains information about future waypoints to visit.
%   If this method returns goal == [], then the robot has accomplished its goal
%   and is ready to replan.
%   

global STATE_NEW;               
global STATE_INITIAL_TURN;      
global STATE_INITIAL_TURN_START;
global STATE_TRAVERSAL;         
global STATE_FINAL_TURN;        
global STATE_FINAL_TURN_START;
global STATE_PHOTOGRAPH;        
global STATE_FINISHED;          
global STATE_ABORTED;        

stateDesc{1} = 'STATE_NEW';
stateDesc{2} = 'STATE_INITIAL_TURN';      
stateDesc{3} = 'STATE_TRAVERSAL';         
stateDesc{4} = 'STATE_FINAL_TURN';        
stateDesc{5} = 'STATE_PHOTOGRAPH';        
stateDesc{6} = 'STATE_FINISHED';          
stateDesc{7} = 'STATE_ABORTED';
stateDesc{8} = 'STATE_INITIAL_TURN_START';
stateDesc{9} = 'STATE_FINAL_TURN_START';

fprintf('[executeFrame.m] The current state is %s \n',stateDesc{goal.state});

% Read in state:
path = goal.path;
pathLength = size(path,1);

% Check that we haven't spent too long on this goal.
if configPars.PRINTLEVEL >= 2 && toc > goal.timeLimit
    fprintf('[executeFrame.m]: Spent %.2f seconds on a goal and could not attain it.\n', goal.timeLimit-goal.startTime);
    [worldState.odom worldState.rPose worldState.rGridPos] = stopBot('blocking', 'doSlam');
    
    goal.state = STATE_ABORTED;
    planningPars.hasGoal = false;
    
    if rand > 0.3
        [configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars);
    end;
    return;
end;

% If we're at the first waypoint, we can perform a spin-turn to arrive at
% the next. goal.step==0 signifies that this is the first time the goal
% has been executed on.
if goal.state == STATE_NEW
    
    initialYaw = goal.initialYaw;
    
    % First turn the robot in place to the correct direction
    yawDelta = thetapp(initialYaw-worldState.rPose(3));
    if abs(yawDelta) > planningPars.YAW_THRESH && planningPars.performInitialTurns
        
        if configPars.PRINTLEVEL >= 3
            fprintf('[executeFrame.m]: Starting swivel toward (initial_yaw=%.2f) (delta=%.2f)\n', goal.initialYaw, yawDelta);
        end;
        
        % Non-blocking turnbot call
        [ goal.desiredYawOdom worldState.odom worldState.rPose worldState.rGridPos ] = turnBot(yawDelta,'slam');
        goal.state = STATE_INITIAL_TURN;
        
    else
        if configPars.PRINTLEVEL >= 3
            fprintf('[executeFrame.m]: Moving to new goal.\n');
        end;
        goal.state = STATE_TRAVERSAL;
    end;
    
end;

% If we're in an 'f'-state or an 'i'-state turn, check if this turn is finished.
if goal.state == STATE_FINAL_TURN || goal.state == STATE_INITIAL_TURN
    diffAngle = abs(thetapp(goal.desiredYawOdom-worldState.odom(3)));
    if diffAngle < 0.1 || (diffAngle < 0.25 && worldState.speeds(3) < 0.1)
        if configPars.PRINTLEVEL >= 3
            fprintf('[executeFrame.m]: Finished swivel turn.\n');
        end;
        
        % Perform a hard stop and slam filtering immediately after a turn:
        [worldState.desiredYawOdom worldState.rPose, worldState.rGridPos] = stopBot('blocking','doSlam');
        if (goal.state == STATE_INITIAL_TURN)
            % Do a couple extra slam calls here so that we have a really awesome map
            % before we start moving:
            [worldState.odom worldState.rPose worldState.rGridPos] = doSlam(3,1);
            goal.state = STATE_TRAVERSAL;
        else
            % This was a finalizing swivel so let's check if we need to take a photo:
            if configPars.exploreOnly
                goal.state = STATE_FINISHED;
                return;
            else
                goal.state = STATE_PHOTOGRAPH;
            end;
        end;
        
    else
        if configPars.PRINTLEVEL >= 5
            fprintf('[executeFrame.m]: Swiveling to %.2frad. Current odom angle = %.2frad. diff = %.2f.\n', goal.desiredYawOdom, worldState.odom(3), diffAngle);
        end;
    end;
end;

% Consider the first step:step+N waypoints in the remaining path:
if goal.state == STATE_TRAVERSAL
    
    leastStepScore = inf;
    for step = max(goal.step,1) : min(goal.step+planningPars.lookAhead,pathLength)
        
        % Convert this waypoint into polar parameters:
        nxtMapCell = path(step,:);
        nxtMapPos = performSlam('get_vecpos_from_gridpos',nxtMapCell);
        vecDelta = nxtMapPos - worldState.rPose(1:2);
        yawDelta = atan2(vecDelta(2),vecDelta(1));
        yawRel = thetapp(yawDelta-worldState.rPose(3));
        distDelta = norm(vecDelta);
        
        % Determine which step to proceed to by scoring each candidate step in the
        % path by considering distance only. Score each candidate step by
        % distance, but penalize steps that are behind the robot and right next to it.
        % stepScore = norm(dirVec) <-- naive distance-based approach
        if distDelta > planningPars.favoredWaypointDistance
            stepScore = distDelta + 2.^(1-cos(yawRel));
        else
            stepScore = 2 * -distDelta + 2.^(1-cos(yawRel));
        end;
        
        if stepScore <= leastStepScore
            leastStepScore = stepScore;
            leastStep = step;
        end;
        
    end;
    
    % Check if we've made progress along the path:
    step = leastStep;
    
    % Recover the polar parameters for this step:
    nxtMapCell = path(step,:);
    nxtMapPos = performSlam('get_vecpos_from_gridpos',nxtMapCell);
    vecDelta = nxtMapPos - worldState.rPose(1:2);
    yawDelta = atan2(vecDelta(2),vecDelta(1));
    distDelta = norm(vecDelta);
    
    if goal.step ~= step
        stepDifference = step - goal.step;
        
        % Calculate desired yaw at the destination:
%         if pathLength >= 2 && step < pathLength
%             nxtMapCell2 = path(step+1,:);
%             nxtMapPos2 = performSlam('get_vecpos_from_gridpos',nxtMapCell2);
%             vecDelta2 = nxtMapPos2 - worldState.rPose(1:2);
%             yawDesired = atan2(vecDelta(2)+vecDelta2(2),vecDelta(1)+vecDelta2(1));
%         elseif goal.type == 'o'
%             yawDesired = goal.finalYaw;
%         else
%             yawDesired = atan2(vecDelta(2),vecDelta(1));
%         end;

        yawDesired = yawDelta;
        
        % Issue a non-blocking command advancing to the next waypoint:
        [ vecDesiredPoseOdom ] = moveBot(worldState.odom, worldState.rPose, worldState.rGridPos, yawDelta, distDelta, yawDesired);
        goal.vecDesiredPoseOdom = vecDesiredPoseOdom;
        
        % Revise eta
        goal.timeLimit = toc + 60; %8 + 2 * (size(goal.path,1) - goal.step);
        if configPars.PRINTLEVEL >= 3
            fprintf('[executeFrame.m]: Advancing %d steps to step %d/%d. Time:%.1fs ETA:%.1fs\n', stepDifference, step, length(goal.path), toc, goal.timeLimit);
        end;
        
        % update goal-state variables:
        goal.step = step;
    end;
    
    % Compare our current distance with our best distance to waypoint.
    % If we've drifted too far away from our goal, then we need to abort and replan.
    if distDelta < goal.bestDistRemain
        goal.bestDistRemain = distDelta;
        goal.timeLimit = toc + 20;
        
    elseif distDelta >= goal.bestDistRemain * 2 + 0.75;
        if configPars.PRINTLEVEL >= 2
            fprintf('[executeFrame.m]: Navigation got to within %.2fm of waypoint but is now at %.2fm. Aborting goal.\n', goal.bestDistRemain, distDelta);
        end;
        
        % Stop the robot
        [worldState.odom worldState.rPose worldState.rGridPos] = stopBot('blocking');
        
        % Take some pictures
        [configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars);
          
        goal.state = STATE_ABORTED;
        return;
        
    end;
    
    % Use odometry target returned from moveBot to check if we should
    % increment our goal step.
    if norm(goal.vecDesiredPoseOdom(1:2)-worldState.odom(1:2)) < 0.35
        goal.step = min(goal.step+1, pathLength);
    end;
    
    % Check if we've actually reached the terminal waypoint of the path:
    arrivedAtTerminus = (goal.step == pathLength) && (norm(goal.vecDesiredPoseOdom(1:2)-worldState.odom(1:2)) < 0.5);
    
    % Check if we're pointing in the right direction:
    if arrivedAtTerminus;
        if configPars.PRINTLEVEL >= 2
            if (goal.type == 'e')
                destinationDesc = 'frontier goal';
            elseif (goal.type == 'o')
                destinationDesc = 'observation goal';
            elseif (goal.type == 'x')
                destinationDesc = 'unspecified goal';
            elseif (goal.type == 't')
                destinationDesc = 'table';
                
                diff = performSlam('get_vecpos_from_gridpos',goal.obsLoc) - worldState.rPose(1:2);
                goal.finalYaw = thetapp(atan2(diff(2),diff(1)));
                
                goal.type = 'o';
            end;
            fprintf('[executeFrame.m]: Arrived at %s!\n', destinationDesc);
        end;
        
        % Perform a blocking stop here.
        [worldState.odom worldState.rPose worldState.rGridPos] = stopBot('blocking','doSlam');
        
        if goal.type == 'o'
            
            % Do we need to turn toward an object?
            if (goal.finalYaw-worldState.rPose(3) > 10 * pi / 180)
                deltaYaw = thetapp(goal.finalYaw - worldState.rPose(3));
                if configPars.PRINTLEVEL >= 3
                    fprintf('[executeFrame.m]: Starting finalizing swivel toward (finalYaw=%.2f) (delta=%.2f)\n', goal.finalYaw, deltaYaw);
                end;
                
                % Turn robot to face the object. No slam needed since we
                % performed a stopBot() with slam above.
                [ goal.desiredYawOdom worldState.odom worldState.rPose worldState.rGridPos ] = turnBot(deltaYaw,'slam');
                %goal.desiredYawOdom = turnBot(deltaYaw,'noslam',worldState.odom,worldState.rPose,worldState.rGridPos);
                goal.state = STATE_FINAL_TURN;
                goal.timeLimit = toc + 30;
	    else
                goal.state = STATE_PHOTOGRAPH;
            end;
            
        elseif goal.type == 'e'
            goal.state = STATE_PHOTOGRAPH;
        end;
        
    end;
    
end;

% Look at interesting objects
if goal.state==STATE_PHOTOGRAPH 
    planningPars.visitedTableRegions = [planningPars.visitedTableRegions ; planningPars.lastTableRegion];
    %planningPars.visitedTables = [planningPars.visitedTables ; planningPars.lastTable];
    fprintf('[executeFrame.m]: Taking pic!\n');
    [ configPars mappingPars planningPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars,true,false);
    goal.state = STATE_FINISHED;
    return;
end;

