%
% This script controls the high level behaviours of the robot during the
% image collection phase of the SRVC
%
% David Meger: May 2, 2007
%
%

% Initialization section...

clear all;
%close all;
clc;

imageCounter = 1;
K = 4;   % Number of interest points to look at

addpath ../../resources/PEImageGrabbing/
addpath ../stereoCode/
addpath ../main/
addpath ../matlab_powerbot/
addpath ../gaze/
addpath ../mapping/pmap

SALIENCY_TOOLBOX = '../../resources/SaliencyToolbox';
addpath(genpath(SALIENCY_TOOLBOX));

configFile = '../stereoCode/bb_id5020143.config';

usingLaptop=0; % Are we running on the laptop?

if usingLaptop
  output_path='/home/lciuser/DRI2/robotImages/';
  output_path2='/home/lciuser/DRI2/robotImages_bbox/';
else
  output_path='/lci/project/raid1/srvc/DRI1/robotImages/';
  output_path2='/lci/project/raid1/srvc/DRI1/robotImages_bbox/';
end;

DISPLAY = 1;
bpattern=0;  % For BB-HICOL-40
canonfl=0;   % Use Canon camera if set, otherwise use Bumblebee

% Available resolutions for Canon camera
res_name=cell(4,1);res=zeros(4,2);
res(1,:)=[3648 2736]; res_name{1}='large';
res(2,:)=[2816 2112]; res_name{2}='medium1';
res(3,:)=[2272 1704]; res_name{3}='medium2';
res(4,:)=[640 480];   res_name{4}='small';
bbtocan_scaling=3.4*[1.2955 1 0.8068 0.2273];  % Scaling from bb to
                                               % given Canon resolution

canon_res=3;          % Use this resolution on Canon camera

% Zoom scalings for Canon camera
% Scalings from zoom 0 to other zoom levels [0:6]
zoom_scaling=[1 1.09 1.228 1.3065 1.44 1.764 2.08];

% Load predictive map and define memory struct
%load gaze_proto_hires338.mat  % 16000 prototypes from 338 saccades
load gaze_proto_hires600rect.mat  % 16000 prototypes from 600 saccades
memory.Ap=Ap;
memory.Up=Up;
memory.Wa=Wa;
memory.Wu=Wu;

% Set parameters for predictive map
pars.itype = 4;
pars.Nn    = 40;
pars.gstd  = 1;

% Define centre marker
xx=[-1 1 0 0 0];
yy=[0 0 0 -1 1];
msz=20;

% Using 16bit or 8bit disparity?
is16BitDisp = 0;

% Connect to robot and get PTU parameters
powerbot('connect','bart.ubcrobots.net',7001);

ptu_lim=powerbot('getlim');
ptu_res=powerbot('getres');

ptu_lim(1:2)=[-90 90];  % Constrain range

state=powerbot('getpos');
pan = state(1);
tilt = state(2);
fprintf('pan=%d tilt=%d\n',pan,tilt);

if usingLaptop
  camera('connect',1);
  camera('setup_capture',1);
  camera('framerate',15);
  if canonfl
    cameraCapture('start');
    cameraCapture('setsize',res_name{canon_res});
  end;
end;

% Mapping
logFile = fopen('map.log', 'wt');
crStr = sprintf('\n');
odom = powerbot('getodom');
objId = 1;
objs = [0 0 0 0];

if canonfl
  CAMERAHEIGHT = 1.5;
else
  CAMERAHEIGHT = 1.45;
end;

%log odometry
odomStr = sprintf('o %g %g %g\n',odom);
fwrite(logFile,odomStr,'char');

performSlam('start',odom);

% Now enter the main behaviour loop. 
%
% reapeat forever...
%
%   (1) grab bumblebee image and pre-process 
%   (2) find K interesting regions
%
%   for region =  1:K
%       
%       (3) centre region in the Canon camera
%       (4) zoom the camera so that the region is somewhat tightly cropped
%       (5) snap an image and save it
%   end;
%
%   (6) move the robot randomly to a new position and direction
%
% end;

tic
for i=1:20

  if i==1

    % Look down
    tilt0=-35;40; % Look 40deg down
    state0=powerbot('setpos',0,tilt0);
    tilt0=state0(2);

    [img,disp,objectImage] = getImages(tilt0,configFile,usingLaptop,is16BitDisp);         % (1)
    
    % Don't look too far down
    angle_left=tilt0-ptu_lim(3);
    objectImage = pruneObjectImage(objectImage,angle_left);

    if DISPLAY,
        figure(1);clf;
        disp8=disp;disp8(disp8>251)=0;disp8=8*disp8;
        image([ rgb2gray(img) disp8; objectImage zeros(size(objectImage),'uint8')]);
        colormap(gray(256));
        drawnow;
    end;
    
    [ objects ] = getObjects( img, objectImage,K );     % (2)
    
    if DISPLAY
        
        figure(2);clf;
        image(img);
        hold on;
        
        for obj = 1:length(objects)
            
          %plot(objects(obj).centre(2), objects(obj).centre(1),'r+');
          h=plot_objnum(objects);
          set(h,'Color',[1 0 0]);
          bbox = objects(obj).bb;
          h=plot_bbox(bbox,'r-');        
        end;
        
        hold off;
    end;

    objId = addObjectsToMap(tilt0,disp,objects,objId,configFile,CAMERAHEIGHT,is16BitDisp,logFile);

  end; % if i==1

    objLocs = performSlam('get_objects');

    if DISPLAY
      figure(11);
      hold on;
      odom = performSlam('get_odom');
      yaw = odom(3);
      plot(odom(1),odom(2),'g+');
      %Matrix to rotate 2d coordinates from robot to global frame
      R_r = [cos(yaw), -sin(yaw); sin(yaw), cos(yaw)];

      objsSize = size(objLocs);
      for j=1:objsSize(1)
        if j < 5000
          xy = R_r*[objLocs(j,1) objLocs(j,2)]';
          xy = xy + [odom(1) odom(2)]';
          plot(xy(1),xy(2),'b.');
        end;
      end;
      drawnow;
    end;
  
    % Try to look at the objects
    objsSize = size(objs);
    objLocsSize = size(objLocs);
    numZeroes = objLocsSize(1) - objsSize(1);
    objs = [ objLocs(1:objsSize(1),:)         objs(1:objsSize(1),4); 
             objLocs(objsSize(1)+1:end,:)     zeros(numZeroes,1)  ];

    
    objs = lookAtPreviousObjects(objs,CAMERAHEIGHT,ptu_lim,configFile,usingLaptop,canonfl);

    % Move to a new location
    powerbot('movebot');

    iter = 1;
    speeds = [0 0 0];
    while iter <= 10 | abs(speeds(1)) > 0.01

      % Do SLAM only if we're not turning too fast
      fprintf('turn speed: %g',abs(speeds(3)));
      if abs(speeds(3)) < 0.5

        odom = powerbot('getodom');
        [lRgs lPts] = powerbot('getlaser');
 
        %log odometry
        odomStr = sprintf('o %g %g %g\n',odom);
        fwrite(logFile,odomStr,'char');

        %log laser
        fwrite(logFile,'l ','char');
        if length(lRgs) > 0
          laserStr = sprintf('%g ',lRgs);
          fwrite(logFile,laserStr,'char');
        end;
        fwrite(logFile,crStr,'char');

        if length(odom) == 3 & length(lRgs) == 361
          performSlam('do_slam', odom, lRgs);
        end;

      end;

      speeds = powerbot('getspeeds');
      iter = iter + 1;
    end;

    coarse = performSlam('get_pmap');
    fine = performSlam('get_rmap');
    if DISPLAY
      figure(10);
      imshow(fine);
      drawnow;
    end;
end;

toc

%!chmod -R g+rwx /lci/project/raid1/srvc/DR1/robotImages
%!chmod -R g+rwx /lci/project/raid1/srvc/DR1/robotImages_bbox

%close log file
fclose(logFile);

%stop SLAM
performSlam('stop');

powerbot('setpos',0,0);
if usingLaptop
  camera('stop_capture');
  camera('disconnect');
  if canonfl
    cameraCapture('stop');
  end;
end;
powerbot('disconnect');

