%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Function plot FOV displays the extents of a camera's field of view
%% based on the input parameters:
%%
%% A: intrinsic parameter matrix
%% resX: number of pixels in the image X direction
%% resY: number of pixels in the image Y direction
%%
%% David Meger, February, 2007
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [ worldCorners ] = plotFOV( A, T_WC, resX, resY, fig, silent )

    if ~silent
        figure(fig);
        hold on;
    end;
    
    A_inv = inv(A);
    
    % Image coordinates for viewpoint extremes
    XCrop = 30;
    YCropTop = 50;
    image_corner{1} = [0+XCrop; 0+YCropTop; 1];
    image_corner{2}  = [resX-XCrop; 0+YCropTop; 1];
    image_corner{3} = [resX-XCrop; resY; 1];
    image_corner{4}  = [0+XCrop; resY; 1];
    
    if ~silent
        % Plot the camera and it's z-axis
        w1 = 0 .* A_inv * [resX/2; resY/2; 1];
        w1 = T_WC(1:3,1:3) * w1 + T_WC(1:3,4);
        w2 = 0.10 .* A_inv * [resX/2; resY/2; 1];
        w2 = T_WC(1:3,1:3) * w2 + T_WC(1:3,4);
        plot3( w1(1),w1(2),w1(3), 'k+');
        plot3( [w1(1) w2(1)],[w1(2) w2(2)], [w1(3) w2(3)], 'k-');
    end;
    
    
    for i=1:4
        
        % Must find the correct projection factor so that the line we draw
        % meets the floor (if this is possible)
        
        z_prime = T_WC(3,1:3) * A_inv * image_corner{i};
        
        if z_prime < 0
            alpha = -T_WC(3,4) ./ z_prime;
        else
            alpha = 3.0;
        end;
        
        % Plot a line connecting two points of the world-coordinate extremes
        w1 = alpha .* A_inv * image_corner{i};
        w1 = T_WC(1:3,1:3) * w1 + T_WC(1:3,4);
        worldCorners{i} = w1;
        w2 = 0 .* A_inv * image_corner{i};
        w2 = T_WC(1:3,1:3) * w2 + T_WC(1:3,4);
        if ~silent
            plot3( [w1(1) w2(1)],[w1(2) w2(2)], [w1(3) w2(3)], 'b--');
        end;
        
%         % Also plot the image points themselves as dots
%         w1 = 1 .* A_inv * image_corner{i};
%         w1 = T_WC(1:3,1:3) * w1 + T_WC(1:3,4);
%         plot3( w1(1),w1(2),w1(3), 'r.' );

        % Now join this point up with the next one to show the workspace
        % extents
        
        z_prime = T_WC(3,1:3) * A_inv * image_corner{mod(i,4)+1};
        
        if z_prime < 0
            alpha = -T_WC(3,4) ./ z_prime;
        else
            alpha = 3.0;
        end;
        
        w2 = alpha .* A_inv * image_corner{mod(i,4)+1};
        w2 = T_WC(1:3,1:3) * w2 + T_WC(1:3,4);
        if ~silent
            plot3( [w1(1) w2(1)],[w1(2) w2(2)], [w1(3) w2(3)], 'b--');
        end;

    end;
    
%     % Plot the floor plane
%     workspace_corner{1} = [ 0; -50; 0 ];
%     workspace_corner{2} = [ 100; -50; 0 ];
%     workspace_corner{3} = [ 100; 50; 0 ];
%     workspace_corner{4} = [ 0; 50; 0 ];
%     
%     for i=1:4
%         w1 = workspace_corner{i};
%         w2 = workspace_corner{mod(i,4)+1};
%         plot3( [w1(1) w2(1)],[w1(2) w2(2)], [w1(3) w2(3)], 'k-');
%     end;
%    
    if ~silent 
        xlabel('X (m)');
        ylabel('Y (m)');
        zlabel('Z (m)');
    end;
    
    
    