/*
 * Matlab Mex file for performing SLAM
 */

#include "mex.h"

#include "lodo.h"
#include "pmap.h"
#include "rmap.h"
#include "omap.h"

/* SLAM constants */
static const int NUMRANGES = 361;
static const double RANGEMAX = 8.0;
static const double RANGERES = 0.1;
static const double RANGESTART = -M_PI/2;
static const double RANGESTEP = M_PI/360;
static const int SAMPLESLEN = 1000;
static const double GRIDWIDTH = 32.0;
static const double GRIDHEIGHT = 24.0;
static const double GRIDSCALE = 0.1;

/* SLAM globals */
static int started = 0;
static int updatePmap = 1;
static int updateRmap = 1;
static lodo_t *lodo;
static pmap_t *pmap;
static rmap_t *rmap;
static omap_t *omap;

void initializeSlam(double *odom)
{
  lodo = lodo_alloc(NUMRANGES, RANGEMAX, RANGERES,
                   RANGESTART, RANGESTEP);

  lodo->laser_pose.pos.x = 0.1;
  lodo->laser_pose.pos.y = 0;
  lodo->laser_pose.rot = 0;

  pmap = pmap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP, SAMPLESLEN,
                   GRIDWIDTH, GRIDHEIGHT, GRIDSCALE);

  rmap = rmap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP,
                   GRIDWIDTH, GRIDHEIGHT);

  omap = omap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP,
                   GRIDWIDTH, GRIDHEIGHT, GRIDSCALE);

  /* Get Initial Odometry */
  pose2_t pose;
  pose.pos.x = odom[0];
  pose.pos.y = odom[1];
  pose.rot = odom[2];

  pmap_set_pose(pmap, pose);
}

void stopSlam()
{
  lodo_free(lodo);
  pmap_free(pmap);
  rmap_free(rmap);
  omap_free(omap);
}

void doSlam(double *odom, double *ranges)
{
  /* Get Odometry */
  pose2_t pose;
  pose.pos.x = odom[0];
  pose.pos.y = odom[1];
  pose.rot = odom[2];

  /* Get laser-stabilized odometry */
  pose2_t lodoPose = lodo_add_scan(lodo, pose, NUMRANGES, ranges);
  
  /* Update pmap */
  lodoPose = pose2_add(lodo->laser_pose, lodoPose);
  pmap_update(pmap, lodoPose, NUMRANGES, ranges);

  updatePmap = 1;
  updateRmap = 1;
}

void build_pmap(int nObstacles, vector3_t *obstacles)
{
  if(!updatePmap)
    return;

  updatePmap = 0;

  /* Populate omap */
  int i;
  pmap_sample_t *sample;
  pmap_scan_t *scan;
  pose2_t pose;

  sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);

  omap_clear(omap);

  /* Create map and write trajectory */
  for (i = 0; i < pmap->step_count; i++)
  {
    scan = pmap->scans + i;
    pose = sample->poses[i];

    omap_add(omap, pose, pmap->num_ranges, scan->ranges);
  }

  if (nObstacles > 0)
  {
    omap_add_obstacles(omap, nObstacles, obstacles);
  }
}

void get_pmap(unsigned char **map, int nObstacles, vector3_t *obstacles)
{
  build_pmap(nObstacles, obstacles);
  omap_get_map(omap, map);
}

/* Save the coarse map */
void save_pmap(int nObstacles, vector3_t *obstacles)
{
  build_pmap(nObstacles, obstacles);

  /* Write map */
  char filename[256];
  snprintf(filename, sizeof(filename), "coarse.pgm");
  fprintf(stderr, "saving %s\n", filename);  
  omap_save_pgm(omap, filename);
  
  return;
}

/* Fine map generation (relax in-memory map) */
void build_rmap(int nObstacles, vector3_t *obstacles)
{
  if(!updateRmap)
    return;

  updateRmap = 0;

  int i;
  pose2_t pose;
  pmap_scan_t *pScan;
  pmap_sample_t *sample;

  rmap_clear(rmap);

  /* Get the current best sample */
  sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);

  /* If the map is empty, initialize it */
  for (i = 0; i < pmap->step_count; i++)
  {
    pScan = pmap->scans + i;
    pose = sample->poses[i];
    rmap_add(rmap, pose, pmap->num_ranges, pScan->ranges);
  }

  /* Match points periodically; HACK */
  /* if (fine_count % 10 == 0) */
  rmap_match(rmap);

  /* Relax map */
  rmap_relax(rmap, 100);

  /* Interpolate */
  rmap_interpolate(rmap);

  /* Show progress */
  fprintf(stderr, "%d %d %d %f \r",
          rmap->num_scans, rmap->num_key_scans, rmap->num_cons, rmap->relax_err);
  fflush(stderr);


  /* Populate omap */
  rmap_scan_t *rScan;
 
  omap_clear(omap);
    
  /* Create map and write trajectory */
  for (i = 0; i < rmap->num_scans; i++)
  {
    rScan = rmap->scans + i;
    pose = rScan->pose;
    omap_add(omap, pose, rmap->num_ranges, rScan->ranges);
  }

  if (nObstacles > 0)
  {
    omap_add_obstacles(omap, nObstacles, obstacles);
  }
}

void get_rmap(unsigned char **map, int nObstacles, vector3_t *obstacles)
{
  build_rmap(nObstacles, obstacles);
  omap_get_map(omap, map);
}

/* Save the fine map */
void save_rmap(int nObstacles, vector3_t *obstacles)
{
  build_rmap(nObstacles, obstacles);

  /* Write map */
  char filename[256];
  snprintf(filename, sizeof(filename), "fine.pgm");
  fprintf(stderr, "saving %s\n", filename);
  omap_save_pgm(omap, filename);
  
  return;
}

vector2_t transformObject( pose2_t oldFrame, pose2_t curFrame, vector2_t objPos )
{
	vector2_t worldVec;
  vector2_t tempVec;
	vector2_t resultVec;

	double c1 = cos(oldFrame.rot);
	double s1 = sin(oldFrame.rot);

  /* Transform objPos to world frame */
	worldVec.x = oldFrame.pos.x + c1*objPos.x - s1*objPos.y;
	worldVec.y = oldFrame.pos.y + s1*objPos.x + c1*objPos.y;

	c1 = cos(curFrame.rot);
	s1 = sin(curFrame.rot);

  /* Transform objPos to correct position of current frame */
  tempVec.x = -curFrame.pos.x+worldVec.x;
  tempVec.y = -curFrame.pos.y+worldVec.y;

  /* Transform objPos to correct rotation of current frame */
	resultVec.x = c1*tempVec.x + s1*tempVec.y;
	resultVec.y = -s1*tempVec.x + c1*tempVec.y;
	
	return resultVec;
}

int calculateLocations(int nObjs, object_t *objects, vector3_t **locations, pose2_t **rPoses)
{
  if(nObjs < 1)
  {
    if(objects != NULL)
      free(objects);
    return 0;
  }

  *locations = mxMalloc(nObjs * sizeof(vector3_t));
  *rPoses = mxMalloc(nObjs * sizeof(pose2_t));

  pose2_t curPose;
  rmap_scan_t *scan;
  pmap_sample_t *sample;

  if(pmap->step_count > 5)
  {
    /* perform relaxation to get accurate poses */
    build_rmap(0, NULL);

    /* Get current pose */
    scan = rmap->scans + rmap->num_scans - 1;
    curPose = scan->pose;
  } else {
    /* Just use poses from pmap */

    /* Get the current best sample */
    sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);

    /* Get current pose */
    curPose = sample->poses[pmap->step_count-1];
  }

  /* Calculate locations of objects as offsets to
    * the current pose */
  pose2_t objPose; 
  int i;
  for (i = 0; i < nObjs; ++i)
  {
    /* Get the pose when the object was detected */
    object_t obj = objects[i];
    int pose_index = obj.pose_index;
    if (pose_index < 0)
    {
      /* No position information to return */
      (*locations)[i].x = 0.0; 
      (*locations)[i].y = 0.0;
      (*locations)[i].z = 0.0;
    } else {
      if(pmap->step_count > 5)
      {
        scan = rmap->scans + pose_index;
        objPose = scan->pose;
      } else {
        objPose = sample->poses[pose_index];
      }
  
      /* loc_current = objPose - curPose + loc_detected */
      vector2_t loc_detected;
      loc_detected.x = obj.location.x;
      loc_detected.y = obj.location.y;

      /*pose2_t temp = pose2_sub(objPose, curPose);
      vector2_t result = pose2_add_pos(loc_detected, temp);*/
      
      vector2_t result = transformObject(objPose,curPose, loc_detected);
      (*locations)[i].x = result.x; 
      (*locations)[i].y = result.y;
      (*locations)[i].z = obj.location.z;

      (*rPoses)[i] = objPose;
    }
  }

  if(objects != NULL)
    free(objects);
  return nObjs;
}

void addObjects(int nObjs, vector3_t *locations)
{
  int i;
  for(i=0; i < nObjs; ++i)
    pmap_add_object(pmap, locations[i]);
}

int getObjects(vector3_t **locations, pose2_t **rPoses)
{
  /** Get salient objects from pmap */
  object_t *objects = NULL;
  int nObjs = pmap_get_objects(pmap, &objects);
  return calculateLocations(nObjs,objects,locations,rPoses);
}

void addObstacles(int nObjs, vector3_t *locations)
{
  int i;
  for(i=0; i < nObjs; ++i)
    pmap_add_obstacle(pmap, locations[i]);

  updatePmap = 1;
  updateRmap = 1;
}

int getObstacles(vector3_t **locations,pose2_t **rPoses)
{
  /** Get salient objects from pmap */
  object_t *obstacles = NULL;
  int nObstacles = pmap_get_obstacles(pmap, &obstacles);
  return calculateLocations(nObstacles,obstacles,locations,rPoses);
}

void getCurVecPos(double *vecPos)
{
  pose2_t curPose;
  if(pmap->step_count > 5)
  {
    /* Perform relaxation to get accurate pose */
    build_rmap(0, NULL);

    rmap_scan_t *scan = rmap->scans + rmap->num_scans - 1;
    curPose = scan->pose;
  } else {
    /* Just use pose from pmap */
    pmap_sample_t *sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);
    curPose = sample->poses[pmap->step_count-1];
  }

  vecPos[0] = curPose.pos.x;
  vecPos[1] = curPose.pos.y;
  vecPos[2] = curPose.rot;
}

void getCurGridPos(int *gridPos)
{
  if(pmap->step_count > 5)
  {
    /* Perform relaxation to get accurate pose */
    build_rmap(0, NULL);
  } else {
    /* Just use pose from pmap */
    build_pmap(0, NULL);
  }

  omap_get_curGridPos(omap,gridPos);

  /* Add 1 because the matlab grid array is 1-indexed */
  gridPos[0]+=1;
  gridPos[1]+=1;
}

void getVecPosFromGridPos(int size, int *gridPos, vector2_t *posVecs)
{
  if(pmap->step_count > 5)
  {
    /* Perform relaxation to get accurate pose */
    build_rmap(0, NULL);
  } else {
    /* Just use pose from pmap */
    build_pmap(0, NULL);
  }

  /* Subtract 1 because the C grid array is 0-indexed */
  int nEntries = size*2;
  int i;
  for(i=0; i < nEntries; ++i)
  {
    gridPos[i]-=1;
  }

  omap_gridPos_to_vecPos(omap,size,gridPos,posVecs);
}

void getGridPosFromVecPos(int size, vector2_t *vecPos, int *gridPos)
{
  if(pmap->step_count > 5)
  {
    /* Perform relaxation to get accurate pose */
    build_rmap(0, NULL);
  } else {
    /* Just use pose from pmap */
    build_pmap(0, NULL);
  }

  omap_vecPos_to_gridPos(omap,size,vecPos,gridPos);

  /* Add 1 because the matlab grid array is 1-indexed */
  int nEntries = size*2;
  int i;
  for(i=0; i < nEntries; ++i)
  {
    gridPos[i]+=1;
  }
}

int parseOdom(const mxArray *prhs, double *odom)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(ndims < 2 || dims[0] != 1 || dims[1] != 3)
  {
    mexPrintf("Error: Expected 3 entries for odometry.\n");
    return -1;
  }

  double *matlab_array=mxGetPr(prhs);
  odom[0] = matlab_array[0];
  odom[1] = matlab_array[1];
  odom[2] = matlab_array[2];

  return 0;
}

int parseReadings(const mxArray *prhs, double *readings)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(ndims < 2 || dims[0] != 1 || dims[1] != NUMRANGES)
  {
    mexPrintf("Error: Expected %d entries for laser.\n", NUMRANGES);
    return -1;
  }

  double *matlab_array=mxGetPr(prhs);
  int i;
  for(i=0; i < NUMRANGES; ++i)
    readings[i] = matlab_array[i]; 

  return 0;
}

int parseObjects(const mxArray *prhs, vector3_t **locations)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(dims[0] == 0)
  {
     /* Empty array, do nothing */
     return 0;
  }
  if(ndims < 2 || dims[1] < 3)
  {
    mexPrintf("Error: Expected x,y,z coordinates for each object.\n");
    return -1;
  }

  int nObjs = dims[0];
  *locations = mxMalloc(nObjs * sizeof(vector3_t));

  double *matlab_array=mxGetPr(prhs);
  int i;
  for(i=0; i < nObjs; ++i)
  {
    (*locations)[i].x = matlab_array[i];
    (*locations)[i].y = matlab_array[i+nObjs];
    (*locations)[i].z = matlab_array[i+2*nObjs];
  }

  return nObjs;
}

int parseObstacles(const mxArray *prhs, vector2_t **locations)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  int nObjs = dims[0];
  if(nObjs == 0)
  {
    /* No locations */
    return 0;
  }

  if(ndims < 2 || dims[1] < 2)
  {
    mexPrintf("Error: Expected x,y coordinates for each object.\n");
    return -1;
  }

  *locations = mxMalloc(nObjs * sizeof(vector2_t));

  double *matlab_array=mxGetPr(prhs);
  int i;
  for(i=0; i < nObjs; ++i)
  {
    (*locations)[i].x = matlab_array[i];
    (*locations)[i].y = matlab_array[i+nObjs];
  }

  return nObjs;
}

int parseGridPos(const mxArray *prhs, int **gridPos)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(ndims < 2 || dims[1] != 2)
  {
    mexPrintf("Error: Expected x,y coordinates for each grid cell.\n");
    return -1;
  }

  double *matlab_array=mxGetPr(prhs);

  int size = dims[0];
  *gridPos = calloc(size*2,sizeof(int));

  int i;
  for(i=0; i < size; ++i)
  {
    (*gridPos)[2*i] = (int)matlab_array[i];
    (*gridPos)[2*i+1] = (int)matlab_array[i+size];
  }
  return size;
}

int parsePosVecs(const mxArray *prhs, vector2_t **posVecs)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(ndims < 2 || dims[1] != 2)
  {
    mexPrintf("Error: Expected [x y] position vector.\n");
    return -1;
  }

  double *matlab_array=mxGetPr(prhs);
  int size = dims[0];
  *posVecs = calloc(size,sizeof(vector2_t));

  int i;
  for(i=0; i < size; ++i)
  {
    (*posVecs)[i].x = matlab_array[i];
    (*posVecs)[i].y = matlab_array[i+size];
  }
  return size;
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/
#define BUF_LEN 32
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
  char cstr[BUF_LEN+1];
  int out_dims[]={0,0,0};
  double *matlab_array;
  unsigned char *matlab_map;

  double odom[3];
  double readings[NUMRANGES];
  vector3_t *locations = NULL;
  pose2_t *rPoses = NULL;
  object_t *objects = NULL;
  int gridPos[2];

 /* Argument count check */
  if(nrhs ==0) goto mexErrExit;
    
  /* Parse arg 1 (request string) */
  if(mxGetClassID(prhs[0]) != mxCHAR_CLASS) {
    mexPrintf("Type mismatch: <request> should be of class char.\n");
    goto mexErrExit;
  }
  mxGetString(prhs[0], cstr,BUF_LEN);

  if(strcmp(cstr,"start")==0) {
    /* Initialize SLAM, parse odometry */
    if(nrhs != 2) {
      mexPrintf("Error: request 'start' expects two arguments.\n");
      goto mexErrExit;
    }

    if(started == 1)
    {
      mexPrintf("SLAM already started.\n");
      return;
    }

    if(parseOdom(prhs[1], odom) < 0)
      goto mexErrExit;

    initializeSlam(odom);
    started = 1;
    mexPrintf("SLAM started.\n");
    return;
  }

  /* All other requests require that SLAM be started */
  if(started != 1) {
    mexErrMsgTxt("SLAM hasn't been started.");
    return;
  }

  if(strcmp(cstr,"do_slam")==0) {
    /* Do SLAM, parse odometry and laser */
    if(nrhs != 3) {
      mexPrintf("Error: request 'doSlam' expects three arguments.\n");
      goto mexErrExit;
    }

    if(parseOdom(prhs[1], odom) < 0)
      goto mexErrExit;
    if(parseReadings(prhs[2], readings) < 0)
      goto mexErrExit;

    doSlam(odom, readings);
    return;
  }

  if(strcmp(cstr,"add_objects")==0) {
    /* Add salient objects */
    if(nrhs != 2) {
      mexPrintf("Error: request 'add_objects' expects two arguments.\n");
      goto mexErrExit;
    }

    /* Parse objects */
    int nObjs = parseObjects(prhs[1], &locations);
    if(nObjs < 0)
    {
      if(locations != NULL)
        mxFree(locations);
      goto mexErrExit;
    }

    if(locations != NULL)
    {
      addObjects(nObjs, locations);
      mxFree(locations);
    }
    return;
  }

  if(strcmp(cstr,"add_obstacles")==0) {
    /* Add visually-detected obstacles */
    if(nrhs != 2) {
      mexPrintf("Error: request 'add_obstacles' expects two arguments.\n");
      goto mexErrExit;
    }

    /* Parse objects */
    int nObstacles = parseObjects(prhs[1], &locations);
    if(nObstacles < 0)
    {
      if(locations != NULL)
        mxFree(locations);
      goto mexErrExit;
    }

    if(locations != NULL)
    {
      addObstacles(nObstacles, locations);
      mxFree(locations);
    }
    return;
  }

  if(strcmp(cstr,"get_curvecpos")==0) {
    /* Get odometry from best map */
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_curvecpos' expects one argument.\n");
      goto mexErrExit;
    }

    getCurVecPos(odom);

    out_dims[0]=1;out_dims[1]=3;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)odom[0];
    matlab_array[1]=(double)odom[1];
    matlab_array[2]=(double)odom[2];
    return;
  }

  if(strcmp(cstr,"get_curgridpos")==0) {
    /* Get robot's position in the best occupancy grid */
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_curgridpos' expects one argument.\n");
      goto mexErrExit;
    }

    getCurGridPos(gridPos);

    out_dims[0]=1;out_dims[1]=2;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)gridPos[0];
    matlab_array[1]=(double)gridPos[1];
    return;
  }

  if(strcmp(cstr,"get_vecpos_from_gridpos")==0) {
    /* Get vector from grid coordinate */
    if(nrhs != 2) {
      mexPrintf("Error: request 'get_vecpos_from_gridpos' expects two arguments.\n");
      goto mexErrExit;
    }

	int *gridPos;
	int size = parseGridPos(prhs[1], &gridPos);
    if(size < 0)
      goto mexErrExit;

    vector2_t *posVecs = calloc(size,sizeof(vector2_t));
    getVecPosFromGridPos(size,gridPos,posVecs);

    out_dims[0]=size;out_dims[1]=2;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
	int i;
	for(i=0; i < size; ++i)
	{
      matlab_array[i]=posVecs[i].x;
      matlab_array[i+size]=posVecs[i].y;
	}

	free(gridPos);
	free(posVecs);
    return;
  }

  if(strcmp(cstr,"get_gridpos_from_vecpos")==0) {
    /* Get vector from grid coordinate */
    if(nrhs != 2) {
      mexPrintf("Error: request 'get_gridpos_from_vecpos' expects two arguments.\n");
      goto mexErrExit;
    }

    vector2_t *posVecs;
	int size = parsePosVecs(prhs[1],&posVecs);
    if(size < 0)
      goto mexErrExit;

	int *gridPos = calloc(size*2,sizeof(int));
    getGridPosFromVecPos(size,posVecs, gridPos);

    out_dims[0]=size;out_dims[1]=2;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
	int i;
	for(i=0; i < size; ++i)
	{
      matlab_array[i]=gridPos[2*i];
      matlab_array[i+size]=gridPos[2*i+1];
	}

    free(posVecs);
    free(gridPos);
    return;
  }

  if(strcmp(cstr,"get_objects")==0) {
    /* Get the salient objects */
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_objects' expects one argument.\n");
      goto mexErrExit;
    }

    if(locations != NULL)
      mexPrintf("locations not NULL!\n");
    int nObjs = getObjects(&locations,&rPoses);

    /* Return the locations */
    if(nlhs>0) {
      if(nObjs > 0) {
        out_dims[0]=nObjs;out_dims[1]=6;out_dims[2]=0;
        plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
        matlab_array=mxGetPr(plhs[0]);
        int i;
        for(i=0; i < nObjs; ++i)
        {
          matlab_array[i] = locations[i].x;
          matlab_array[i+nObjs] = locations[i].y;
          matlab_array[i+2*nObjs] = locations[i].z;

          matlab_array[i+3*nObjs] = rPoses[i].pos.x;
          matlab_array[i+4*nObjs] = rPoses[i].pos.y;
          matlab_array[i+5*nObjs] = rPoses[i].rot;
        }
      } else {
      out_dims[0]=out_dims[1]=out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      }
    }

    if(locations != NULL)
      mxFree(locations);
      mxFree(rPoses);
    return;
  }

  if(strcmp(cstr,"get_obstacles")==0) {
    /* Get visually-detected obstacles */
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_obstacles' expects one argument.\n");
      goto mexErrExit;
    }

    if(locations != NULL)
      mexPrintf("locations not NULL!\n");
    int nObstacles = getObstacles(&locations,&rPoses);

    /* Return the locations */
    if(nlhs>0) {
      if(nObstacles > 0) {
        out_dims[0]=nObstacles;out_dims[1]=6;out_dims[2]=0;
        plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
        matlab_array=mxGetPr(plhs[0]);
        int i;
        for(i=0; i < nObstacles; ++i)
        {
          matlab_array[i] = locations[i].x;
          matlab_array[i+nObstacles] = locations[i].y;
          matlab_array[i+2*nObstacles] = locations[i].z;

          matlab_array[i+3*nObstacles] = rPoses[i].pos.x;
          matlab_array[i+4*nObstacles] = rPoses[i].pos.y;
          matlab_array[i+5*nObstacles] = rPoses[i].rot;
        }
      } else {
      out_dims[0]=out_dims[1]=out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      }
    }

    if(locations != NULL)
      mxFree(locations);
      mxFree(rPoses);
    return;
  }

  if(strcmp(cstr,"get_pmap")==0) {
    /* Return the pmap */
    if(nrhs < 1 || nrhs > 2) {
      mexPrintf("Error: request 'get_pmap' expects one or two arguments.\n");
      goto mexErrExit;
    }

    int grid_sx = omap->grid_sx;
    int grid_sy = omap->grid_sy;
    unsigned char **map = calloc(grid_sy,sizeof(char *));
    int i,j;
    for(j=0; j < grid_sy; ++j)
      map[j] = calloc(grid_sx,sizeof(char));
    
    int nObstacles = getObstacles(&locations,&rPoses);
    get_pmap(map, nObstacles, locations);
    
    /* Write the matlab array */
    out_dims[0]=grid_sy;out_dims[1]=grid_sx;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
    matlab_map=(unsigned char *)mxGetPr(plhs[0]);
    for (j=0; j < grid_sy; ++j)
      for (i=0; i < grid_sx; ++i)
        matlab_map[j+i*grid_sy]=map[j][i];

    if(locations != NULL)
      mxFree(locations);
    return;
  }

  if(strcmp(cstr,"save_pmap")==0) {
    /* Save the pmap */
    if(nrhs < 1 || nrhs > 2) {
      mexPrintf("Error: request 'save_pmap' expects one or two arguments.\n");
      goto mexErrExit;
    }

    mexPrintf("Saving pmap.. ");
    int nObstacles = getObstacles(&locations,&rPoses);
    save_pmap(nObstacles, locations);
    mexPrintf("done\n");
    return;
  }

  if(strcmp(cstr,"get_rmap")==0) {
    /* Build and return rmap */
    if(nrhs < 1 || nrhs > 2) {
      mexPrintf("Error: request 'get_rmap' expects one or two arguments.\n");
      goto mexErrExit;
    }

    int grid_sx = omap->grid_sx;
    int grid_sy = omap->grid_sy;
    unsigned char **map = calloc(grid_sy,sizeof(char *));
    int i,j;
    for(j=0; j < grid_sy; ++j)
      map[j] = calloc(grid_sx,sizeof(char));
    
    int nObstacles = getObstacles(&locations,&rPoses);
    get_rmap(map, nObstacles, locations);
    
    /* Write the matlab array */
    out_dims[0]=grid_sy;out_dims[1]=grid_sx;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
    matlab_map=(unsigned char *)mxGetPr(plhs[0]);
    for (j=0; j < grid_sy; ++j)
      for (i=0; i < grid_sx; ++i)
        matlab_map[j+i*grid_sy]=map[j][i];

    if(locations != NULL)
      mxFree(locations);
    return;
  }

  if(strcmp(cstr,"save_rmap")==0) {
    /* Build and save rmap */
    if(nrhs < 1 || nrhs > 2) {
      mexPrintf("Error: request 'save_rmap' expects one or two arguments.\n");
      goto mexErrExit;
    }

    mexPrintf("Saving rmap... ");
    int nObstacles = getObstacles(&locations,&rPoses);
    save_rmap(nObstacles, locations);
    mexPrintf("done\n");
    return;
  }

  if(strcmp(cstr,"get_map")==0) {
    /* Return the best map */
    if(nrhs < 1 || nrhs > 2) {
      mexPrintf("Error: request 'get_map' expects one or two arguments.\n");
      goto mexErrExit;
    }

    int grid_sx = omap->grid_sx;
    int grid_sy = omap->grid_sy;
    unsigned char **map = calloc(grid_sy,sizeof(char *));
    int i,j;
    for(j=0; j < grid_sy; ++j)
      map[j] = calloc(grid_sx,sizeof(char));
    
    int nObstacles = getObstacles(&locations,&rPoses);
    if(pmap->step_count > 5)
    {
      get_rmap(map, nObstacles, locations);
    } else {
      get_pmap(map, nObstacles, locations);
    }
    
    /* Write the matlab array */
    out_dims[0]=grid_sy;out_dims[1]=grid_sx;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
    matlab_map=(unsigned char *)mxGetPr(plhs[0]);
    for (j=0; j < grid_sy; ++j)
      for (i=0; i < grid_sx; ++i)
        matlab_map[j+i*grid_sy]=map[j][i];

    if(locations != NULL)
      mxFree(locations);
    return;
  }

  if(strcmp(cstr,"stop")==0) {
    /* Stop SLAM */
    if(nrhs != 1) {
      mexPrintf("Error: request 'stop' expects one argument.\n");
      goto mexErrExit;
    }

    stopSlam();
    started = 0;
    mexPrintf("SLAM stopped.\n");
    return;
  }

  if(strcmp(cstr,"get_gridtransform")==0) {
    /* Get the slam->grid coordinate transform matrix */
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_gridtransform' expects one argument.\n");
      goto mexErrExit;
    }

    
    /* Convenience variables: */
    double r = omap->grid_res;
    double xs = omap->grid_sx;
    double ys = omap->grid_sy;

    /* Initialize the world->grid transformation matrix: */
    mxArray *trMatrix = mxCreateDoubleMatrix (3, 3, mxREAL);
    mxSetCell (trMatrix, 0, mxCreateDoubleScalar(1/r));    /* ie: topleft */
    mxSetCell (trMatrix, 4, mxCreateDoubleScalar(-1/r));   /* ie: middle */
    mxSetCell (trMatrix, 6, mxCreateDoubleScalar(xs/2.0)); /* ie: topright */
    mxSetCell (trMatrix, 7, mxCreateDoubleScalar(ys/2.0)); /* ie: midright */
    mxSetCell (trMatrix, 8, mxCreateDoubleScalar(1.0));    /* ie: botright */
    plhs[0] = trMatrix;
    return;
  }

  if(strcmp(cstr,"get_worldtransform")==0) {
    if(nrhs != 1) {
      mexPrintf("Error: request 'get_worldtransform' expects one argument.\n");
      goto mexErrExit;
    }
    

    /* Convenience variables: */
    double r = omap->grid_res;
    double xs = omap->grid_sx;
    double ys = omap->grid_sy;

    mxArray *trMatrix = mxCreateDoubleMatrix (3, 3, mxREAL);
    mxSetCell (trMatrix, 0, mxCreateDoubleScalar(r));         /* ie: topleft */
    mxSetCell (trMatrix, 4, mxCreateDoubleScalar(-r));        /* ie: middle */
    mxSetCell (trMatrix, 6, mxCreateDoubleScalar(-r*xs/2.0)); /* ie: topright */
    mxSetCell (trMatrix, 7, mxCreateDoubleScalar(r*ys/2.0));  /* ie: midright */
    mxSetCell (trMatrix, 8, mxCreateDoubleScalar(1.0));       /* ie: botright */
    plhs[0] = trMatrix;
    return;
  }
  
  /* Exit with error message */
  mexErrExit:
  mexErrMsgTxt("Usage: [<map>]=performSlam(<request>,...)\n");
  mexExit: {}
}
