/*
**  Matlab client for still camera
*/

#include <string.h>
#include "mex.h"
#include "capture_mini.h"

static int started = 0;

typedef struct camera_packet {
  int width;
  int height;
  int size;
  char *data;
} camera_packet;

void free_camera_packet(camera_packet *camera) {
  if(camera->data != NULL)
    free(camera->data);
  free(camera);
}

void sig_shutdown(int arg)
{
  capture_stop();
  exit(arg);
}

camera_packet *request_takephoto(char *fname) {
  capture_grab(fname);
  
  camera_packet *cpacket = NULL;
  char *img;
  int fread_bytes=capture_loadimage(fname,&img);
  if(fread_bytes<0) printf("Ouch!\n");
  else {
    printf("Compose packet\n");
    cpacket = (camera_packet *)calloc(1,sizeof(camera_packet));
    cpacket->width=fread_bytes/64; /* Ugly fix to fit into short */
    cpacket->height=64;
    cpacket->size=fread_bytes;
    cpacket->data=(char *)img;
  }
  return cpacket;
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/
#define BUF_LEN 256
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
  const int *arg_dims;
  int arg_ndims;
  char cstr[BUF_LEN+1];
  int zoom = 6;      /* Default */
  int out_dims[]={0,0,0};
  camera_packet *cpack;
  double *matlab_array;

 /* Argument count check */
  if(nrhs ==0) goto mexErrExit;
    
  /* Parse arg 1 (request string) */
  arg_dims = mxGetDimensions(prhs[0]);
  arg_ndims = mxGetNumberOfDimensions(prhs[0]);
  if(mxGetClassID(prhs[0]) != mxCHAR_CLASS) {
    mexPrintf("Type mismatch: <request> should be of class char.\n");
    goto mexErrExit;
  }
  mxGetString(prhs[0], cstr,BUF_LEN);

  if(strcmp(cstr,"start")==0) {
    capture_start();
    started = 1;
    mexPrintf("Capture started.\n");
    return;
  }

  /* All other requests require that Capture be started */
  if(started != 1) {
    mexErrMsgTxt("Capture hasn't been started.");
    return;
  }

  /* SETSIZE */
  if(strcmp(cstr,"setsize")==0) {
    if((nrhs >2)||(nrhs <1)) {
      mexPrintf("Error: request 'setsize' expects one argument.\n");
      goto mexErrExit;
    }
    if(nrhs==2) {
      mxGetString(prhs[1], cstr,BUF_LEN);
    }

    capture_setsize(cstr);
     mexPrintf("Capture size set to %s.\n", cstr);
    return;
  }

  /* SETZOOM */
  if(strcmp(cstr,"setzoom")==0) {
    if((nrhs >2)||(nrhs <1)) {
      mexPrintf("Error: request 'setzoom' expects one argument.\n");
      goto mexErrExit;
    }
    if(nrhs==2) {
      zoom=mxGetScalar(prhs[1]);
    }

    capture_setzoom(zoom);
     mexPrintf("Zoom set to %dx.\n", zoom);
    return;
  }

  /* SETFLASH */
  if(strcmp(cstr,"setflash")==0) {
    if((nrhs >2)||(nrhs <1)) {
      mexPrintf("Error: request 'setflash' expects one argument.\n");
      goto mexErrExit;
    }
    if(nrhs==2) {
      mxGetString(prhs[1], cstr,BUF_LEN);
    }
    
    int flash = strcmp(cstr,"on")==0? 1:0;
    capture_setflash(flash);
     mexPrintf("Flash set to %s.\n", cstr);
    return;
  }

  /* TAKE_PHOTO */
  if(strcmp(cstr,"take_photo")==0) {
    if(nrhs < 2)
      strncpy(cstr,"lastframe.jpg",BUF_LEN);
    else
      mxGetString(prhs[1], cstr,BUF_LEN);

    cpack = request_takephoto(cstr);
    if(!cpack)
    {
      mexPrintf("Failed to grab photo.\n");
      sig_shutdown(0);
    }

    out_dims[0]=cpack->width;out_dims[1]=cpack->height;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
    memcpy(mxGetPr(plhs[0]),cpack->data,cpack->size);
    free_camera_packet(cpack);
    return;
  }

  /* STOP */
  if(strcmp(cstr,"stop")==0) {
    capture_stop();
    started = 0;
    return;
  }

  mexErrMsgTxt("cameraCapture: Unknown request");
  goto mexExit;

  /* Exit with error message */
 mexErrExit:
  mexErrMsgTxt("Usage: [<Limg>,<Rimg>]=cameraCapture(<request>,...)\n");
 mexExit: {}
}
