/*
 *  Minimal interface to PTU (modified Player int code)
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <strings.h>
#include <errno.h>
#include <math.h>
#include "ptu46_mini.h"

int PTU46_Write(int fd,char * data, int length) {

  if (fd < 0)
    return -1;
		
  // autocalculate if using short form
  if (length == 0)
    length = strlen(data);
  
  // ugly error handling, if write fails then shut down unit
  if(write(fd, data, length) < length)
    {
      fprintf(stderr,"Error writing to Pan Tilt Unit\n");
      return -1;
    }
  else {
    return 0;
  }
}

/* set position in degrees */
#define RESOLUTION (46.2857/3600)
/* Default parameters from PTU-D46-70 */
#define PTU46_TMin -3672
#define PTU46_TMax 2447
#define PTU46_PMin -12374
#define PTU46_PMax 12375
#define PTU46_RTilt RESOLUTION
#define PTU46_RPan  RESOLUTION
#define POS_FAIL -10000

/*
 *  Send a request to PTU
 *  result is in buffer and return pointer
 *  returns NULL on error.
 */
char *PTU46_Request(int fd,char *buffer,char *cmd) {
  int len = 0;
  if (fd < 0)
    return NULL;
  PTU46_Write(fd,cmd,0);  /* Send request */
  len = read(fd, buffer, PTU46_BUFFER_LEN );
  buffer[len] = '\0';
  printf("received string:%s\n",buffer); 
  if (len < 3 || buffer[0] != '*')
    return NULL;
  return buffer;  
}

/*
 * Get resolution in degrees per count
 * Returns -1 on error
 */
float PTU46_GetRes(int fd,char *buffer,char type) {

  char cmd[4] = " r ";
  cmd[0] = type;

  if(PTU46_Request(fd,buffer,cmd) == NULL) {
    fprintf(stderr,"Error getting pan-tilt res\n");
    return -1;
  }
  return strtod(&buffer[2],NULL)/3600;	
}

/* get PTU position in degrees
 * Returns POS_FAIL on error
 */
float PTU46_GetPos(int fd,char *buffer,PTU46_data *pars,char type) {

  char cmd[4] = " p ";
  cmd[0] = type;

  if(PTU46_Request(fd,buffer,cmd) == NULL) {
    fprintf(stderr,"Error getting pan-tilt pos\n");
    return POS_FAIL;
  }
  return (float)(strtod(&buffer[2],NULL) * 
		 (type == PTU46_TILT ? pars->rtilt : pars->rpan));
}


/*
 * Get a position limit
 */
float PTU46_GetLimit(int fd,char *buffer,PTU46_data *pars,
		     char type, char LimType) {

  char cmd[4] = "   ";
  cmd[0] = type;
  cmd[1] = LimType;
	
  if(PTU46_Request(fd,buffer,cmd) == NULL) {
    fprintf(stderr,"Error getting pan-tilt limit\n");
    return -1;    
  }

  return (float)(strtod(&buffer[2],NULL) * 
		 (type == PTU46_TILT ? pars->rtilt : pars->rpan));
}

/*
 * Initialise serial connection to PTU
 * and return a file descriptor.
 * This function also allocates the parameters
 * struct and fills it with default values.
 * Call PTU46_ReadParams to ensure that it contains
 * the correct values.
 */
int PTU46_Init(char *port,char **bufferp,PTU46_data **ptu_paramsp) {
  struct termios oldtio;
  struct termios newtio;
  int fd;
  char *buffer;
  PTU46_data *params;

  buffer=(char *)calloc(PTU46_BUFFER_LEN,sizeof(char));
  bufferp[0]=buffer; // Return buffer pointer

  params=(PTU46_data *)calloc(1,sizeof(PTU46_data));
  ptu_paramsp[0]=params; // Return struct pointer

  // open the serial port
  printf("Opening serial connection.\n");
	
  fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
  if ( fd<0 ) {
    fprintf(stderr, "Could not open serial device %s\n",port);
    return -1;
  }
  fcntl(fd,F_SETFL, 0);
  
// save the current io settings
  tcgetattr(fd, &oldtio);

  memset(&newtio, 0,sizeof(newtio));
  newtio.c_cflag = /*(rate & CBAUD) |*/ CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_lflag = ICANON;
  // activate new settings
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, &newtio);

  // now set up the pan tilt camera
  PTU46_Write(fd," ",0); // terse feedback
  usleep(100000);
  tcflush(fd, TCIFLUSH);
  
  PTU46_Write(fd,"ft ",0); // terse feedback
  PTU46_Write(fd,"ed ",0); // disable echo
  PTU46_Write(fd,"ci ",0); // position mode

  //PTU46_Write(fd,"pa9000 pu6000 ta9000 tu6000 ",0); // high speed operation
  PTU46_Write(fd,"pa2000 pu2000 pb60 ta2000 tu2000 tb60 ",0); // heavy duty operation

/*   PTU46_Write(fd,"pu ",0); // upper pan speed */
/*   PTU46_Write(fd,"pl ",0); // lower pan speed */
/*   PTU46_Write(fd,"ps ",0); // pan speed */
/*   PTU46_Write(fd,"pb ",0); // pan base speed */
/*   PTU46_Write(fd,"pa ",0); // pan acceleration */

/*   PTU46_Write(fd,"tu ",0); // upper tilt speed */
/*   PTU46_Write(fd,"tl ",0); // lower tilt speed */
/*   PTU46_Write(fd,"ts ",0); // tilt speed */
/*   PTU46_Write(fd,"tb ",0); // tilt base speed */
/*   PTU46_Write(fd,"ta ",0); // tilt acceleration */

  usleep(100000);
  
  int len;
  int k;

  for(k=0;k<9;k++) {
    len = read(fd, buffer, PTU46_BUFFER_LEN );
    buffer[len] = '\0';
    printf( "Message was: %s .\n",buffer);
  }

  tcflush(fd, TCIFLUSH); /* Eat stars */

  /* Fill in default parameters in struct */
  params->tmin  = PTU46_TMin;
  params->tmax  = PTU46_TMax;
  params->pmin  = PTU46_PMin;
  params->pmax  = PTU46_PMax;
  params->rtilt = PTU46_RTilt;
  params->rpan  = PTU46_RPan;

  return fd;
}

/*
 * Should always be called after PTU46_Init to ensure
 * that correct limits and resolution are used
 */
int PTU46_ReadParams(int fd,char *buffer,PTU46_data *pars) {
   // Read resolution
  if(PTU46_Request(fd,buffer,"pr ") == NULL) {
    fprintf(stderr,"Error getting pan res\n");
  }
  pars->rpan = strtod(&buffer[2],NULL)/3600;
  if(PTU46_Request(fd,buffer,"tr ") == NULL) {
    fprintf(stderr,"Error getting tilt res\n");
  }
  pars->rtilt = strtod(&buffer[2],NULL)/3600;

    // Read limits
  if(PTU46_Request(fd,buffer,"pn ") == NULL) {
    fprintf(stderr,"Error getting pan min pos\n");
  }
  pars->pmin = (int)strtod(&buffer[2],NULL);
  if(PTU46_Request(fd,buffer,"px ") == NULL) {
    fprintf(stderr,"Error getting pan max pos\n");
  }
  pars->pmax =  (int)strtod(&buffer[2],NULL);
  if(PTU46_Request(fd,buffer,"tn ") == NULL) {
    fprintf(stderr,"Error getting tilt min pos\n");
  }
  pars->tmin =  (int)strtod(&buffer[2],NULL);
  if(PTU46_Request(fd,buffer,"tx ") == NULL) {
    fprintf(stderr,"Error getting tilt max pos\n");
  }
  pars->tmax =  (int)strtod(&buffer[2],NULL);
}

int PTU46_PrintParams(PTU46_data *ptu_params) {
  printf("PTU46 params:\n");
  printf("tmin=%d tmax=%d \n",ptu_params->tmin,ptu_params->tmax);
  printf("pmin=%d pmax=%d \n",ptu_params->pmin,ptu_params->pmax);
  printf("rtilt=%g rpan=%g \n",ptu_params->rtilt,ptu_params->rpan);
}


// Release buffer and close file
int PTU46_Release(int fd,char *buffer,PTU46_data *params) {
  close(fd);
  free(buffer);
  free(params);
}

/* Set pan or tilt position */
int PTU46_SetPos(int fd,char *buffer,PTU46_data *pars,
		 char type, float pos, int Block) {
  int Count;

  /* Convert degrees to raw encoder count to move to */
  Count = (int)rint(pos/(type == PTU46_TILT ? pars->rtilt : pars->rpan));

  /* Check limits for desired position */
  if (Count < (type == PTU46_TILT ? pars->tmin : pars->pmin) || Count > (type == PTU46_TILT ? pars->tmax : pars->pmax))
    {
      fprintf(stderr,"Pan Tilt Value out of Range: %c %g(%d) (%d-%d)\n", type, pos, Count, (type == PTU46_TILT ? pars->tmin : pars->pmin),(type == PTU46_TILT ? pars->tmax : pars->pmax));
      return -1;
    }
  
  char cmd[16];
  snprintf(cmd,16,"%cp%d ",type,Count);
  fprintf(stdout,"PTU46_SetPos: sending request %c %g [%s]\n",type,pos,cmd);

  if(PTU46_Request(fd,buffer,cmd) == NULL) {
    fprintf(stderr,"Error setting pan-tilt pos\n");
    return -1;
  }
  
  if (Block) {
    while (fabs(PTU46_GetPos(fd,buffer,pars,type)-pos)>(type == PTU46_TILT ? pars->rtilt : pars->rpan));
  }
  return 0;
}

/*
 * Function to set pan and tilt position
 * optionally waiting until completion
 */
int PTU46_SetPanTilt(int fd,char *buffer,PTU46_data *pars,
		     float pan,float tilt,int Block) {
  float cpan,ctilt;
  if(PTU46_SetPos(fd,buffer,pars,PTU46_PAN,pan,PTU46_NOBLOCK)==-1)
    return -1;
  if(PTU46_SetPos(fd,buffer,pars,PTU46_TILT,tilt,PTU46_NOBLOCK)==-1)
    return -1;
  if (Block) {
    do {
      cpan  = PTU46_GetPos(fd,buffer,pars,PTU46_PAN);
      if(cpan==POS_FAIL) break;
      ctilt = PTU46_GetPos(fd,buffer,pars,PTU46_TILT);
      if(ctilt==POS_FAIL) break;
      printf("cpan=%g ctilt=%g\n",cpan,ctilt); 
    } while ((fabs(cpan-pan)>pars->rpan)||(fabs(ctilt-tilt)>pars->rtilt));
    if((cpan==POS_FAIL)||(ctilt==POS_FAIL)) return -1;
  }
  return 0;
}
