/*
 *  server_mini.c
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/poll.h>
#include <string.h>
#include <pthread.h>

#include "server_mini.h"
#include "ptu46_mini.h"
#include "bb_mini.h"
#include "socket_mini.h"
#include "capture_mini.h"
#include "pb_mini.h"

int client_socket=-1;
int client_port=7001;
int backlog=5;
char * logfile="/tmp/seriald.log";
unsigned char reset_str[2];

/* Serial port variables */
int serial_fd;
char *serial_buffer;

/* PTU variables */
PTU46_data *ptu_params;

/* dc1394 camera variables */
char *dma_dev_name = DMA_DEV_NAME;
dc1394_cameracapture camera;
raw1394handle_t handle;
int frame_no=0;

/* Sockets */
struct sockaddr_in client_skname, peername;
int client_connection=-1;  // file descriptor

/* Bumblebee */
int activate_bb = 0;

/* Canon camera */
int activate_camera=0;
int activate_robot=1;

/* PTU */
int activate_ptu=1;

void handleOptions(int argc, char *argv[]) {
  int c;

  while((c=getopt(argc, argv, "v:p:nru"))!=EOF) 
    switch (c) {
    case 'v': dma_dev_name=optarg; activate_bb=1; break;
    case 'p': client_port=atoi(optarg); break;
    case 'n': activate_camera=1; break; /* Don't activate Canon camera */
    case 'r': activate_robot=0; break;
    case 'u': activate_ptu=0;break;
    default: fprintf(stderr, "Usage: %s [-v <video-device>|'skip'] [-p <port>][-n][-u][-r]\n", argv[0]); exit(0);
    }
}


/* The SIGTERM handler for shutting down */
void sig_shutdown(int arg)
{
  unsigned char buf[]={250,251,3,2,0,2};
  if (client_connection>0 && client_connection!=client_socket) close(client_connection);
  if (client_socket>0) close(client_socket);
  /* Release serial port */
  PTU46_Release(serial_fd,serial_buffer,ptu_params);  
  /* Close bumblebee */
  BB_release(handle,&camera);
  /* Close canon */
  if(activate_camera) capture_stop();
  exit(0);
}


/* the SIGHUP handler for restarting */
void sig_restart(int arg)
{
}

Message *read_client_command(int fd) {
  unsigned char buf[256];
  int numread=read(fd,buf,255);
  Message *msg;
  msg=(Message *)calloc(1,sizeof(Message));
  msg->size=numread;
  msg->data=(unsigned char*)calloc(numread,sizeof(char));
  memcpy(msg->data,buf,numread);
  return msg;
}

void parse_client_command(Message *msg,PTU46_data *pars) {
  Message *reply_msg,*img_msg;
/*   char filename[MSG_BUFSIZE]; */
  float odometry[3];
  float xSpeed, ySpeed, yawSpeed;
  float *ranges; /* laser ranges */
  float *points; /* laser points */
  pthread_t motion_thread;
  float pos_pan,pos_tilt;
  float max_pan,min_pan,max_tilt,min_tilt;
  int width,height,size,zoom,*bufptr;
  int fread_bytes;
  int senderr=0;
  char *img;
  camera_packet *cpacket;
  reply_msg = (Message *)calloc(1,sizeof(Message));
  reply_msg->data=(unsigned char*)calloc(MSG_BUFSIZE,sizeof(char));
  switch(msg->data[0]) {
  case 'p':
    printf("Set PTU position request.\n");
    // Parse set request
    sscanf((char *)(msg->data),"p%g;%g\n",&pos_pan,&pos_tilt);
    // Move PTU to new position
    if(PTU46_SetPanTilt(serial_fd,serial_buffer,pars,pos_pan,pos_tilt,
			PTU46_BLOCK)
       ==-1) printf("Error detected.\n");
    // Read current position
    pos_pan  = PTU46_GetPos(serial_fd,serial_buffer,pars,PTU46_PAN);
    pos_tilt = PTU46_GetPos(serial_fd,serial_buffer,pars,PTU46_TILT);
    // Send current position back
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"POS:%g:%g\n",
	     pos_pan,pos_tilt);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;
  case 'r':
    printf("Get PTU position request.\n");
    // Read current position
    pos_pan  = PTU46_GetPos(serial_fd,serial_buffer,pars,PTU46_PAN);
    pos_tilt = PTU46_GetPos(serial_fd,serial_buffer,pars,PTU46_TILT);
    // Send current position back
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"POS:%g:%g\n",
	     pos_pan,pos_tilt);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;
  case 'o':
    printf("Get odometry request.\n");
    // Read odometry
    PB_GetOdometry(odometry);
    // Send odometry back
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"ODOM:%g:%g:%g\n",
	     odometry[0],odometry[1],odometry[2]);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;
  case 'v':
    printf("Get speeds request.\n");
    // Read Speeds
    PB_GetSpeeds(&xSpeed, &ySpeed, &yawSpeed);
    // Send speeds back
    printf("speed: %g %g %g\n",xSpeed,ySpeed,yawSpeed);
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"SPEEDS:%g:%g:%g\n",
	     xSpeed,ySpeed,yawSpeed);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;
  case 's':
  {
    printf("Get laser request.\n");
    // Read laser
    int rgsSize = PB_GetLaserRanges(ranges);
    int ptsSize = PB_GetLaserPoints(points);

    int arrSize = rgsSize+ptsSize;
    float *readings = (float *)calloc(arrSize,sizeof(float));
    memcpy(readings,ranges,sizeof(float)*rgsSize);
    memcpy(readings+rgsSize,points,sizeof(float)*ptsSize);

    static const int MAX_SIZE = 12000; // maximum 12000 bytes
    char data[MAX_SIZE];

    // Write all the floats into the string of data
    int remaining = MAX_SIZE;
    char *cur = data;
    for(int i=0; i < arrSize; ++i)
    {
      cur += snprintf(cur,remaining,
             "%g ", readings[i]);
      remaining = MAX_SIZE-(cur-data);
    }

    int numBytes = strlen(data);
    // First tell the client how many readings we'll be sending
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"LAS_SIZE:%d %d %d\n",
	     rgsSize,ptsSize,numBytes);
    reply_msg->size=MSG_BUFSIZE;//strlen((char *)reply_msg->data);
    Message_send(reply_msg,client_connection);

    // Now send the readings
    Message *msg = Message_new(data);
    Message_send(msg,client_connection);

    Message_free(reply_msg);
    Message_free(msg);
    
    free(ranges);
    free(points);
    free(readings);
    fprintf(stderr,"parse_client_command: done.\n");
    fflush(stderr);
    return;
  }
  case 'm':
  {
    printf("Move robot request.\n");
    float *speedLims = (float *)calloc(2,sizeof(float));

    // Parse move request
    sscanf((char *)(msg->data),"m%g:%g\n",
           speedLims,speedLims+1);

    pthread_create(&motion_thread, (pthread_attr_t *) NULL, PB_motion, speedLims);
    break; 
  }
  case 'n':
  {
    printf("Move robot to request.\n");

    float *odom = (float *)calloc(3,sizeof(float));

    // Parse move to request
    sscanf((char *)(msg->data),"n%g;%g;%g\n",
           odom,odom+1,odom+2);

    pthread_create(&motion_thread, (pthread_attr_t *) NULL, PB_moveTo, odom);
    break; 
  }
  case 't':
    printf("Turn robot request.\n");
    sscanf((char *)(msg->data),"t%g\n",
           odometry);
    PB_turn(odometry[0]);
    break; 	
  case 'd':
    printf("Get PTU resolution request.\n");
    // Read resolution
    pos_pan  = PTU46_GetRes(serial_fd,serial_buffer,PTU46_PAN);
    pos_tilt = PTU46_GetRes(serial_fd,serial_buffer,PTU46_TILT);
    // Send resolution back
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"RES:%g:%g\n",
	     pos_pan,pos_tilt);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;   
  case 'l':
    printf("Get PTU limits request.\n");
    // Read limits
    min_pan  = PTU46_GetLimit(serial_fd,serial_buffer,pars,PTU46_PAN,'n');
    max_pan  = PTU46_GetLimit(serial_fd,serial_buffer,pars,PTU46_PAN,'x');
    min_tilt = PTU46_GetLimit(serial_fd,serial_buffer,pars,PTU46_TILT,'n');
    max_tilt = PTU46_GetLimit(serial_fd,serial_buffer,pars,PTU46_TILT,'x');
    // Send resolution back
    snprintf((char *)reply_msg->data,MSG_BUFSIZE,"LIM:%g:%g:%g:%g\n",
	     min_pan,max_pan,min_tilt,max_tilt);
    reply_msg->size=strlen((char *)reply_msg->data);
    break;      
  case 'c':
    printf("Get bumblebee frame request.\n");

    printf("BB:single_capture\n"); 
    if(BB_single_capture(handle,&camera)==-1) return;
    printf("Release buffer.\n");
    BB_buffer_free(&camera);
    printf("BB:single_capture\n"); 
    if(BB_single_capture(handle,&camera)==-1) return;
 
    printf("Compose packet\n");
    cpacket = (camera_packet *)calloc(1,sizeof(camera_packet));
    cpacket->width=camera.frame_width;
    cpacket->height=camera.frame_height;
/*     cpacket->size=camera.dma_frame_size; */
    cpacket->size=camera.quadlets_per_frame*4;
    cpacket->data=(char *)camera.capture_buffer;
    img_msg = Message_camera_packet(cpacket);
    free(cpacket);

    printf("Send package\n");
    senderr=Message_send(img_msg,client_connection);
    if(senderr==-1) {
      fprintf(stderr,"Closing connection.\n");
      if (client_connection>0 && client_connection!=client_socket)
	close(client_connection);
      if (client_socket>0) close(client_socket);
      client_socket=client_connection=-1;
    }
    Message_free(img_msg);
    printf("Release buffer.\n");
    BB_buffer_free(&camera);
    break;
  case 'x':
    printf("Capture still image request.\n");
    capture_grab("lastframe.jpg");
    fread_bytes=capture_loadimage("lastframe.jpg",&img);
    if(fread_bytes<0) printf("Ouch!\n");
    else {
      printf("Compose packet\n");
      cpacket = (camera_packet *)calloc(1,sizeof(camera_packet));
      cpacket->width=fread_bytes/64; /* Ugly fix to fit into short */
      cpacket->height=64;
      cpacket->size=fread_bytes;
      cpacket->data=(char *)img;
      img_msg = Message_camera_packet(cpacket);
      free(cpacket);
    }
    capture_freeimage(img);

    printf("Send package\n");
    senderr=Message_send(img_msg,client_connection);
    if(senderr==-1) {
      fprintf(stderr,"Closing connection.\n");
      if (client_connection>0 && client_connection!=client_socket)
	close(client_connection);
      if (client_socket>0) close(client_socket);
      client_socket=client_connection=-1;
    }
    Message_free(img_msg);
    break;
  case 'z':
    printf("Set zoom request.\n");
    /* Parse zoom request */
    sscanf((char *)(msg->data),"z%d\n",&zoom);
    capture_setzoom(zoom);
    break;
  case 'q':
    fprintf(stderr,"Close connection request.\n");
    fflush(stderr);
    if (client_connection>0 && client_connection!=client_socket)
      close(client_connection);
    if (client_socket>0) close(client_socket);
    client_socket=client_connection=-1;
    break;
  }
  if(reply_msg->size) {
    Message_send(reply_msg,client_connection);
  }
  Message_free(reply_msg);
  fprintf(stderr,"parse_client_command: done.\n");
  fflush(stderr);
}

int main(int argc, char ** argv) {
  /* PTU46_mini variables */
  char *port,*cmd;
  float pos_pan,pos_tilt;
  /* Other stuff */
  int infofl=1;       /* Print info if set */
  char filename[50];
  Message *request;
  dc1394_miscinfo misc_info;

  handleOptions(argc,argv);

  /* Setup for Canon camera */
  if(activate_camera) capture_start();
  else printf("Skipping camera initialisation.\n");

  /*Setup for Powerbot */
  if(activate_robot){
	if( PB_init() !=1 )
	{		
		printf("Error initializing robot.\n");
		exit(1);
	}
	else
	{
		printf("Robot activated.\n");
	}
  }
  else 
  {	
	printf("Skipping robot initialisation.\n");
  }

  if( activate_ptu )
  {
    /* Setup for PTU46 */
    port=strdup(PTU46_PORTNAME);
    serial_fd=PTU46_Init(port,&serial_buffer,&ptu_params);
    /* Read resolution and limits for error checking */
    PTU46_ReadParams(serial_fd,serial_buffer,ptu_params);
    PTU46_PrintParams(ptu_params);
  }

  if( activate_bb && strcmp(dma_dev_name,"skip")!=0) {
    /* Camera connect and setup code */

    /* Find first camera on 1394 bus */
    if(BB_find_camera(&handle,&camera,infofl)==-1) return -1;

    if(BB_setup_capture(handle,&camera,dma_dev_name)==-1) return -1;

    /* Read current video mode */
    if(dc1394_get_camera_misc_info(handle,camera.node,&misc_info)
       !=DC1394_SUCCESS) {
      fprintf(stderr,"Could not get camera misc info\n");
      dc1394_destroy_handle(handle);
      return -1;
    }
    if(misc_info.format==FORMAT_SVGA_NONCOMPRESSED_1) {
      printf("BB-HICOL detected.\n");
      /* Hardcoded whitebalance setting for BB-HICOL40 */
      if(BB_setwhitebalance(handle,&camera,50,33)==-1) return -1;
    }
    if(misc_info.format==FORMAT_VGA_NONCOMPRESSED) {
      printf("BB-COL detected.\n");
      /* Hardcoded whitebalance setting for BB-COL40 */
      if(BB_setwhitebalance(handle,&camera,53,41)==-1) return -1;
    }

    if(BB_start_iso_transmission(handle,&camera)==-1) return -1;
  } else {
    fprintf(stderr,"Warning: Skipping camera setup.\n");
  }

  signal(SIGHUP, sig_restart);
  signal(SIGTERM, sig_shutdown);

  while (1) {
    client_socket=establish_local_socket(client_port, &client_skname);
    fprintf(stderr,"client_socket %d %d\n", client_port, client_socket);
    client_connection=client_socket;
    while (client_connection !=-1) {
      struct pollfd pfd[1];
      pfd[0].fd=client_connection;
      pfd[0].events=POLLIN;
      int rval;
      while ((rval=poll(pfd,1,5000))==0) {
	fprintf(stderr,"Poll...\n");
      }
      if (rval==-1) {
	perror("poll");
	sig_shutdown(0);
	break;
      } 
      if (pfd[0].revents==POLLIN) {
	fprintf(stderr,"client_connection=%d client_socket\n",client_connection,client_socket);
	fflush(stderr);
	if (client_connection==client_socket) {
	  client_connection=accept_connection(client_socket);
	  fprintf(stderr,"Client connection: %d\n", client_connection);
	} else {
	  request=read_fixlength_packet(client_connection);
	  if (request->size) {
	    Message_printf("received:",request);
	    parse_client_command(request,ptu_params);
            printf("done parse_client_command\n");
	    Message_free(request);
	  } else {
	    fprintf(stderr,"Here.\n");
	    fflush(stderr);
	    sig_shutdown(0);
	  }
	}
      }
    }
    fprintf(stderr,"Client disconnected. Reopening connection.\n");
  }
  exit(0);
}


