function [ poseMap ] = buildPoseMap ( parents )
% [ poseMap ] = buildPoseMap ( parents )
%
% For each potential destination on the occupancy grid, we'd like to estimate
% what the robot's position 'would' be if it were to navigate to this position
% from where it currently stands.
% 
% Since we already have a map showing each cell's 'parent' (the cell that would
% precede it in a path) we can simple trace backward a few parents and get a
% rough idea of the robot's angle of approach to any particular cell.
% 

height = size(parents,1);
width = size(parents,2);
TRACEBACK = 8;
poseMap = zeros(height,width);

for i = 1:width
    for j = 1:height
        
        % Start at given cell (x,y) = (i,j):
        x = i;
        y = j;
        if (parents(y,x,1)==0 || parents(y,x,2)==0)
            continue;
        end;
        
        % Trace backward by a number of parents:
        for k = 1:TRACEBACK
            px = parents(y,x,1);
            py = parents(y,x,2);
            if (px == 0 || py == 0)
                break;
            end;
            x = px;
            y = py;
        end;
        
        % Now we've jumped a few pixels away. Let's calculate the yaw angle
        % between the origin location and its ancestor. 
        % The result is undefined if dx=dy=0.
        dx = double(x - px);
        dy = double(y - py);
        if (dy == 0)
            halfpi = pi/2;
            theta = sign(-dx) * halfpi + halfpi;
        else
            theta = atan(dx/dy);
        end;
        poseMap(j,i) = theta;
        
    end;
end;

