function [objectiveMap varargout ] = buildScoreMap2 ( worldState, cameraPars, maxViewDist )
%   This builds an alternative objective function for the planner that is
% multiplicative across camera shots within the same bin, but that is
% additive across overlapping bins. The benefits of this approach are that
% that it can be copied and modified easily after a picture has been taken.
%
%   We seek an optimization of score over time over all possible waypoints on 
% the map. We define score as the sum of all the bin scores that are viewable 
% from a particular pose. The time is then that required to travel to the 
% waypoint, set up for a camera shot, then take as many photos as we have 
% viewable objects from that waypoint. This gives us a total time that we spend
% before moving to a new waypoint.
%
% worldState: the occupancy grid, the distance map, the objects, and the pose.
% cameraPars: the camera height and the ptu angular limits.
% maxViewDist: the farthest horizontal distance at which we're willing to
% take a shot.
%
% Scoring Constants:
% alpha: the score of a bin is the result of raising alpha to the negative of the
% bin's count: score = alpha^(-count). Larger values of alpha incur greater
% penalties on the score for a given bin for each subsequent picture.
alpha = 1.5;

% sigma: this is the coefficient to the Gaussian smoothing function that is
% convolved with the bins.
sigma = 2;

% Unpack the parameters:
oGrid = worldState.oGrid;
objs = worldState.objs;
dists = worldState.dists;

% Static parameters, guesstimated for now - derive empirically later:
% Costs are given in seconds.
overheadCost = 4;
pictureCost = 3;
velocity = 25; % I think this is roughly cm/s

% This is the map of the sums of "shot" scores.
scoreMap = zeros(size(oGrid));

% This is the map of the number of shots available.
shotsMap = zeros(size(oGrid));
objPositions = [];

% We might only need to return the scores, not the objs
if nargout > 1
    varargout{1} = [];
end;

% If there are no objects, return:
nObs = size(objs,1);
if (nObs == 0)
    objectiveMap = zeros(size(oGrid));
    return;
end

% Initialize grid:
[rows,cols] = size(oGrid);
cl=-maxViewDist:maxViewDist;
[x y] = meshgrid(cl,cl);

% Calculate the x2y2 matrix - this is used as the basis for the halos
% around each salient object.
x2y2 = x.^2+y.^2;

% 1D Gaussian filter:
gf = exp(-((-sigma*3:sigma*3).^2)/(2*sigma^2));
gf = gf/sum(gf);

% Convert SLAM coordinates to grid positions:
gridPositions = performSlam('get_gridpos_from_vecpos',objs(:,1:2));

% Blur bins and get rid of invalid objects:
scoredObjects = zeros(nObs(1),40);
minHorzDists = zeros(nObs(1),1);
nValid = 0;
for j=1:nObs
    
    object = objs(j,:);
    
    % Object with unknown location, skip it:
    if all(object(1:3) == 0)
        continue;
    else
        nValid = nValid + 1;
    end;
    
    % Figure the minimum distance we need to be from each object:
    minHorzDists(nValid) = calcMinHorizonDistance(double(object(3)), cameraPars);
    
    % Each bin contains a count of the number of photos taken from that
    % angle. Use these counts to generate a score for that angle.
    viewScores = alpha .^ (-object(5:40));
    
    % Wraparound entries at the edges:
    wrappedScores = [viewScores(37-sigma*3:36) viewScores viewScores(1:sigma*3)];
    blurredScores = conv2(wrappedScores,gf,'valid');
    
    % Blurred frequencies:
    object(5:40) = blurredScores;
    scoredObjects(nValid,:) = object;
    
end;

% objs = temp(1:nValid,:);
minHorzDists = minHorzDists(1:nValid);
nObs = size(scoredObjects,1);

% Rings around objects initially start with a prior score of 1:
for j=1:nValid
    [ir ic] = find( x2y2<maxViewDist^2 & x2y2>minHorzDists(j)^2 );
    ir = (ir-maxViewDist)+gridPositions(j,1);
    ic = (ic-maxViewDist)+gridPositions(j,2);
    validl = find((ir>0)&(ir<rows)&(ic>0)&(ic<cols));
    il = sub2ind([rows cols],ir(validl),ic(validl));
    scoreMap(il) = 1;
end;

% Perform a sequence of conical "sweeps" around each object:
for angle=1:36
    
    maxRange = (angle-0.001) * pi/18;
    minRange = (maxRange) - pi/18;
    
    for j=1:nValid
        
        % Find all the cells that are within acceptable viewing range:
        [ir ic] = find( x2y2<maxViewDist^2 & x2y2>minHorzDists(j)^2 & mod(atan2(y,x),2*pi)>minRange & mod(atan2(y,x),2*pi)<maxRange);
        
        % Convert these cell coordinates into worldgrid coordinates:
        ir=(ir-maxViewDist)+gridPositions(j,1);
        ic=(ic-maxViewDist)+gridPositions(j,2);
        
        % Find which of these entries are off the worldgrid altogether:
        validl=find((ir>0)&(ir<rows)&(ic>0)&(ic<cols));
        
        % Calc linear indices into the score map using the row,col indices
        il = sub2ind([rows cols],ir(validl),ic(validl));
        
        % Add in the score for this conical sweep
        scoreMap(il) = scoreMap(il) + scoredObjects(j,4+angle);
        shotsMap(il) = shotsMap(il) + 1;
    end;
end;

% Now calculate the travel time, overhead cost, and "shots" time for each waypoint.
timeMap = (double(dists) ./ velocity) + (shotsMap .* pictureCost) + overheadCost;

% Now calculate the objective function as the ratio of score to cost:
objectiveMap = scoreMap ./ timeMap;

% Invalidate obstacles
if nargout > 1
    varargout{1} = scoredObjects(:,1:2);
end;
