function [ sampleRewards sampleScores ] = evaluateSamples ( samples, worldState, mappingPars, planningPars, configPars, photoRewardMap)
% [ sampleRewards sampleScores ] = evaluateSamples 
%   ( samples, worldState, mappingPars, planningPars, photoRewardMap, frontierRewardMap )
% 
%   This function accepts a set of samples of a given size and generates a
% column of scores of that size, one for each sample row. 
% 
% sampleRewards(:,1) are the frontier rewards
% sampleRewards(:,2) are the coverage rewards
% sampleRewards(:,3) are the photograph rewards
% 
% sampleScores(:,1) is the sum of all three scores, divided by the distance vector
% 
nNumSamples = size(samples, 1);
sampleRewards = zeros(nNumSamples,3);
sampleScores = zeros(nNumSamples,1);
oGrid = worldState.oGrid;
dimGrid = size(oGrid);
[ mx my ] = meshgrid(1:dimGrid(2), 1:dimGrid(1));

% create a dense map of photograph reward on the basis of zoomed photo potential:
photoWeight = planningPars.photographImportance;
photoReward = photoWeight .* photoRewardMap;

for s=1:nNumSamples
    
    % grab the details out of the sample. treat these as SLAM coordinates:
    i = samples(s,1);
    j = samples(s,2);
    theta = samples(s,3);
    idx = sub2ind(size(oGrid),j,i);
    
    % create a wedge mask of the area we'd like to cover:
      % (?fixme: this is the slow way of doing things, since we're essentially 
      %  copying a bunch of arrays around just to make a wedge shape ?)
    maxViewDist = planningPars.maxViewDist;
    minViewDist = planningPars.minViewDist;
    wedgeHalfAngle = pi/6;
    dx = i-mx; 
    dy = j-my;
    x2y2 = dx .* dx + dy .* dy;
    wedgeMask = find(x2y2 <= maxViewDist^2 & x2y2 >= minViewDist^2 & (mod(atan2(my,mx)-theta, 2*pi) < wedgeHalfAngle));
    
    % calculate the amount of coverage we've obtained over this wedge
    % (?hack: for now I'm taking the number of coverage shots and averaging it
    % over the size of a coverage wedge to get the 'expected' number of coverage
    % shots for a given cell seen from this location.)
    coverageReward = sum(mappingPars.coverageMap(wedgeMask));
    stdWedgeSize = (wedgeHalfAngle / pi) * pi * (maxViewDist^2 - minViewDist^2);
    coverageReward = coverageReward ./ stdWedgeSize;
    base = planningPars.coverageScoreDegradation;
    coverageReward = (base).^(-coverageReward);
    
    % total reward is a linear combination of frontier reward, coverage reward,
    % and photo reward:
    sampleRewards(s,1) = coverageReward;
    sampleRewards(s,2) = photoReward(idx);
    
    % now derive sample score by taking distance into account:
    % (?hack: avoid div by zero by adding a constant. later we can change this to reflect
    %         time spent taking photos, turning, etc)
    sampleScores(s) = sum(sampleRewards(s),2) ./ (worldState.dists(idx) + 5);
    if configPars.PRINTLEVEL >= 4
%        fprintf('[evaluateSamples.m] Debug: sample(%d,:)=[%d %d] sampleRewards(%d,:)=[%d %d]\n', s, i, j, s, sampleRewards(1), sampleRewards(2));
    end;
    
end;
