function [ goal planningPars ] = generateGoal (worldState, cameraPars, configPars, mappingPars, planningPars)
% [ goal planningPars ] = generateGoal
%   (worldState, cameraPars, configPars, mappingPars, planningPars)
% 
%   Generates a goal object, governing where we move to next. Our approach uses 
% a scoring function that comprises frontier exploration, coverage pictures, and 
% photographs. The scoring function will be defined up to a set of parameters 
% governing the relative weights of the rewards for exploration, coverage, and 
% photograph collection. 
% 
%   Strategy: 
%   
% Define a dense reward map over the o-grid that indicates the 
% frontier-exploration rewards. 
%
% Define a dense map over the o-grid that indicates previous camera coverage.
% 
% Define a dense map over the o-grid that indicates the indexes to other
% objects. Each cell will contain a structure that contains a list of objects
% present at that cell.
% 
% Sample from the pose space (x,y,yaw). At each sample, compute the reward
% achieved by exploring frontiers at this location in terms of
% frontier-exploration, coverage, and zoom photos.
% 

global STATE_NEW;               
global STATE_INITIAL_TURN;      
global STATE_TRAVERSAL;         
global STATE_FINAL_TURN;        
global STATE_PHOTOGRAPH;        
global STATE_FINISHED;          
global STATE_ABORTED;           

goal = [];
MAX_COST = planningPars.MAX_COST;

% Create a dense map over the oGrid indicating reward for exploring frontiers
%[ frontierRewardMap numFrontierCells ] = calcFrontierRewards ( worldState.oGrid, planningPars );
if ~planningPars.exploreDone
    
    % Explore? Check total length of frontiers / length of perimeter
    [frontiers lengths] = getFrontiers(worldState.oGrid,floor(255*planningPars.OBSTACLE_THRESHOLD),20,planningPars.explored);
    nFrontiers = length(frontiers);
    if nFrontiers > 0
        [rows cols] = find(worldState.obstacle_map);
        perimeter = (max(rows)-min(rows) + max(cols)-min(cols))*2;
        
        % Look at length of all frontiers remaining:
        if sum(lengths) / perimeter < 0.1
            
            % Good enough, let's go into picture mode.
            fprintf('[generateGoal.m]: Frontiers are tiny compared to perimeter. Time to take some pictures.\n');
            planningPars.exploreDone = true;
            
        else
            
            % Create a frontier goal:
            minScore = -inf;
            [height width]=size(worldState.oGrid);
            for i=1:length(frontiers)
                
                frontier = frontiers{i};
                % Find the "best" cell of this frontier
                center = frontier(1,:);
                
                if norm(center-worldState.rGridPos,2) < 10
                    continue;
                end;
                minCost = MAX_COST;
                min_dist = inf;
                for j=1:worldState.nFreeIndices
                    index = worldState.free_indices(j);
                    thisRow = mod(index-1,height)+1;
                    thisCol = floor(index/height)+1;
                    thisDist = norm( [ thisRow thisCol ] - center, 2 );
                    
                    cost = worldState.dists(thisRow,thisCol);
                    if( thisDist < min_dist && cost < MAX_COST)
                        min_dist = thisDist;
                        minCost = double(cost);
                        bestFCell = [ thisRow thisCol ];
                    end;
                end;
                
                % Choose this frontier if it has the highest score
                if minCost < MAX_COST
                    score = lengths(i)/minCost;
                    if score > minScore
                        minScore = score;
                        goal.frontier = frontier;
                        goal.frontiers = frontiers;
                        bestFCenter = center;
                        bestGoal = bestFCell;
                    end;
                end;
            end;
            
            if minScore > -inf
                % Successfully found a path to a frontier
                goal.path = buildGridPath(configPars,worldState.parents,worldState.rGridPos,bestGoal);
                goal.type = 'e';
                % Angle to face center of frontier from goal
                y = bestGoal(1)-bestFCenter(1); % high grid row is low y in odometry
                x = bestFCenter(2)-bestGoal(2);
                goal.initialYaw = mod(atan2(y,x),2*pi);
            else
                fprintf('[generateGoal.m]: Frontiers remain, but no feasible path exists. Time to take some pictures.\n');
                planningPars.exploreDone = true;
            end;
            
        end;
        
    else
        fprintf('[generateGoal.m]: Frontiers are completely gone. Time to take some pictures.\n');
        planningPars.exploreDone = true;
    end;
end;

if planningPars.exploreDone
    
    % call generateGoalTable here until all tables are visited
    % uncomment for table integration
    % if ~planningPars.tablesDones
    %   [ goal planningPars ] = generateGoalTable( worldState, cameraPars, configPars, mappingPars, planningPars );
    % else
    
    % If we got here, we're done exploring frontiers and we should take some
    % pictures.
    
    % Create a dense map over the oGrid indicating rewards for taking photographs.
    [ photoRewardMap ] = buildScoreMap2 ( worldState, cameraPars, planningPars.maxViewDist );
    
    % Search over the oGrid _and_ the viewing angles to find the reward for
    % achieving coverage from a given pose. 
    % fixme: don't pass in the dense maps into this function as it's too slow.
    % instead, sample only over the coverage maps and return the result out
    % here.
    [ candidate ] = sampleFromGoalSpace ( worldState, mappingPars, planningPars, configPars, photoRewardMap );
  
    % Take this sample and turn it into a goal structure.
    destination = [ candidate(2) candidate(1) ]; % switch to row,col format
    goal.path = buildGridPath ( configPars, worldState.parents, worldState.rGridPos, destination );
    goal.type = 'o';
    
    % Calculate initial yaw
    diff = performSlam('get_vecpos_from_gridpos',goal.path(1,:)) - worldState.rPose(1:2);
    goal.initialYaw = thetapp(atan2(diff(2),diff(1)));
    goal.finalYaw = candidate(3);
end;

if isempty(goal)
    return;
end;

goal.yaw = goal.initialYaw;
goal.step = 0;
goal.startTime = toc;
goal.timeLimit = toc + 8 + 2 * size(goal.path,1);
goal.bestDistRemain = inf;
goal.lastSlamTime = goal.startTime;

% goal.state has possible state values:
% STATE_NEW := new goal, no initial turn cmd given
% STATE_INITIAL_TURN := turning bot toward initial yaw
% STATE_TRAVERSAL := traversing waypoints
% STATE_FINAL_TURN := turning bot toward final yaw
% STATE_PHOTOGRAPH := take pictures (happens in last frame)
% STATE_FINISHED := finished 
goal.state = STATE_NEW;

planningPars.numGoals = planningPars.numGoals + 1;

if configPars.DISPLAY
    DisplayFrame;
end;

%end %uncomment for table integration
