function [ goal planningPars ] = generateGoalTable (worldState, cameraPars, configPars, mappingPars, planningPars)
% [ goal planningPars ] = generateGoal
%   (worldState, cameraPars, configPars, mappingPars, planningPars)
% 
%   Generates a goal object, governing where we move to next. Our approach uses 
% a scoring function that comprises frontier exploration, coverage pictures, and 
% photographs. The scoring function will be defined up to a set of parameters 
% governing the relative weights of the rewards for exploration, coverage, and 
% photograph collection. 
% 
%   Strategy: 
%   
% Define a dense reward map over the o-grid that indicates the 
% frontier-exploration rewards. 
%
% Define a dense map over the o-grid that indicates previous camera coverage.
% 
% Define a dense map over the o-grid that indicates the indexes to other
% objects. Each cell will contain a structure that contains a list of objects
% present at that cell.
% 
% Sample from the pose space (x,y,yaw). At each sample, compute the reward
% achieved by exploring frontiers at this location in terms of
% frontier-exploration, coverage, and zoom photos.
% 

global STATE_NEW;               
global STATE_INITIAL_TURN;      
global STATE_TRAVERSAL;         
global STATE_FINAL_TURN;        
global STATE_PHOTOGRAPH;        
global STATE_FINISHED;          
global STATE_ABORTED;           

goal = [];
MAX_COST = planningPars.MAX_COST;

% Create a dense map over the oGrid indicating reward for exploring frontiers
%[ frontierRewardMap numFrontierCells ] = calcFrontierRewards ( worldState.oGrid, planningPars );

obstacle_map = worldState.obstacle_map;
figure(2); 
%imshow(obstacle_map);
hold on;
%find connected components corresponding to tables
[Tables num] =  bwlabel(obstacle_map);

fprintf('Found %d tables!\n', num-1); 
goal.type = 'x';
goal.path = [];
goal.step=0;

if num<2 
    return;
end;

if(size(planningPars.visitedTableRegions,1)>0)
    numVisitedTables = size(unique(planningPars.visitedTableRegions(:,1:2),'rows'),1);
    fprintf('[generateGoalTable.m]: Photographed at %d tables at least once!\n', numVisitedTables);
    fprintf('[generateGoalTable.m]: Photographed at %d table regions!\n', size(planningPars.visitedTableRegions,1));
    if(numVisitedTables==num-1)
        planningPars.tablesDone = true;
        fprintf('[generateGoalTable.m]: Photographed at all tables!\n');
        %return;
    end
end
        

minScore = inf;
bestTCenter = [];
bestCenter = []; 
bestGoal = [];

[height width]=size(worldState.oGrid);
for i=1:num
    %pixels of ith table
    [r c] = find(Tables==i);
    if min(r)==1 && min(c)==1 && max(r)==height && max(c)==width
        continue;
    end
    tableCenter = [floor(median(r)) floor(median(c))];

    % Find the closest point on the table to the robot
    diff = repmat(worldState.rGridPos,length(r),1)-[r c];
    dist =sqrt(diff(:,1).^2 + diff(:,2).^2);
    [closestDist index] = min(dist);
    center = [r(index) c(index)];

    if( size(planningPars.visitedTableRegions ,1) > 0 )

        %check if we've already visited this table region
        diff = repmat(center,size(planningPars.visitedTableRegions,1),1)-planningPars.visitedTableRegions(:,3:4);
        dist =sqrt(diff(:,1).^2 + diff(:,2).^2);
        [closestDist index] = min(dist);
        if closestDist < planningPars.TABLE_DISTANCE
          continue;
        end
    end;

    minCost = MAX_COST;
    min_dist = inf;

    for j=1:worldState.nFreeIndices
        index = worldState.free_indices(j);
        thisRow = mod(index-1,height)+1;
        thisCol = floor(index/height)+1;
        thisDist = norm( [ thisRow thisCol ] - center, 2 );
        %thisDist = worldState.dists(center(1),center(2));
        cost = worldState.dists(thisRow,thisCol);
        if( thisDist < min_dist && cost < MAX_COST)
            min_dist = thisDist;
            minCost = double(cost);
            bestTCell = [ thisRow thisCol ];
        end;
    end;
    %bestTCell
    % Choose this frontier if it has the highest score
    score = minCost;
    if score < minScore
            minScore = score;
            bestTCenter = tableCenter;
            bestCenter = center;
            bestGoal = bestTCell;
    end;
end;
%plot(bestGoal(1), bestGoal(2));
if minScore < inf
     % Successfully found a path to a table
    goal.path = buildGridPath(configPars,worldState.parents,worldState.rGridPos,bestGoal);
    goal.path = bestGoal; %for testing purposes only
    goal.obsLoc = bestTCenter;
    goal.type = 't';
    
    diff = performSlam('get_vecpos_from_gridpos',goal.path(1,:)) - worldState.rPose(1:2);
    goal.initialYaw = thetapp(atan2(diff(2),diff(1)));
    
    % Angle to face center of frontier from goal
    %y = bestGoal(1)-bestTCenter(1); % high grid row is low y in odometry
    %x = bestTCenter(2)-bestGoal(2);
    %goal.initialYaw = mod(atan2(y,x),2*pi);
else
    if ~(num==1)
        fprintf('[generateGoal.m]: Tables remain, but no more feasible/undiscovered paths exist. Random walk to discover other table regions.\n');
        %planningPars.tablesDone = true;
    end    
    return;
    
end;

goal.step = 0;
        
if isempty(goal)
    return;
end;

goal.yaw = goal.initialYaw;
diff = performSlam('get_vecpos_from_gridpos',bestTCenter) - performSlam('get_vecpos_from_gridpos',goal.path(end,:));
goal.finalYaw = thetapp(atan2(diff(2),diff(1)));
goal.step = 0;
goal.startTime = toc;
goal.timeLimit = toc + 60;%toc + 8 + 2 * size(goal.path,1);
goal.bestDistRemain = inf;

% goal.state has possible state values:
% STATE_NEW := new goal, no initial turn cmd given
% STATE_INITIAL_TURN := turning bot toward initial yaw
% STATE_TRAVERSAL := traversing waypoints
% STATE_FINAL_TURN := turning bot toward final yaw
% STATE_PHOTOGRAPH := take pictures (happens in last frame)
% STATE_FINISHED := finished 
goal.state = STATE_NEW;
planningPars.numGoals = planningPars.numGoals + 1;
if( size(planningPars.visitedTableRegions ,1) > 0 )

    %check if we've already visited this table
    diff = repmat(center,size(planningPars.visitedTableRegions,1),1)-planningPars.visitedTableRegions(:,1:2);
    dist =sqrt(diff(:,1).^2 + diff(:,2).^2);
    [closestDist index] = min(dist);
    %To ignore any drift errors - we're at the same table, so don't record
    %a different table center
    if closestDist < 5
      bestTCenter = planningPars.visitedTableRegions(index,1:2);
    end
end;

planningPars.lastTableRegion = [bestTCenter bestCenter]; 
plot(bestTCenter(2), bestTCenter(1), 'r+');
plot(bestCenter(2), bestCenter(1), 'g+');
plot(bestGoal(2), bestGoal(1), 'b+');
if configPars.DISPLAY
   DisplayFrame;
end;


