function [ worldState ] = generateWorldState ( cameraPars, gazePars, mappingPars, configPars, planningPars, varargin )
%
% [worldState] = GENERATEWORLDSTATE 
%   ( lastWorldState, cameraPars, gazePars, mappingPars, configPars, planningPars )
%
% Generates an encapsulated 'worldState' object that contains all the
% information about the robot's and the world's current state.
%
% Returns: worldState structure with fields:
%  odom
%  speeds
%  rPose: the robot's position in SLAM coordinates
%  rGridPos: the robot's position in the current grid coordinates
%  obstacleMap: a dilation of the oGrid to exclude/avoid space around the
%    objects.
% 

% Stamp this worldState with the clock
worldState.time = toc;

% Grab the odometry info from the powerbot, and grab all the
% relevant pose information from the SLAM module.
worldState.speeds   = powerbot('getspeeds');
[ worldState.odom worldState.rPose worldState.rGridPos ] = doSlam(1,1);

% These are each 3x3 transformation matrices that allow us to take any pair of
% x,y coordinates and transform them as a (x,y,1) homogeneous tuple into another
% space. The two spaces we are concerned with are grid coordinates (2d cells,
% ints) and 'world' coordinates (2d real-valued).
% gridTransform converts world coords to grid coords
% worldTransform converts grid coords to world coords
%worldState.gridTransform  = performSlam('get_gridtransform');
%worldState.worldTransform = performSlam('get_worldtransform');

worldState.oGrid    = performSlam('get_map');

% Object locations have to be updated regularly since the SLAM module will
% change coordinate frames as it updates the map.
mappingPars.objs    = updateObjectLocations ( mappingPars.objs );
worldState.objs     = mappingPars.objs;

% The obstacle_map is a dilation of the occupancy grid so that areas
% near obstacles are excluded from planning.
worldState.obstacle_map = ~im2bw(worldState.oGrid,planningPars.OBSTACLE_THRESHOLD);
worldState.obstacle_map = imdilate(worldState.obstacle_map,strel('disk',planningPars.dilationRadius));
worldState.obstacle_map = imerode(worldState.obstacle_map,strel('disk',planningPars.dilationRadius));

% Build map of costs to every location:
[dists parents nFreeIndices free_indices distMap] = buildPathMap(worldState.oGrid,worldState.rGridPos,planningPars.OBSTACLE_THRESHOLD);
worldState.dists        = dists;
worldState.parents      = parents;
worldState.nFreeIndices = nFreeIndices;
worldState.free_indices = free_indices;
worldState.distMap      = distMap;

