/*
 * Perform A-Star planning on the grid input. From start to goal
 *
 *
 * David Meger - June 2007
 */
/*===========================================================================
** System Includes
**=========================================================================*/
#include <stdio.h>
#include <stdlib.h>

/*===========================================================================
** Matlab Includes
**=========================================================================*/
#include "mex.h"

extern "C" {
#include "matrix.h"
#include "mexutil.h"
#include "sp.h"
}


#define use_heuristic 0		/* don't use a heuristic */
#define abort_at_dest 0		/* don't abort at destination */

#ifdef max
#undef max
#endif
#ifdef abs
#undef abs
#endif

#define max(a,b) ((a) > (b) ? (a) : (b))
#define abs(a)   ((a) < 0 ? -(a) : (a))

int SIZE_R;
int SIZE_C;
int **map;
int **dist;
int ***parent;
int startr, startc,endr,endc,rows_read,cols_read,pops,pushes;
int debug_level=0;
extern int max_memory;   

/* Two helpful macros to make the code more readable.  drow(row,i) and
 * dcol(col,i) will apply the delta table to the indicated row/column
 * to return a square around the one specified.
 * The delta table lists row/column offsets of the 8 squares which
 * surround a square (ie: up-left, up, up-right, left, etc).
 */
#define drow(row,i) (row+delta[i][0])
#define dcol(col,i) (col+delta[i][1])

/* delta is defined with the diagonals first and the "natural" directions
 * second which corresponds with our priority queue to give us "nicer"
 * looking paths.  It would probably be more appropriate to make the
 * algorithm simply pick a natural direction when possible.  I would
 * recommend that if actually using this algorithm with a real priority
 * queue since you will not be able to predict order in which the elements
 * will be returned.
 */
int delta[8][2] = { -1, -1, -1, 1, 1, -1, 1, 1, -1, 0, 0, 1, 1, 0, 0, -1 };

static int
h(int r1, int c1, int r2, int c2)
{
    if (use_heuristic) return(max(abs(r2 - r1), abs(c2 - c1)));
    return(0);
}

/**************************************************************************
 *
 * (void) build_path()
 *
 * This routine is entire algorithm:
 *
 * push the starting point onto the queue
 * while there is an element of least weight in the queue, say (r,c)
 *    foreach square around (r,c), say (r',c') where
 *	  dist[to (r,c)] + weight(r',c') < dist[to (r',c')]
 *    do
 *	  dist[to (r',c')] = dist[to (r,c)] + weight(r',c')
 *	  push (r',c') onto the queue
 **************************************************************************/
void
build_path(void)
{
PQUEUE * pq;
int      row, col;
int	 i;

    pq = pqueue_new();
    if (startr < 0 || startr >= rows_read || startc < 0 || startc >= cols_read)
    {
        mexPrintf("The start position was invalid.\n");
	return;
    }

    if( endr < 0 || endr >= rows_read || endc < 0 || endc >= cols_read )
    {
        mexPrintf("The end position was invalid.\n");
    }

    dist[startr][startc] = 0;
    parent[startr][startc][0] = -1;
    parent[startr][startc][1] = -1;
    pqueue_insert(pq, startr, startc, 0 + h(startr,startc,endr,endc));
    pushes++;

    while (pqueue_popmin(pq, &row, &col) == 0) {
	/* DEBUG INFORMATION */
	pops++;
	if (debug_level > 0)
    	    printf("\nSelected (row,col)=(%d,%d).  Adding:\n", row, col);

	/* See if we only wanted the one path */
	if (abort_at_dest && row == endr && col == endc) {
		/* STOP HERE */
		break;
	}
	/* FOREACH square adjacent to (row,col), call it drow(row,i), dcol(col,i) */
	for (i = 0; i < 8; i++) {
	    /* Check to make sure the square is on the map */
	    if (drow(row,i) < 0 || drow(row,i) >= rows_read ||
		dcol(col,i) < 0 || dcol(col,i) >= cols_read)
		continue;
        
        /* Check to see if the map value is max_int, if so, then skip the cell */
        
        if( map[drow(row,i)][dcol(col,i)] == INT_MAX )
            continue;

	    /* Check to see if we've found a shorter path */
	    if (dist[row][col] + map[drow(row,i)][dcol(col,i)] <
		dist[drow(row,i)][dcol(col,i)]) {
		/* DEBUG INFORMATION */
		if (debug_level > 0) {
		    printf("\t\t(%d,%d) Weight was/is: ", drow(row,i), dcol(col,i));
                    printf("%d/%d\n", dist[drow(row,i)][dcol(col,i)], map[drow(row,i)][dcol(col,i)] + dist[row][col]);
		}

		/* We have a shorter path, update the information */
		dist[drow(row,i)][dcol(col,i)] = map[drow(row,i)][dcol(col,i)] + dist[row][col];
		parent[drow(row,i)][dcol(col,i)][0] = row;
		parent[drow(row,i)][dcol(col,i)][1] = col;
		/* Push the modified square onto the priority queue */
		pqueue_insert(pq, drow(row,i), dcol(col,i), dist[drow(row,i)][dcol(col,i)] + h(drow(row, i), dcol(col, i), endr, endc));
		pushes++;
	    } else if (debug_level >= 5) {
		printf("\t\t(%d,%d) Weight Stays: %d (would have been %d)\n", drow(row,i), dcol(col,i), dist[drow(row,i)][dcol(col,i)], dist[row][col] + map[drow(row,i)][dcol(col,i)]);
	    }
	}
    }
    return;
}

/* (void) limit_path(void)
 * This routine will set the parent of every square which is not on the
 * shortest path we selected between the starting and ending points.
 * It does this by tracing back the path and multiplying the distance of
 * the shortest path to that point by -1 to identify the squares which
 * lie on the path.
 * Then, the entire set of squares is run through to return the distance
 * values to their original values and clear the parent information for
 * all squares in the path.
 * This is not a pretty way of getting the path but it's the easiest way
 * to do it in place in such a way that we can easily dump the map to
 * the screen.
 */

void
limit_path(void)
{
int row, col;
int tmp;

    if (endr < 0 || endr >= rows_read || endc < 0 || endc >= cols_read) return;
    row = endr;
    col = endc;
    do {
	dist[row][col] = -dist[row][col];
	tmp = parent[row][col][0];
	col = parent[row][col][1];
	row = tmp;
    } while(row >= 0 && col >= 0);

    for (row = 0; row < rows_read; row++)
	for (col = 0; col < cols_read; col++)
	    if (dist[row][col] < 0)
		dist[row][col] = -dist[row][col];
	    else
		parent[row][col][0] = -1;
    return;
}

void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[]) 
{
  const int *arg_dims;
  int arg_ndims;
  int msize[3];
  int x,y,rows,cols;
  int arg_rows,arg_cols;
  int start_rows,start_cols,end_rows,end_cols;
  int *matlab_data_in,*matlab_data_parent,*matlab_data_dist;
  int *start, *end;
  int i,j;

  if(nrhs ==0) {
    mexPrintf("Usage: [path dist] = gridAStar( grid, start, goal)\n");
    return;
  }

  if(nrhs!=3) {
    mexPrintf("Error: 3 input arguments expected.\n");
    return;
  }

  /* Parse arg 1 which is the map image */
  msize[0]=-1;msize[1]=-1;
  if(!type_check("image",prhs[0],mxINT32_CLASS,2,msize))
  {
    mexPrintf("Checking of input image type failed.\n");
    return;
  }

  arg_dims = mxGetDimensions(prhs[0]);
  arg_rows=arg_dims[0];
  arg_cols=arg_dims[1];
  
  matlab_data_in = (int*)mxGetPr(prhs[0]);
  


  // Need to reallocate the row pointers (which will be to the rows in Matlab's representation. Note that we are switching rows and columns
  // here to piece things together more easily

    
  SIZE_R = arg_cols;
  SIZE_C = arg_rows;

  rows_read = SIZE_R;
  cols_read = SIZE_C;

    /*mexPrintf("The input map had %d rows and %d cols.\n",rows_read,cols_read);*/

  pops = 0;
  pushes = 0;

  if( (map = (int**)malloc(SIZE_R*sizeof(int*))) == NULL ){
    mexPrintf("Error allocating map data.\n");
    return;
  }

  // Now point our row pointers to the right places in the already existing matlab data

  for( i=0; i<SIZE_R; i++ )
  {
    map[i] = NULL;

    if( (map[i] = matlab_data_in+i*SIZE_C) == NULL )
    {
        mexPrintf("Error allocating map data.\n");
        return;
    }
  }

  if( (dist = (int**)malloc(SIZE_R*sizeof(int*)))==NULL){
    mexPrintf("Error allocating distance matrix.\n");
    return;
    }


  if( (parent = (int***)malloc(SIZE_R*sizeof(int**)))==NULL){
    mexPrintf("Error allocating parent matrix.\n");
    return;
  }

    /* Allocate the return structure for parent and point to it properly locally (mixing rows/cols)*/
  int numDims = 3;
  msize[0]=arg_rows;
  msize[1]=arg_cols;
  msize[2]=2;
  plhs[1]=mxCreateNumericArray(numDims,msize,mxINT32_CLASS,mxREAL);
  matlab_data_parent=(int*)mxGetPr(plhs[1]);

   /* Now likewise for dist */
  numDims = 2;
  msize[0]=arg_rows;
  msize[1]=arg_cols;
  plhs[0]=mxCreateNumericArray(numDims,msize,mxINT32_CLASS,mxREAL);
  matlab_data_dist=(int*)mxGetPr(plhs[0]);


  for( i=0; i<SIZE_R; i++ )
  {
    dist[i] = NULL;
    if( (dist[i] = matlab_data_dist+i*SIZE_C) == NULL )
    {
        mexPrintf("Error pointing to dist data.\n");
        return;
    }
 /*    if( ( dist[i] = (int*)malloc(SIZE_C*sizeof(int)))== NULL ){
        mexPrintf("Error allocating distance data.\n");
        return;
    }
*/
    parent[i] = NULL;
    if( (parent[i] = (int**)malloc(SIZE_C*sizeof(int*)))==NULL){
        mexPrintf("Error allocating parent matrix.\n");
        return;
    }
    
    for( j=0; j<SIZE_C; j++ )
    {
        parent[i][j] = NULL;
/*
        if((parent[i][j] = matlab_data_parent+2*(i*SIZE_C+j))==NULL){
            mexPrintf("Error pointing to parent data.\n");
            return;
        }
*/
        if( (parent[i][j]=(int*)malloc(2*sizeof(int)))==NULL){
            mexPrintf("Error allocating parent data.\n");
            return;
        }

        dist[i][j] = INFINITY; 
        parent[i][j][0] = -1;
        parent[i][j][1] = -1;
    }
  }

  arg_dims = mxGetDimensions(prhs[1]);
  start_rows=arg_dims[0];
  start_cols=arg_dims[1];

  if( start_rows != 1 || start_cols != 2 )
  {
    mexPrintf("Start had %d rows and %d cols. Need 2x1.\n",start_rows,start_cols);
    return;
  }

  start = (int*)mxGetPr(prhs[1]);
  
  startr = start[1]-1;
  startc = start[0]-1;

  /*mexPrintf("Starting row: %d, col: %d.\n", startr,startc);*/

  arg_dims = mxGetDimensions(prhs[2]);
  end_rows=arg_dims[0];
  end_cols=arg_dims[1];

  if( start_rows != 1 || start_cols != 2 )
  {
    mexPrintf("End had %d rows and %d cols. Need 2x1.\n",end_rows,end_cols);
    return;
  }
  
  end = (int*)mxGetPr(prhs[2]);

  endr = end[1]-1;
  endc = end[0]-1;

  /*mexPrintf("Ending row: %d,col %d.\n", endr,endc);*/

  build_path();

  /*mexPrintf("The value of the map at the start was %d.\n",map[startr][startc]);
  mexPrintf("The value of the map at the end was %d.\n",map[endr][endc]);
  mexPrintf("The distance to the goal is %d.\n",dist[endr][endc]);
  mexPrintf("The last step to the goal is %d, %d.\n",parent[endr][endc][0],parent[endr][endc][1]);
  */

  for(i=0;i<arg_cols;i++) {
      for(j=0;j<arg_rows;j++) {
        matlab_data_parent[j+i*arg_rows+0*arg_rows*arg_cols]=parent[i][j][1]+1;
        matlab_data_parent[j+i*arg_rows+1*arg_rows*arg_cols]=parent[i][j][0]+1;
      }
    }

  return;
}

