function [ vecDesiredPoseOdom ] = moveBot(odom, rPose, rGridPos, yawDelta, distDelta, yawDesired)
% [] = moveBot (odom, rPose, rGridPos, yawDelta, distDelta, yawDesired)
%
% Issues a non-blocking movement command to move along a vector distDelta from
% the current position, where (yawDelta, distDelta) are the (theta,r) polar
% coordinates of a vector from the robot's current physical position. 
% 
% This method abstracts away the notion of odometry- or slam- coordinate frames
% so that you can issue movement commands that are always relative to the
% position of the robot. The exception to this is yawDesired, which is given in
% the SLAM coordinate frame.
%
% This method _assumes_ that odom, rPose, and rGridPos are up-to-date. If they
% are not up-to-date, performance may vary. However, having old data is still
% not as bad as having out-of-sync data, so don't try to insert your own
% 'getodom' commands before or after this method.
%
% If you _must_ get new data immediately before this command, try calling
% doSlam() and fetch the refreshed / filtered pose information from that
% function.
%
% Difference between yawDelta and yawDesired: 
%   yawDelta is the absolute angle between robot and goal
%   yawDesired is the absolute angle that the robot should assume _at_ the goal

global configPars;

% Find the angle diff that takes the SLAM frame to the odom frame:
yawDiff = thetapp(odom(3) - rPose(3));

% Take the desired yaw (which is in SLAM coords) into the odometry frame:
yawDesiredOdom = mod(yawDesired + yawDiff + 2*pi, 2 * pi);

% Calc the vector delta of the destination in the odom frame:
vecPositionOdom = odom(1:2) + distDelta .* [ cos(yawDelta) sin(yawDelta) ];

% Calc the pose of the destination in the odom frame:
vecDesiredPoseOdom = [ vecPositionOdom(1) vecPositionOdom(2) yawDesiredOdom ];

% Print the command
if configPars.PRINTLEVEL >= 4
    fprintf('[moveBot.m] powerbot(''movebotTo'', [ %.2f %.2f %.2f ])\n', vecPositionOdom(1), vecPositionOdom(2), yawDesiredOdom);
end;

% Issue the non-blocking powerbot command with instructions in the odom frame:
powerbot('movebotTo', vecDesiredPoseOdom);

