/*********************************************************************
	Module: simulator.c (A MATLAB C module that performs an RPC call)
	Author: Marius Muja (2007)
	Input: a real matrix
	Output: a real row vector
***********************************************************************/

#define SHM_NAME "/srvc_shm"
#define SHM_SIZE 10*1024*1024
#define SEM_NAME "/srvc_sem"


#include "mex.h"
#include <stdio.h>
#include <string.h>
#include <semaphore.h>
#include <fcntl.h>

#include <opencv/cv.h>
#include <opencv/highgui.h>


#include "shared_memory.h"
#include "bumblebee.h"

typedef struct {
   int width;
   int height;
   int bpp;
} Image;


void read_from_shared_memory(SharedMemory& shmem, IplImage** dst)
{
    IplImage* h = shmem.reserve<IplImage>();
    *dst = h;
    char* imageData = shmem.reserve<char>(h->imageSize);
    h->imageData = imageData;
}

template <typename T>
void mat_copy(mxArray* dst, IplImage* src)
{
    T* dstPtr = (T*) mxGetData(dst);
    T* srcPtr = (T*) src->imageData;

    for (int i=0;i<src->height;++i) {
        for (int j=0;j<src->width;++j) {
            int row_skip = j*src->height+i;
            int mat_size = src->width*src->height;
            for (int k=src->nChannels-1;k>=0;--k) {
                *(dstPtr+k*mat_size+row_skip) = *(srcPtr++);
            }
        }
    }
}


mxArray* IplImage2Matlab(IplImage* img)
{
    int sizes[3];
    sizes[0] = img->height;
    sizes[1] = img->width;
    sizes[2] = img->nChannels;
    
    int dims = 3;
    if (sizes[2]==1) {
        dims = 2;
    }
    
    mxArray* mat;
            
    switch (img->depth) {
        case IPL_DEPTH_8U:
            mat = mxCreateNumericArray(dims, sizes, mxUINT8_CLASS, mxREAL);
            mat_copy<unsigned char>(mat,img);
            break;
        
        case IPL_DEPTH_16U:
            mat = mxCreateNumericArray(dims, sizes, mxUINT16_CLASS, mxREAL);
            mat_copy<unsigned short>(mat,img);
            break;
        
        case IPL_DEPTH_32F:
            mat = mxCreateNumericArray(dims, sizes, mxSINGLE_CLASS, mxREAL);
            mat_copy<float>(mat,img);
            break;
    }
    
    return mat;
}

mxArray* read_objects_from_shared_memory( SharedMemory &shmem )
{
	int* num_obj = shmem.reserve<int>();
	double* objects = shmem.reserve<double>(12*(*num_obj));
	double *mxObject;

	int sizes[2];
    sizes[0] = 12;
    sizes[1] = *num_obj;

	mxArray* mat = mxCreateNumericArray(2, sizes, mxDOUBLE_CLASS, mxREAL);
	mxObject = mxGetPr( mat );

	memcpy( mxObject, objects, 12*(*num_obj)*sizeof(double) );

	return mat;
}

mxArray* read_pc_from_shared_memory( SharedMemory &shmem )
{
        int* num_pts = shmem.reserve<int>();
        double* points = shmem.reserve<double>(6*(*num_pts));
        double *mxPoint;

        int sizes[2];
    sizes[0] = 6;
    sizes[1] = *num_pts;

        mxArray* mat = mxCreateNumericArray(2, sizes, mxDOUBLE_CLASS, mxREAL);
        mxPoint = mxGetPr( mat );

        memcpy( mxPoint, points, 6*(*num_pts)*sizeof(double) );

        return mat;
}


template <typename T>
T read_data_from_shared_memory(SharedMemory &shmem)
{
    T* ptr = shmem.reserve<T>();
    return *ptr;
}

void mexFunction(int nOutArray, mxArray *OutArray[], int nInArray, const mxArray *InArray[])
{
	if( nOutArray!=2 && nOutArray != 6 && nOutArray != 8 )
	{
		mexErrMsgTxt("bumblebeeCapture expects 2,6 or 8 outputs.");
	}
		
    SharedMemory shmem(SHM_NAME);
    if (shmem.attach() == NULL) {
        mexErrMsgTxt("Cannot read. Shared memory is not available.");
    }

    //    sem_unlink("srvc_sem");
    sem_t* sem = sem_open(SEM_NAME,O_CREAT,S_IRWXU | S_IRWXG,1);
    if (sem==SEM_FAILED) {
        printf("Cannot create semaphore\n");
        exit(1);
    }
    sem_wait(sem);

    
    int size = 1;

    OutArray[0] = mxCreateNumericArray(1, &size, mxDOUBLE_CLASS, mxREAL);
    double* ptr = mxGetPr( OutArray[0] );
    *ptr = read_data_from_shared_memory<float>(shmem);
    
    OutArray[1] = read_objects_from_shared_memory(shmem);
    
    for (int i=2;i<((nOutArray<7) ? nOutArray: 7);++i) {
        IplImage* img;
	if(i==5)
	{	OutArray[i] = read_pc_from_shared_memory(shmem);
	}
	else
	{
        	read_from_shared_memory(shmem,&img);
		OutArray[i] = IplImage2Matlab(img);
	}
   	 
    }

    if( nOutArray > 7 )
    {
	   OutArray[7] = read_objects_from_shared_memory(shmem);
    }
    

    sem_post(sem);
    
}
