#ifndef _MATCH_H_
#define _MATCH_H_

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <algorithm>
#include <functional>
#include <vector>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include "flann.h"

// SIFT Keypoints includes
extern "C" {
#include "key.h"
}

using namespace std;





const int MAX_FEATURES = 200000;
const int MAX_CAT = 100;
const int MAX_TEST_SIZE = 10000;
const int NN = 10;


const int SCALE_BINS = 20;
const int SCALE_BINS_2 = SCALE_BINS/2;
const int ORI_BINS = 12;

const int N_MATCH = 10;
const int K_MATCH = 2;
const float DIST_EPS = 1.5;

const int TOP_MATCHES = 10;

float distSquared(float *v1, float *v2, int veclen)
{
    float diff, distsq = 0.0;
    float diff0, diff1, diff2, diff3;
    float *final, *finalgroup;

    final = v1 + veclen;
    finalgroup = final - 3;

    /* Process 4 pixels with each loop for efficiency. */
    while (v1 < finalgroup) {
      diff0 = v1[0] - v2[0];
      diff1 = v1[1] - v2[1];
      diff2 = v1[2] - v2[2];
      diff3 = v1[3] - v2[3];
      distsq += diff0 * diff0 + diff1 * diff1 + diff2 * diff2 + diff3 * diff3;
      v1 += 4;
      v2 += 4;
    }
    /* Process last 0-3 pixels.  Not needed for standard vector lengths. */
    while (v1 < final) {
      diff = *v1++ - *v2++;
      distsq += diff * diff;
    }
    return distsq;
}



class SiftMatcher {

    struct SiftKeypoint{
        float row, col;      /* Row, column location relative to input image.  */
        float scale;         /* Scale (in sigma units for smaller DOG filter). */
        float ori;           /* Orientation in radians (-PI to PI). */
        int cat;
        int img;
    };

    SiftKeypoint* feature_info;
    float* features;
    int count;
    
    SiftKeypoint* testset_info;
    float* testset;    
    int test_count;
    
    int* matches;

    char test_image_name[100];
    int size;
    int veclen;
    FLANN_INDEX index;
    IndexParameters index_params;
    
    int cat_count;
    int* image_count;
    int max_image_count;
    float max_scale;
    float min_scale;
    
    int* rev_lookup;
    
    char* feature_path;
    
    
    struct Match {
        int first;
        int second;
        Match* next;
    };
    
    class MatchCmp {
        Match* m;
        SiftKeypoint* testset_info;
    public:
        MatchCmp(Match* ref, SiftKeypoint* info) : m(ref), testset_info(info) {};
        
        float sq_dist(SiftKeypoint a, SiftKeypoint b)
        {
            return ((a.row-b.row)*(a.row-b.row)+(a.col-a.col)*(a.col-a.col));
        }
    
        int operator() (Match* a, Match* b) 
        {
            if (m==a) {
                return true;
            }
            else if (m==b) {
                return false;
            }
            else if (a->first == m->first) {
                return true;
            }
            else if (b->first==m->first) {
                return false;
            }
            else {
                return sq_dist(testset_info[m->first],testset_info[a->first])<sq_dist(testset_info[m->first],testset_info[b->first]);
            }
            
        }
    };
    
    struct Bin {
        int img;
        int cat;
        int scale;
        int ori;
        int count;
        Match* m;
    };
    
    class BinCmp {
    public:
        int operator() (Bin a, Bin b) 
        {
            greater<int> cmp;
            return cmp(a.count,b.count);
        }
    };

    bool is_directory(char* name) 
    {
        struct stat s;
        int ret  = stat(name,&s);
        if (ret==0 && S_ISDIR(s.st_mode)) {
            return true;
        }        
        return false;
    }

    bool is_file(char* name) 
    {
        struct stat s;
        int ret  = stat(name,&s);
        if (ret==0 && S_ISREG(s.st_mode)) {
            return true;
        }        
        return false;
    }



    void loadClassFeatures(char* path, int cat)
    {
        int img = 0;
        char name[500];
        sprintf(name,"%s/%d.sift",path,img);
        
        while (is_file(name)) {
            printf("Loading SIFT features from: %s\n",name);
            loadSiftFile(name,cat,img);
            
            img++;
            sprintf(name,"%s/%d.sift",path,img);
        }
        image_count[cat] = img;
        if (img>max_image_count) {
            max_image_count = img;
        }
    }
    
    void loadSiftFile(char* path, int cat, int img)
    {
/*        if (cat>=cat_count) {
            cat_count = cat+1;
        }*/
        
        FILE* file = fopen(path,"r");
        if (file==NULL) {
            perror("Error opening file");
            exit(1);
        }
    
        count += loadSiftFeatures(file, features+count*veclen, feature_info+count, cat, img);
        
        
        for (int i=0;i<count;++i) {
            if (max_scale<feature_info[i].scale) {
                max_scale = feature_info[i].scale;
            }
            if (min_scale>feature_info[i].scale) {
                min_scale = feature_info[i].scale;
            }
        }
        
        fclose(file);
    }
    
    
    int loadSiftFeatures(FILE* file, float* ptr, SiftKeypoint* info_ptr, int cat, int img)
    {
        int cnt;
        int vlen;
        
        fread(&cnt, sizeof(cnt), 1, file);
        fread(&vlen, sizeof(vlen), 1, file);       
        printf("File has %d features of size %d\n",cnt,vlen);
        assert(vlen==veclen);
        
        // read feature data
        for (int i=0;i<cnt;++i) {
            unsigned char values[1000];
            fread(&values, vlen, 1, file);
            
            for (int j=0;j<vlen;++j) {
                ptr[j] = values[j];
            }
            ptr += vlen;
        }
        
        // read feature info data
        for (int i=0;i<cnt;++i) {
            fread(&info_ptr[i].scale, sizeof(float), 1, file);
            fread(&info_ptr[i].ori, sizeof(float), 1, file);
            fread(&info_ptr[i].row, sizeof(float), 1, file);
            fread(&info_ptr[i].col, sizeof(float), 1, file);
            
            info_ptr[i].cat = cat;
            info_ptr[i].img = img;
            
        }
                
        return cnt;
    }



    int get_bin_number(float value, float bin_size) 
    {
        int bin_no = 0;
        
        if (value>=0) { 
            while (value >= bin_size) {
                bin_no++;
                value -= bin_size;
            }
        }
        else {
            while (value <= -bin_size) {
                bin_no--;
                value += bin_size;
            }
        }
        
        return bin_no;
    }
    
    float get_orientation_difference(SiftKeypoint a, SiftKeypoint b)
    {
        float diff = a.ori - b.ori;
        
        float two_pi = 2*M_PI;
        
        if (diff<0) {
            diff += two_pi;
        } 
        else if (diff>two_pi) {
            diff -= two_pi;
        }
        
        return diff;
    }
    
    float normalized_sq_dist(SiftKeypoint a, SiftKeypoint b)
    {
        return ((a.row-b.row)*(a.row-b.row)+(a.col-a.col)*(a.col-a.col))/(a.scale*a.scale);
    }
    
    
    float largest_ratio(float v) 
    {
        return (v>1)?v:(1/v);
    }
    
    bool is_consistent_match(Match* m1, Match* m2)
    {
        float dist1 = normalized_sq_dist(testset_info[m1->first],testset_info[m2->first]);
        float dist2 = normalized_sq_dist(feature_info[m1->second],feature_info[m2->second]);
//         printf("Dist1: %g\n",dist1);
//         printf("Dist2: %g\n",dist2);
        
        float dist_ratio = largest_ratio(dist1/dist2);
                
//         printf("Dist ratio: %g\n",dist_ratio);
        if (dist_ratio<DIST_EPS) {
            return true;
        }
        return false;
    }
    
    Match* geometric_consistency(Match* m, int &count)
    {
        int size = count;
        vector<Match*> matches(count);
        
        int k = 0;
        for (Match* crt = m;crt!=NULL;crt = crt->next) {
            matches[k++] = crt;
        }
        
        for (Match* crt = m;crt!=NULL;crt = crt->next) {
            MatchCmp compare(crt,testset_info);
            int n = min(N_MATCH,size);            
            partial_sort(matches.begin(),matches.begin()+n,matches.begin()+size,compare);
            int cnt = 0;
            assert(crt==matches[0]);
            for (int i=1;i<n;++i) {
                if (is_consistent_match(crt,matches[i])) {
                    cnt++;
                }
            }
            if (cnt<K_MATCH) {
                swap(matches[0],matches[size-1]);
                --size;
            }
        }
        
        count = size;
        if (size>0) {
            for (int i=0;i<size;++i) {
                matches[i]->next = matches[i+1];
            }
            matches[size-1]->next = NULL;
            for (int i=size;i<count;++i) {
                free(matches[i]);
            }
            return matches[0];
        }
        else {
            return NULL;
        }
        
    
    
/*      if (m!=NULL) {
            Match* crt = m;
            Match* prev = NULL; 
            while (crt!=NULL) {
                int cnt = 0;
                int good = 0;
                Match* ptr = crt->next;
                if (ptr == NULL) ptr = m;
                while (cnt<N_MATCH && ptr != crt) {
                    if (is_consistent_match(crt,ptr)) {
                        good++;
                    }
                    ptr = ptr->next;
                    if (ptr == NULL) ptr = m;
                    cnt++;
                }
                if (good < K_MATCH) { //elliminate start node
                    if (crt==m) {
                            prev = m;
                            m = m->next;
                            crt = m;
                            free(prev);
                            prev = NULL;
                    }
                    else { //remove node from list
                        prev->next = crt->next;
                        free(crt);
                        crt = prev->next;
                    }
                } 
                else {
                    prev = crt;
                    crt = crt->next;
                }
            }
        }*/
    }


public:
    
    SiftMatcher()
    {
        size = MAX_FEATURES;
        veclen = 128;
        count = 0;
        features = (float*) malloc(size*veclen*sizeof(float));
        feature_info = (SiftKeypoint*) malloc(size*sizeof(SiftKeypoint));
        image_count = (int*) malloc(MAX_CAT);
        rev_lookup = NULL;
        
        testset = (float*) malloc(MAX_TEST_SIZE*veclen*sizeof(float));
        testset_info = (SiftKeypoint*) malloc(MAX_TEST_SIZE*sizeof(SiftKeypoint));

        matches = (int*)malloc(MAX_TEST_SIZE*NN*sizeof(int));
        
        cat_count = 0;
        max_image_count = 0;
        max_scale = 0;
        min_scale = 1e10;
    }
    
    ~SiftMatcher()
    {
        free(features);
        free(feature_info);
        free(image_count);
        if (rev_lookup!=NULL) delete rev_lookup;
        free(testset);
        free(testset_info);
        free(matches);
    }
    

    void loadFeatures(char* path)
    {    
        feature_path = path;
    
        int c = 0;
        
        char name[500];
        sprintf(name,"%s/%d",path,c);
        
        
        while (is_directory(name)) {
            loadClassFeatures(name,c);
            
            c++;
            sprintf(name,"%s/%d",path,c);
        }
        
        cat_count = c;
        printf("Total: %d features\n",count);
        printf("Categories: %d\n",cat_count);
        
        
        // fill in reverse lookup table
        rev_lookup = (int*) malloc(cat_count*max_image_count*sizeof(int));
        for (int i=0;i<count;++i) {
            rev_lookup[feature_info[i].cat*cat_count+feature_info[i].img] = i;
        }        
    }
    

    
    void build_index()
    {
        FLANNParameters fp;
        fp.log_level = LOG_INFO;
        fp.log_destination = NULL;
        fp.random_seed = 112; //just a number
        
        index_params.checks=256;
        index_params.algorithm = KDTREE;
        index_params.trees=8;
        index_params.branching=32;
        index_params.iterations=7;
        index_params.target_precision = -1;
        float speedup;
        
        printf("Computing index.\n");
        index = flann_build_index(features, count, veclen, &speedup, &index_params, &fp);
        printf("Done computing index.\n");
//        flann_find_nearest_neighbors_index(index_id, testset, tcount, result, nn, p.checks, &fp);
    }
    

    
    

    
    void show_matches(char* test_name, char* matched_name, Match* m)
    {
        IplImage* test_img = cvLoadImage(test_name);
        IplImage* matched_img = cvLoadImage(matched_name);
        CvMat sub_rect;
            
        CvSize size;
        size.height = max(test_img->height, matched_img->height);
        size.width = test_img->width + matched_img->width;
//         int small_width = size.width/4;
//         int small_height = size.height/4;
               
        IplImage* result_image = cvCreateImage(size, test_img->depth, test_img->nChannels);
        cvZero(result_image);
        cvGetSubRect( result_image, &sub_rect, cvRect(0,0, test_img->width, test_img->height));
        cvCopy(test_img, &sub_rect);
        cvGetSubRect( result_image, &sub_rect, cvRect(test_img->width,0, matched_img->width, matched_img->height));
        cvCopy(matched_img, &sub_rect);
        
/*        cvResize(matched_img,small_match);
        cvGetSubRect( test_img, &sub_rect, cvRect(0,0, small_width, small_height));
        cvCopy(small_match,&sub_rect);*/
        
//         cvNamedWindow( "Matched image", 1 );
//         cvShowImage("Matched image", matched_img);

        Match* ptr;
        for (ptr = m;ptr!=NULL;ptr = ptr->next) {
        
            CvPoint p1,p2;
            p1.x = (int)testset_info[ptr->first].col;
            p1.y = (int)testset_info[ptr->first].row;
            p2.x = (int)feature_info[ptr->second].col+test_img->width;
            p2.y = (int)feature_info[ptr->second].row;
        
//             printf("(%d,%d) -> (%d,%d)\n",p1.x,p1.y,p2.x,p2.y);
        
            cvLine(result_image,p1,p2,CV_RGB(0,255,0));
        }
        
        for (int i=0;i<test_count;++i) {
            cvCircle(result_image, cvPoint((int)testset_info[i].col,(int)testset_info[i].row), 10, CV_RGB( 0, 255, 0 ));
        }
        
        cvNamedWindow( "Test image", 1 );
        cvShowImage("Test image", result_image);
        
//         static int cnt = 0;
        
/*        char buffer[100];
        sprintf(buffer,"image%3.0d.jpg",cnt++);*/
//         cvSaveImage(buffer,result_image);
        
//         printf("Before\n");
        cvWaitKey(0);
//         sleep(1);
//         printf("After\n");
        
        cvReleaseImage(&result_image);
        cvReleaseImage(&matched_img);
        cvReleaseImage(&test_img);
        
    }
    
    
    void free_matches(Match* m)
    {
        Match* p = m;
        while (m) {
            p = m->next;
            free(m);
            m = p;
        }
    }


    
    void process_matches(SiftKeypoint* test_keys, int* matches, int test_count, int nn) 
    {
        float ori_bin_size = 2*M_PI/ORI_BINS;
        float scale_bin_max = log(max_scale)-log(min_scale);
        float scale_bin_size = ((scale_bin_max+1)*2)/SCALE_BINS;
    
        
        Bin voting_space[cat_count][max_image_count][ORI_BINS][SCALE_BINS];        
        memset(voting_space,0,cat_count*max_image_count*ORI_BINS*SCALE_BINS*sizeof(Bin));
        
        for (int c = 0;c<cat_count;++c) 
        for (int i = 0;i<image_count[c];++i) 
        for (int o = 0;o<ORI_BINS;++o) 
        for (int s = 0;s<SCALE_BINS;++s) {
            voting_space[c][i][o][s].cat = c;
            voting_space[c][i][o][s].img = i;
            voting_space[c][i][o][s].scale = s;
            voting_space[c][i][o][s].ori = o;
        }               

        
        for (int i=0;i<test_count;++i) {            
            SiftKeypoint crt_key = test_keys[i];
            for (int j=0;j<nn;++j) {
                SiftKeypoint nn_key = feature_info[matches[i*nn+j]];
                float orientation = get_orientation_difference(crt_key,nn_key);
                float log_scale = log(nn_key.scale)-log(crt_key.scale);
                int ori_bin = get_bin_number(orientation,ori_bin_size);
                int scale_bin = get_bin_number(log_scale,scale_bin_size)+SCALE_BINS_2;
                
                assert(ori_bin < ORI_BINS);
                if (scale_bin >= SCALE_BINS) {
                    printf("!!!!!!! Out of bin condition. Max scale bins: %d, crt. scale bin: %d\n",SCALE_BINS,scale_bin);
                }
                
                // count vote
                voting_space[nn_key.cat][nn_key.img][ori_bin][scale_bin].count++;
                // add match to list
                Match* ptr = (Match*) malloc(sizeof(Match));
                ptr->first = i;
                ptr->second = matches[i*nn+j];
                ptr->next = voting_space[nn_key.cat][nn_key.img][ori_bin][scale_bin].m;
                voting_space[nn_key.cat][nn_key.img][ori_bin][scale_bin].m = ptr;
                
                // save to alloc list
//                 alloc_list.push_back(ptr);
            }
        }
        
        BinCmp cmp;
        partial_sort((Bin*)voting_space,(Bin*)voting_space+TOP_MATCHES,(Bin*)voting_space+(cat_count*max_image_count*ORI_BINS*SCALE_BINS),cmp);
        

        char matched_image[100];
        
        Bin* best_bin = (Bin*)voting_space;
        bool have_match = false;
        for (int i=0;i<TOP_MATCHES;++i) {
//             Bin* best
            sprintf(matched_image,"%s/%d/%d.jpg",feature_path,best_bin->cat,best_bin->img);
        
            best_bin->m = geometric_consistency(best_bin->m, best_bin->count);
            
            if (best_bin->count>K_MATCH) {
                printf("Best image, cat: %d, img: %d, scale_bin: %d, ori_bin: %d, count:%d\n",best_bin->cat,best_bin->img, best_bin->scale, best_bin->ori, best_bin->count);
    //             Match* m = best_bin->m;
                //show_matches(test_image_name, matched_image, best_bin->m);
                have_match = true;
            }
            else {
                //printf("Skipping not matching image.\n");
            }
            best_bin++;
        }
        if (!have_match) {
            printf("No match found\n");
        }
        
        for (int i = 0;i<cat_count*max_image_count*ORI_BINS*SCALE_BINS;++i) {
            free_matches(((Bin*)voting_space)[i].m);
        }
    }
    
    
    void test_image(char* test_path)
    {
        // loading sift features
        printf("Loading file %s\n",test_path);
        FILE* file = fopen(test_path,"r");
        if (file==NULL) {
            perror("Error opening file");
            exit(1);
        }        
        test_count = loadSiftFeatures(file, testset, testset_info, -1, -1);
        
        FLANNParameters fp;
        fp.log_level = LOG_INFO;
        fp.log_destination = NULL;
        
        flann_find_nearest_neighbors_index(index, testset, test_count, matches, NN, 128, &fp);

        strcpy(test_image_name,test_path);
        char* dot_ptr = strrchr(test_image_name,'.');
        strcpy(dot_ptr,".jpg");

        process_matches(testset_info,matches, test_count, NN);
        
    }
    
    void opencvToNative(IplImage* cvImage, Image* native)
    {
        int x, y;
    
        *native = CreateImage(cvImage->height, cvImage->width, IMAGE_POOL);
        for ( y = 0; y < cvImage->height; ++y ) {
            for ( x = 0; x < cvImage->width; ++x ) {
            (*native)->pixels[y][x] = ((uchar*)(cvImage->imageData + cvImage->widthStep*y))[x] / 255.0f;
            }
        }
    }
    
    void test_image(IplImage* image)
    {
        // extracting sift features
        Image im1 = NULL;
        IplImage* gray_image = NULL;
        gray_image = cvCreateImage(cvGetSize(image),IPL_DEPTH_8U,1); 
        cvCvtColor(image, gray_image, CV_RGB2GRAY); // need gray image
        
        opencvToNative(gray_image, &im1);
        Keypoint keys = GetKeypoints(im1, -1);
        Keypoint p = keys;
        test_count = 0;
        float* ptr = testset;
        while (p) {
            testset_info[test_count].row = p->row;
            testset_info[test_count].col = p->col;
            testset_info[test_count].scale = p->scale;
            testset_info[test_count].ori = p->ori;
            
            for (int j=0;j<veclen;++j) {
                ptr[j] = p->ivec[j];
            }
            ptr += veclen;
            p = p->next;
            
            test_count++;
        }
        
        FreeStoragePool(KEY_POOL);
        FreeStoragePool(IMAGE_POOL);
        
        
        
//         drawKeypoints(left_rect,keys, (double)left_rect->width/sift_image->width);
//         printf("Found %d keypoints\n",length(keys));
        
    
        FLANNParameters fp;
        fp.log_level = LOG_INFO;
        fp.log_destination = NULL;
        
        flann_find_nearest_neighbors_index(index, testset, test_count, matches, NN, 128, &fp);

        process_matches(testset_info,matches, test_count, NN);
    }
    
    void test_image2(char* image_name)
    {
        IplImage* image = cvLoadImage(image_name);
        strcpy(test_image_name, image_name);
        test_image(image);
        cvReleaseImage(&image);
    }

    
};


/*

int main(int argc, char** argv)
{
    if (argc<3) {
        printf("Usage: %s train_images_path test_image\n",argv[0]);
        exit(1);
    }
    SiftMatcher sm;
    sm.loadFeatures(argv[1]);
    sm.build_index();
    sm.test_image2(argv[2]);
    
    return 0;
}*/



#endif //_MATCH_H_
