function [ bestSample ] = sampleFromGoalSpace ( worldState, mappingPars, planningPars, configPars, photoRewardMap )
% [ bestSample ] = sampleFromGoalSpace ( worldState, mappingPars, planningPars, configPars, photoRewardMap )
%
% Create a batch of samples. We'll need two random vars for the location,
% and one random var for the pose for each row.
% 
% bestSample has the form ( x y theta )
%

% These parameters are described in InitializeRobot.m:
batchSize = planningPars.batchSize;
maxEpochs = planningPars.maxEpochs;
sampleTimeLimit = planningPars.sampleTimeLimit;

% Create the initial samples:
samples = createPoseSamples ( batchSize, configPars, worldState );
epoch = 1;

% Randomly assign the best sample to start off:
bestSample = samples(1,:);
bestSampleScore = -inf;
newBestEpoch = 0;

% Simulated annealing:
% Set up two evolutionary profiles, one for strong exploitation and another for
% strong exploration. If we've just found a new best sample, favor strong
% exploitation. If we've had the same best sample for a while, favor strong
% exploration. The first number is the culling factor. The second number is an
% elitism factor.
exploration = [ 0.5 0.0 ];
exploitation = [ 0.8 0.1 ];

% Loop a fixed number of times
% (?todo: change this so that we check for convergence)
while epoch <= maxEpochs && (toc - worldState.time) < sampleTimeLimit
    
    % Evaluate all the samples once
    [ sampleRewards sampleScores ] = evaluateSamples ( samples, worldState, mappingPars, planningPars, configPars, photoRewardMap );
    
    % Sort the samples by score. Avoid duplicates.
    [ sortedScores sortedIdx ] = unique(sampleScores);
    samples = samples(sortedIdx,:);
    
    % Check the best sample
    if bestSampleScore < sortedScores(1)
        newBestEpoch = epoch;
        bestSampleScore = sortedScores(1);
        bestSample = samples(1,:);
    end;
    
    % If we need to bail, might as well do it now
    if (epoch > maxEpochs && (toc-worldState.time) >= sampleTimeLimit)
        break;
    end;
    
    % Calculate how long the highest-scoring sample has been known
    annealTime = epoch - newBestEpoch;
    w = (2 / ( 1 + 2.^(-annealTime))) - 1;
    annealParms = w .* exploitation + (1-w) .* exploration;
    
    % Cull the bad samples
    numReplacements = floor(batchSize * annealParms(1));
    numRemain = batchSize - numReplacements;
    lower = (batchSize-numReplacements+1);
    upper = batchSize;
    samples(lower:upper,:) = zeros(numReplacements,3);
    
    % Check for a batch that all scored zero, jic.
    batchScore = sum(sampleScores(1:numRemain));
    if (batchScore <= 0)
        fprintf('[sampleFromGoalSpace.m] Warning: batch of samples all scored zero.\n');
    end;
    
    % Make a few copies of the best sample
    copies = floor(annealParms(2) * numReplacements);
    numReplacements = numReplacements - copies;
    upper = lower + copies - 1;
    samples(lower:upper,:) = repmat(samples(1,:),copies,1);
    
    % Permute the existing samples
    samples = permuteSamples ( samples, worldState );
    
    % Add in enough random samples so that we make up a full batch
    newSamples = createPoseSamples ( numReplacements, configPars, worldState );
    lower = batchSize - numReplacements + 1;
    upper = batchSize;
    samples(lower:upper,:) = newSamples;
    
    % Count epochs:
    epoch = epoch + 1;
    
end;

if configPars.PRINTLEVEL >= 4
    fprintf('[sampleFromGoalSpace.m] Debug: performed %d sample epochs.\n', epoch);
end;

if configPars.PRINTLEVEL >= 3 && bestSampleScore <= 0
    fprintf('[sampleFromGoalSpace.m] Warning: best sample had score of %d.\n', bestSampleScore);
end;

% At the end of this loop, bestSample is the sample that scored the highest out
% of all the batches.
